package gov.va.vss.model;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import java.time.Month;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;

import javax.persistence.Column;
import javax.persistence.Convert;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.Table;
import javax.persistence.Transient;
import javax.validation.constraints.Max;
import javax.validation.constraints.Min;
import javax.validation.constraints.NotNull;

import org.apache.commons.lang3.builder.CompareToBuilder;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.hibernate.validator.constraints.Length;
import org.hibernate.validator.constraints.NotBlank;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonProperty;

import gov.va.shared.model.core.AbstractAuditedVersionedPersistent;
import gov.va.vss.model.lookup.NACStatus;
import gov.va.vss.model.lookup.VtkInstitution;
import gov.va.vss.model.lookup.VtkOrganizationType;
import gov.va.vss.persistence.conversion.OrganizationScopeTypeConverter;

@Entity
@Table(name = "VtkVolunteerOrganizations", schema = "dbo")
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public class Organization extends AbstractAuditedVersionedPersistent<Organization>
		implements Comparable<Organization>, ObjectScopedToStationNumbers {
	private static final long serialVersionUID = 6904844123870655771L;

	// -------------------------------------- Fields

	@NotNull
	private OrganizationScopeType scope;
	@NotNull(message = "Type is required.")
	private VtkOrganizationType type;

	private int code;
	@NotBlank(message = "Display Name is required.")
	@Length(max = 35)
	private String name;
	@NotBlank(message = "Full Name is required.")
	@Length(max = 255)
	private String fullName;
	@Length(max = 6)
	private String abbreviation;
	private boolean inactive;
	private VtkInstitution vtkInstitution;
	private boolean onNationalAdvisoryCommittee;
	private boolean nacExecutiveMember;
	private NACStatus nacMembershipStatus;

	@Min(1)
	@Max(12)
	private Integer annualJointReviewMonth;

	// -------------------------------------- Business Methods

	@Transient
	public Collection<String> getScopedToStationNumbers() {
		VtkInstitution institution = getVtkInstitution();
		if (institution == null)
			return new ArrayList<>();
		return Arrays.asList(institution.getStationNumber());
	}

	@Transient
	public String getAJRMonthName() {
		if (this.annualJointReviewMonth != null) {
			return Month.of(annualJointReviewMonth.intValue()).name();
		} else {
			return "";
		}
	}

	@Transient
	public boolean isActive() {
		return !inactive;
	}
	
	@Transient
	@JsonProperty
	public String getDisplayName(){
		return getName();
	}


	// -------------------------------------- Common Methods

	@Override
	protected boolean requiredEquals(Organization oo) {
		return new EqualsBuilder().append(code, oo.getCode())
				.append(nullSafeGetId(vtkInstitution), nullSafeGetId(oo.getVtkInstitution())).isEquals();
	}

	@Override
	protected int requiredHashCode() {
		return new HashCodeBuilder().append(code).append(nullSafeGetId(vtkInstitution)).toHashCode();
	}

	@Override
	public int compareTo(Organization o) {
		if (equals(o))
			return 0;

		return new CompareToBuilder().append(nullSafeLowercase(name), nullSafeLowercase(o.getName()))
				.append(code, o.getCode()).append(nullSafeGetId(vtkInstitution), nullSafeGetId(o.getVtkInstitution()))
				.toComparison() > 0 ? 1 : -1;
	}

	public String toString() {
		return getName() + " (code " + getCode() + ", station ID " + nullSafeGetId(vtkInstitution) + ")";
	}

	// -------------------------------------- Accessor Methods
	
	
	@Column(name = "CodeNUM", length = 5)
	public int getCode() {
		return code;
	}

	public void setCode(int code) {
		this.code = code;
	}

	@Column(length = 6)
	public String getAbbreviation() {
		return abbreviation;
	}

	public void setAbbreviation(String abbreviation) {
		this.abbreviation = abbreviation;
	}

	@Column(name = "OrganizationName", length = 35, nullable = false)
	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	@Column(name = "FullName", length = 255, nullable = false)
	public String getFullName() {
		return fullName;
	}

	public void setFullName(String fullName) {
		this.fullName = fullName;
	}

	@Column(name = "IsInactive", nullable = false)
	public boolean isInactive() {
		return inactive;
	}

	public void setInactive(boolean inactive) {
		this.inactive = inactive;
	}

	@Column(nullable = false)
	@Convert(converter = OrganizationScopeTypeConverter.class)
	public OrganizationScopeType getScope() {
		return scope;
	}

	public void setScope(OrganizationScopeType scopeType) {
		this.scope = scopeType;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "OrganizationTypeFK")
	public VtkOrganizationType getType() {
		return type;
	}

	public void setType(VtkOrganizationType type) {
		this.type = type;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "StationFK")
	public VtkInstitution getVtkInstitution() {
		return vtkInstitution;
	}

	public void setVtkInstitution(VtkInstitution station) {
		this.vtkInstitution = station;
	}

	@Column(name = "IsOnNationalAdvisoryCommittee", nullable = false)
	public boolean isOnNationalAdvisoryCommittee() {
		return onNationalAdvisoryCommittee;
	}

	public void setOnNationalAdvisoryCommittee(boolean onNationalAdvisoryCommittee) {
		this.onNationalAdvisoryCommittee = onNationalAdvisoryCommittee;
	}

	@Column(name = "IsNACExecutiveMember", nullable = false)
	public boolean isNacExecutiveMember() {
		return nacExecutiveMember;
	}

	public void setNacExecutiveMember(boolean nacExecutiveMember) {
		this.nacExecutiveMember = nacExecutiveMember;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "NACMembershipStatusFK")
	public NACStatus getNacMembershipStatus() {
		return nacMembershipStatus;
	}

	public void setNacMembershipStatus(NACStatus nacMembershipStatus) {
		this.nacMembershipStatus = nacMembershipStatus;
	}

	public Integer getAnnualJointReviewMonth() {
		return annualJointReviewMonth;
	}

	public void setAnnualJointReviewMonth(Integer annualJointReviewMonth) {
		this.annualJointReviewMonth = annualJointReviewMonth;
	}

}
