package gov.va.vss.model.lookup;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Transient;
import javax.validation.constraints.NotNull;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.builder.CompareToBuilder;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.hibernate.annotations.BatchSize;
import org.hibernate.annotations.Cache;
import org.hibernate.annotations.CacheConcurrencyStrategy;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnore;

import gov.va.shared.model.core.AbstractAuditedVersionedPersistent;
import gov.va.shared.model.lookup.sds.VAFacility;
import gov.va.vss.model.DonGenPostFund;
import gov.va.vss.model.DonationReference;
import gov.va.vss.model.StationParameters;
import gov.va.vss.model.VoluntaryServiceParameters;

@Entity
@Table(name = "VtkInstitutions_tbl", schema = "dbo")
@Cache(usage = CacheConcurrencyStrategy.NONSTRICT_READ_WRITE)
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public class VtkInstitution extends AbstractAuditedVersionedPersistent<VtkInstitution>
		implements Comparable<VtkInstitution> {
	private static final long serialVersionUID = -8678395783438462990L;

	/**
	 * Factory method to create a new VtkInstitution which reproduces all
	 * trigger-based logic in the DB (creating related entries) - CPB
	 * 
	 * @param facility
	 * @param state
	 * @param type
	 * @param parent
	 * @return
	 */
	public static VtkInstitution createNew(VAFacility facility, VtkState state, VtkInstitutionType type,
			VtkInstitution parent) {
		VtkInstitution newInstitution = new VtkInstitution(facility, state, type, null);
		newInstitution.populateMissingRequiredFields();
		return newInstitution;
	}

	public void populateMissingRequiredFields() {
		StationParameters sp = getStationParameters();
		if (sp == null) {
			sp = new StationParameters(this);
			setStationParameters(sp);
		}
		VoluntaryServiceParameters vsp = getVoluntaryServiceParameters();
		if (vsp == null) {
			vsp = new VoluntaryServiceParameters(StringUtils.left(getName() + " Voluntary Service", 120));
			setVoluntaryServiceParameters(vsp);
		}

		List<DonGenPostFund> funds = getDonGenPostFunds();
		if (funds.isEmpty()) {
			DonGenPostFund f = new DonGenPostFund("None");
			f.setVtkInstitution(this);
			funds.add(f);
		}

		List<DonationReference> donationReferences = getDonationReferences();
		if (donationReferences.isEmpty()) {
			DonationReference dr = new DonationReference("None");
			dr.setVtkInstitution(this);
			donationReferences.add(dr);
		}
	}

	// ---------------------------------------- Fields

	@NotNull
	private String name;
	@NotNull
	private String stationNumber;
	private VtkInstitution parent;
	private VtkAdministrativeUnit administrativeParent;
	private String addressLine1;
	private String addressLine2;
	private String city;
	@NotNull
	private VtkState state;
	private String zip;
	private VtkInstitutionType type;
	private Boolean mealPrintOCX;

	private VAFacility vaFacility;

	/*
	 * Only contains one item, lazy-loaded; mapping like this instead
	 * of @OneToOne so it can be optional and still lazy-load - CPB
	 */
	private List<StationParameters> stationParametersList;

	/*
	 * Only contains one item, lazy-loaded; mapping like this instead
	 * of @OneToOne so it can be optional and still lazy-load - CPB
	 */
	private List<VoluntaryServiceParameters> voluntaryServiceParametersList;

	private List<DonGenPostFund> donGenPostFunds;

	private List<DonationReference> donationReferences;

	// ---------------------------------------- Constructors

	public VtkInstitution() {
	}

	private VtkInstitution(VAFacility facility, VtkState state, VtkInstitutionType type, VtkInstitution parent) {
		this.name = facility.getName();
		this.stationNumber = facility.getStationNumber();
		this.parent = parent;
		if (parent != null)
			this.administrativeParent = parent.getAdministrativeParent();
		this.addressLine1 = facility.getAddressLine1();
		this.addressLine2 = facility.getAddressLine2();
		this.city = facility.getCity();
		this.state = state;
		this.zip = facility.getZip();
		this.type = type;
		this.vaFacility = facility;
	}

	// ---------------------------------------- Business Methods

	@Transient
	@JsonIgnore
	public StationParameters getStationParameters() {
		List<StationParameters> l = getStationParametersList();
		return l.isEmpty() ? null : l.get(0);
	}

	public void setStationParameters(StationParameters stationParameters) {
		getStationParametersList().clear();
		if (stationParameters != null) {
			getStationParametersList().add(stationParameters);
			stationParameters.setVtkInstitution(this);
		}
	}

	@Transient
	@JsonIgnore
	public VoluntaryServiceParameters getVoluntaryServiceParameters() {
		List<VoluntaryServiceParameters> l = getVoluntaryServiceParametersList();
		return l.isEmpty() ? null : l.get(0);
	}

	public void setVoluntaryServiceParameters(VoluntaryServiceParameters vsp) {
		getVoluntaryServiceParametersList().clear();
		if (vsp != null) {
			getVoluntaryServiceParametersList().add(vsp);
			vsp.setVtkInstitution(this);
		}
	}

	@Transient
	public String getDisplayName() {
		return getName() + " (" + getStationNumber() + ")";
	}
	
	// ---------------------------------------- Common Methods

	@Override
	protected boolean requiredEquals(VtkInstitution oo) {
		return new EqualsBuilder().append(getStationNumber(), oo.getStationNumber()).isEquals();
	}

	@Override
	protected int requiredHashCode() {
		return new HashCodeBuilder().append(getStationNumber()).toHashCode();
	}

	@Override
	public int compareTo(VtkInstitution o) {
		if (equals(o))
			return 0;

		return new CompareToBuilder().append(getStationNumber(), o.getStationNumber()).toComparison() > 0 ? 1 : -1;
	}

	// ---------------------------------------- Accessor Methods

	@Column(name = "NameOfInstitution", length = 30, nullable = false)
	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true, fetch = FetchType.LAZY, mappedBy = "vtkInstitution")
	@BatchSize(size = 500)
	@JsonIgnore
	private List<StationParameters> getStationParametersList() {
		if (stationParametersList == null)
			stationParametersList = new ArrayList<>();
		return stationParametersList;
	}

	@SuppressWarnings("unused")
	private void setStationParametersList(List<StationParameters> stationParameters) {
		this.stationParametersList = stationParameters;
	}

	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true, fetch = FetchType.LAZY, mappedBy = "vtkInstitution")
	@BatchSize(size = 500)
	private List<VoluntaryServiceParameters> getVoluntaryServiceParametersList() {
		if (voluntaryServiceParametersList == null)
			voluntaryServiceParametersList = new ArrayList<>();
		return voluntaryServiceParametersList;
	}

	@SuppressWarnings("unused")
	private void setVoluntaryServiceParametersList(List<VoluntaryServiceParameters> voluntaryServiceParametersList) {
		this.voluntaryServiceParametersList = voluntaryServiceParametersList;
	}

	@Column(length = 7, nullable = false)
	public String getStationNumber() {
		return stationNumber;
	}

	public void setStationNumber(String stationNumber) {
		this.stationNumber = stationNumber;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "ParentInstitutionFK")
	@BatchSize(size = 500)
	@JsonIgnore
	public VtkInstitution getParent() {
		return parent;
	}

	public void setParent(VtkInstitution parent) {
		this.parent = parent;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "AdministrativeParentFK")
	@BatchSize(size = 500)
	@JsonIgnore
	public VtkAdministrativeUnit getAdministrativeParent() {
		return administrativeParent;
	}

	public void setAdministrativeParent(VtkAdministrativeUnit administrativeParent) {
		this.administrativeParent = administrativeParent;
	}

	@Column(name = "StreetAddress1", length = 35)
	public String getAddressLine1() {
		return addressLine1;
	}

	public void setAddressLine1(String addressLine1) {
		this.addressLine1 = addressLine1;
	}

	@Column(name = "StreetAddress2", length = 35)
	public String getAddressLine2() {
		return addressLine2;
	}

	public void setAddressLine2(String addressLine2) {
		this.addressLine2 = addressLine2;
	}

	@Column(length = 30)
	public String getCity() {
		return city;
	}

	public void setCity(String city) {
		this.city = city;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "StateFK", nullable = false)
	@BatchSize(size = 500)
	@JsonIgnore
	public VtkState getState() {
		return state;
	}

	public void setState(VtkState state) {
		this.state = state;
	}

	@Column(name = "ZipCode", length = 10)
	public String getZip() {
		return zip;
	}

	public void setZip(String zip) {
		this.zip = zip;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "TypeFK", nullable = false)
	@BatchSize(size = 500)
	@JsonIgnore
	public VtkInstitutionType getType() {
		return type;
	}

	public void setType(VtkInstitutionType type) {
		this.type = type;
	}

	public Boolean getMealPrintOCX() {
		return mealPrintOCX;
	}

	public void setMealPrintOCX(Boolean mealPrintOCX) {
		this.mealPrintOCX = mealPrintOCX;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "SDS_INSTITUTION_ID")
	@BatchSize(size = 500)
	public VAFacility getVaFacility() {
		return vaFacility;
	}

	public void setVaFacility(VAFacility facility) {
		this.vaFacility = facility;
	}

	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true, fetch = FetchType.LAZY, mappedBy = "vtkInstitution")
	@BatchSize(size = 500)
	@JsonIgnore
	public List<DonGenPostFund> getDonGenPostFunds() {
		if (donGenPostFunds == null)
			donGenPostFunds = new ArrayList<>();
		return donGenPostFunds;
	}

	public void setDonGenPostFunds(List<DonGenPostFund> donGenPostFunds) {
		this.donGenPostFunds = donGenPostFunds;
	}

	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true, fetch = FetchType.LAZY, mappedBy = "vtkInstitution")
	@BatchSize(size = 500)
	@JsonIgnore
	public List<DonationReference> getDonationReferences() {
		if (donationReferences == null)
			donationReferences = new ArrayList<>();
		return donationReferences;
	}

	public void setDonationReferences(List<DonationReference> donationReferences) {
		this.donationReferences = donationReferences;
	}

}
