package gov.va.vss.persistence;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.sql.Blob;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.sql.DataSource;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.SerializationUtils;
import org.springframework.dao.DataAccessException;
import org.springframework.dao.InvalidDataAccessApiUsageException;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.core.support.AbstractInterruptibleBatchPreparedStatementSetter;
import org.springframework.jdbc.core.support.AbstractLobCreatingPreparedStatementCallback;
import org.springframework.jdbc.support.lob.DefaultLobHandler;
import org.springframework.jdbc.support.lob.LobCreator;
import org.springframework.jdbc.support.lob.LobHandler;
import org.springframework.jdbc.support.rowset.ResultSetWrappingSqlRowSet;
import org.springframework.jdbc.support.rowset.SqlRowSet;
import org.springframework.stereotype.Repository;

import com.hazelcast.core.MapStore;

import gov.va.vss.util.CompressionUtil;
import gov.va.vss.util.SerializationUtil;

/**
 * A generic class which stores a {key, value} pair in the DB. Derived classes
 * can be empty if storage into the default table name is acceptable (and their
 * key and value conform to the data types stored in this table). Otherwise,
 * they can define their own table, or optionally their own storage methods
 * entirely.
 * 
 * @author vhaisdbarryc
 * 
 * @param <K>
 *            The key type
 * @param <V>
 *            The value type
 */
@Repository
public class BasicDataSourceClusterStore<K, V extends Serializable> implements MapStore<K, V> {

	private NamedParameterJdbcTemplate npjt;
	private JdbcTemplate jt;

	public void setDataSource(DataSource ds) {
		this.npjt = new NamedParameterJdbcTemplate(ds);
		this.jt = new JdbcTemplate(ds);
	}

	String tableName;

	public void setTableName(String tableName) {
		this.tableName = tableName;
	}

	protected String getSelectQuery() {
		return "select [key], [value] from " + tableName + " where [key] in (:keys)";
	}

	protected String getSelectAllKeysQuery() {
		return "select [key] from " + tableName;
	}

	protected String getMergeQuery() {
		return "merge " + tableName + " as target" //
				+ " using (values(?, ?)) as source([key], [value])" //
				+ " on target.[key] = ?" //
				+ " when matched then update set [value] = source.[value]" //
				+ " when not matched then" //
				+ " insert ([key], [value]) values (source.[key], source.[value]);";
	}

	protected String getDeleteQuery() {
		return "delete from " + tableName + " where [key] in (:keys)";
	}

	private LobHandler lobHandler = new DefaultLobHandler();

	@Override
	public void store(final K key, final V value) {
		jt.execute(getMergeQuery(), new AbstractLobCreatingPreparedStatementCallback(lobHandler) {
			@Override
			protected void setValues(PreparedStatement ps, LobCreator lobCreator)
					throws SQLException, DataAccessException {
				ps.setObject(1, key);
				byte[] data = getBytesFromValue(value);
				ps.setBinaryStream(2, new ByteArrayInputStream(data), (int) data.length);
				ps.setObject(3, key);
			}
		});
	}

	@Override
	public void storeAll(final Map<K, V> map) {
		final List<Map.Entry<K, V>> entries = new ArrayList<>(map.entrySet());

		jt.batchUpdate(getMergeQuery(), new AbstractInterruptibleBatchPreparedStatementSetter() {
			@Override
			protected boolean setValuesIfAvailable(PreparedStatement ps, int i) throws SQLException {
				if (i >= entries.size())
					return false;

				Map.Entry<K, V> entry = entries.get(i);
				byte[] data = getBytesFromValue(entry.getValue());

				ps.setObject(1, entry.getKey());
				ps.setBinaryStream(2, new ByteArrayInputStream(data), (int) data.length);
				ps.setObject(3, entry.getKey());
				return true;
			}
		});
	}

	@Override
	public V load(K key) {
		Map<String, Object> paramMap = new HashMap<>();
		paramMap.put("keys", key);
		ResultSetWrappingSqlRowSet rowSet = (ResultSetWrappingSqlRowSet) npjt.queryForRowSet(getSelectQuery(),
				paramMap);

		if (!rowSet.next())
			return null;

		try {
			Object o = rowSet.getObject(2);
			byte[] bytes = null;
			if (o instanceof Blob) {
				Blob blob = (Blob) rowSet.getObject(2);
				bytes = getBytesFromBlob(blob);
			} else if (o instanceof byte[]) {
				bytes = (byte[]) o;
			}

			V item = getValueFromBytes(bytes);
			return item;
		} catch (IOException e) {
			throw new InvalidDataAccessApiUsageException("Couldn't deserialize the persistent byte[]", e);
		} catch (ClassNotFoundException e) {
			throw new InvalidDataAccessApiUsageException("Couldn't find the serialized class", e);
		} catch (SQLException e) {
			throw new InvalidDataAccessApiUsageException("Database error retrieving value", e);
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public Map<K, V> loadAll(Collection<K> keys) {
		Map<String, Object> paramMap = new HashMap<>();
		paramMap.put("keys", keys);
		SqlRowSet rowSet = npjt.queryForRowSet(getSelectQuery(), paramMap);

		try {
			Map<K, V> results = new HashMap<>();
			while (rowSet.next()) {
				Object o = rowSet.getObject(2);
				byte[] bytes = null;
				if (o instanceof Blob) {
					Blob blob = (Blob) o;
					bytes = getBytesFromBlob(blob);
				} else if (o instanceof byte[]) {
					bytes = (byte[]) o;
				} else
					throw new RuntimeException(
							"Unexpected type returned by getObject() from database: " + o.getClass());

				V item = getValueFromBytes(bytes);
				results.put((K) rowSet.getObject(1), item);
			}
			return results;
		} catch (IOException e) {
			throw new InvalidDataAccessApiUsageException("Couldn't deserialize the persistent byte[]", e);
		} catch (ClassNotFoundException e) {
			throw new InvalidDataAccessApiUsageException("Couldn't find the serialized class", e);
		} catch (SQLException e) {
			throw new InvalidDataAccessApiUsageException("Database error retrieving value", e);
		}

	}

	@SuppressWarnings("unchecked")
	@Override
	public Set<K> loadAllKeys() {
		SqlRowSet rowSet = npjt.queryForRowSet(getSelectAllKeysQuery(), new HashMap<String, Object>());
		Set<K> results = new HashSet<>();
		while (rowSet.next())
			results.add((K) rowSet.getObject(1));
		return results;
	}

	@Override
	public void delete(K key) {
		Map<String, Object> paramMap = new HashMap<>();
		paramMap.put("keys", key);
		npjt.update(getDeleteQuery(), paramMap);
	}

	@Override
	public void deleteAll(Collection<K> keys) {
		Map<String, Object> paramMap = new HashMap<>();
		paramMap.put("keys", keys);
		npjt.update(getDeleteQuery(), paramMap);
	}

	@SuppressWarnings("unchecked")
	protected V getValueFromBytes(byte[] bytes) throws IOException, ClassNotFoundException {
		return (V) SerializationUtil.deserializeObject(CompressionUtil.decompress(bytes, true));
	}

	protected byte[] getBytesFromValue(V value) {
		return CompressionUtil.compress(SerializationUtils.serialize(value));
	}

	private byte[] getBytesFromBlob(Blob obj) throws SQLException, IOException {
		InputStream inStream = null;
		try {
			inStream = obj.getBinaryStream();
			byte[] byteBuffer = IOUtils.toByteArray(inStream);
			return byteBuffer;
		} finally {
			if (inStream != null) {
				try {
					inStream.close();
				} catch (IOException e) {
				}
			}
		}
	}

}
