package gov.va.vss.service.impl;

import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;

import javax.annotation.Resource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import com.hazelcast.core.HazelcastInstance;

import gov.va.shared.util.StationsUtil;
import gov.va.vss.model.VistaStationData;
import gov.va.vss.service.VistaStationDataService;
import gov.va.vss.service.cluster.ClusterInstanceName;
import gov.va.vss.service.cluster.ClusterLockType;

@Service
public class VistaStationDataServiceImpl extends AbstractServiceImpl implements VistaStationDataService {
	private static final Logger log = LoggerFactory.getLogger(VistaStationDataServiceImpl.class);

	@Autowired
	@Qualifier(ClusterInstanceName.ALL)
	protected HazelcastInstance clusterOpsAll;
	@Autowired
	@Qualifier(ClusterInstanceName.UNATTENDED)
	private HazelcastInstance clusterOpsUnattended;
	@Resource(name = "vistaStationDataServiceImpl")
	private VistaStationDataServiceImpl selfProxy;

	/*
	 * We currently establish a cluster lock on the station number for the
	 * getOrCreate* methods below; this may not be performant since 99% of the
	 * time we'll be retrieving an existing value instead of saving a new one.
	 * In the event that HazelCast doesn't optimize its locking behavior (i.e.
	 * like how Java synchronized block is very quick to exit if no lock
	 * exists), we could change it as such: 1 - ensure we have a unique
	 * constraint in the DB on stationNumber, 2 - change method to get, then if
	 * null, create, if we hit a unique constraint error, assume somebody else
	 * just created it, so open a new session & transaction (required by
	 * Hibernate after any DB error) and attempt to get it again. CPB
	 */
	@Value("${vistaStationData.lockAquisitionTimeoutMillis}")
	private long lockAquisitionTimeoutMillis;

	@Override
	public VistaStationData getOrCreateByStation(String stationNumber) {
		stationNumber = StationsUtil.getThreeDigitStationNumber(stationNumber);
		VistaStationData d = vistaStationDataDAO.findByStationNumber(stationNumber);
		if (d != null)
			return d;

		String lockName = ClusterLockType.VISTA_STATION_DATA.getLockName(stationNumber);
		Lock lock = clusterOpsAll.getLock(lockName);
		try {
			if (!lock.tryLock(lockAquisitionTimeoutMillis, TimeUnit.MILLISECONDS))
				throw new RuntimeException("Timeout waiting for cluster to release lock '" + lockName + "'");
		} catch (InterruptedException e) {
			throw new RuntimeException("Interrupted waiting for cluster lock on '" + lockName + "'", e);
		}

		try {
			VistaStationData data = selfProxy.getStationNumberInNewTransaction(stationNumber);
			return data;
		} finally {
			lock.unlock();
		}
	}

	@Transactional(propagation = Propagation.REQUIRES_NEW)
	public VistaStationData getStationNumberInNewTransaction(String stationNumber) {
		VistaStationData d = vistaStationDataDAO.findByStationNumber(stationNumber);
		if (d != null)
			return d;

		d = new VistaStationData();
		d.setStationNumber(stationNumber);
		d = vistaStationDataDAO.saveOrUpdate(d);
		vistaStationDataDAO.clearCache();

		return d;
	}

	@Override
	@Transactional(propagation = Propagation.REQUIRES_NEW)
	public int saveOrUpdateWithFieldsInNewTransaction(String stationNumber,
			ZonedDateTime cprsHeaderSyncLastExecutionDate, ZonedDateTime patientInactivationLastExecutionDate,
			ZonedDateTime vistaCleanupProcessLastExecutionDate, ZonedDateTime staffUpdateLastExecutionDate) {
		int rowsUpdated = vistaStationDataDAO.updateWithFields(stationNumber, null, cprsHeaderSyncLastExecutionDate,
				null, patientInactivationLastExecutionDate, null, staffUpdateLastExecutionDate, null,
				vistaCleanupProcessLastExecutionDate, null);
		if (rowsUpdated == 0) {
			getOrCreateByStation(stationNumber);
			rowsUpdated = vistaStationDataDAO.updateWithFields(stationNumber, null, cprsHeaderSyncLastExecutionDate,
					null, patientInactivationLastExecutionDate, null, staffUpdateLastExecutionDate, null,
					vistaCleanupProcessLastExecutionDate, null);
		}
		return rowsUpdated;
	}

	public int updateCPRSHeaderSyncLastExecutionDatesForEnabledStations(Collection<String> stationNumbers,
			ZonedDateTime executionDate) {
		return vistaStationDataDAO.updateCPRSHeaderSyncLastExecutionDatesForEnabledStations(stationNumbers,
				executionDate);
	}

	@Override
	public int setProcessesEnabled(String stationNumber, Boolean cprsHeaderSyncProcessEnabled,
			Boolean patientInactivationProcessEnabled, Boolean vistaSyncProcessEnabled,
			Boolean vistaCleanupProcessEnabled, Boolean staffUpdateProcessEnabled) {
		int rowsUpdated = vistaStationDataDAO.updateWithFields(stationNumber, cprsHeaderSyncProcessEnabled, null,
				patientInactivationProcessEnabled, null, staffUpdateProcessEnabled, null, vistaCleanupProcessEnabled,
				null, vistaSyncProcessEnabled);
		if (rowsUpdated == 0) {
			getOrCreateByStation(stationNumber);
			rowsUpdated = vistaStationDataDAO.updateWithFields(stationNumber, cprsHeaderSyncProcessEnabled, null,
					patientInactivationProcessEnabled, null, staffUpdateProcessEnabled, null,
					vistaCleanupProcessEnabled, null, vistaSyncProcessEnabled);
		}
		return rowsUpdated;
	}

	@Override
	public List<VistaStationData> getStationsNeedingAutoInactivationProcessing() {
		List<VistaStationData> map = vistaStationDataDAO.findByCriteria(null, null, true);
		return map;
	}

}
