package gov.va.vss.service.impl;

import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;

import gov.va.shared.model.lookup.sds.State;
import gov.va.shared.model.lookup.sds.VAFacility;
import gov.va.vss.model.lookup.VtkAdministrativeUnit;
import gov.va.vss.model.lookup.VtkInstitution;
import gov.va.vss.model.lookup.VtkInstitutionType;
import gov.va.vss.model.lookup.VtkInstitutionType.VtkInstitutionTypeValue;
import gov.va.vss.model.lookup.VtkState;
import gov.va.vss.service.VtkService;

@Service
public class VtkServiceImpl extends AbstractServiceImpl implements VtkService {
	private static final Logger log = LoggerFactory.getLogger(VtkServiceImpl.class);

	/**
	 * Gets the VtkAdministrativeUnit for the specified SDS VISN VAFacility. If
	 * no VtkAdministrativeUnit exists which references the specified VISN
	 * facility, one is created and saved. This method is intentionally not
	 * synchronized across cluster; instead, it expects and depends upon a
	 * unique constraint in DB to prevent duplicates. CPB
	 */
	@Override
	public synchronized VtkAdministrativeUnit getOrCreateAdministrativeUnitForSDSVisnFacility(long visnVaFacilityId) {
		VtkAdministrativeUnit u = vtkAdministrativeUnitDAO.findBySDSInstitution(visnVaFacilityId);
		if (u != null)
			return u;

		VAFacility facility = vaFacilityDAO.findRequiredByPrimaryKey(visnVaFacilityId);
		u = new VtkAdministrativeUnit(facility);
		u = vtkAdministrativeUnitDAO.saveOrUpdate(u);
		return u;
	}

	/**
	 * Gets the VSS institution object for the specified SDS VAFacility object.
	 * If no VSS institution object exists with a matching stationnumber, one is
	 * created and saved. This method is intentionally not synchronized across
	 * cluster; instead, it expects and depends upon a unique constraint in DB
	 * to prevent duplicates. CPB
	 */
	@Override
	public synchronized VtkInstitution getOrCreateInstitutionForSDSFacility(long vaFacilityId) {
		VtkInstitution i = vtkInstitutionDAO.findBySDSFacility(vaFacilityId);
		if (i != null) {
			i.populateMissingRequiredFields();
			i = vtkInstitutionDAO.saveOrUpdate(i);
			return i;
		}

		/* Temporary workaround */
		VAFacility facility = vaFacilityDAO.findRequiredByPrimaryKey(vaFacilityId);

		if (System.currentTimeMillis() > 0) {
			VtkState state = getOrCreateForSDSState(facility.getState().getId());
			VtkInstitutionType type = vtkInstitutionTypeDAO.findByLookup(VtkInstitutionTypeValue.TIMEKEEPING);
			VtkInstitution newInstitution = VtkInstitution.createNew(facility, state, type, null);
			newInstitution = vtkInstitutionDAO.saveOrUpdate(newInstitution);
			return newInstitution;
		}

		VAFacility visnFacility = null;
		VtkInstitution root = null;

		/* top-down list of nodes needing addition */
		List<VAFacility> facilitiesToAdd = new ArrayList<>();
		facilitiesToAdd.add(facility);

		if (facility.getVisn() != null) {
			visnFacility = facility.getVisn();
		}

		while (true) {
			/* move to parent node */
			facility = facility.getParent();

			if (facility.getVisn() != null)
				visnFacility = facility.getVisn();

			/*
			 * If we don't have a parent defined, or we have ourselves defined
			 * as our own parent, exit
			 */
			if (facility == null || facilitiesToAdd.contains(facility))
				break;

			/*
			 * If the parent is the VISN, break - this is the best parent
			 */
			if (facility.isVISN()) {
				visnFacility = facility;
				break;
			}

			/*
			 * Our parent exists and is a non-VISN. Check to see if we already
			 * have a VtkInstitution defined for this facility. If so, use it as
			 * the parent and bail; otherwise, queue it for addition as a new
			 * VtkInstitution and continue walking up the tree - CPB
			 */
			VtkInstitution iParent = vtkInstitutionDAO.findBySDSFacility(facility.getId());
			if (iParent != null) {
				root = iParent;

				break;
			} else {
				facilitiesToAdd.add(0, facility);
			}
		}

		/*
		 * Once we leave the loop we have a list of nodes to add in top-down
		 * order, and hopefully either a parent or VISN defined.
		 */

		// FIXVSS todo
		VtkInstitution parent = null; // visnFacility
		if (visnFacility != null) {
		} else if (root != null) {
			parent = root;
			for (VAFacility facilityToAdd : facilitiesToAdd) {
				VtkState vtkState = getOrCreateForSDSState(facilityToAdd.getState().getId());
				VtkInstitutionType vtkInstitutionType = vtkInstitutionTypeDAO
						.findByLookup(VtkInstitutionTypeValue.TIMEKEEPING);
				VtkInstitution newInstitution = VtkInstitution.createNew(facilityToAdd, vtkState, vtkInstitutionType,
						parent);
				newInstitution = vtkInstitutionDAO.saveOrUpdate(newInstitution);
				parent = newInstitution;
			}
		}

		return parent;
	}

	/**
	 * Gets the VtkState for the specified SDS State. If no VtkState exists with
	 * a matching fips code, one is created and saved. This method is
	 * intentionally not synchronized across cluster; instead, it expects and
	 * depends upon a unique constraint in DB to prevent duplicates. CPB
	 */
	@Override
	public synchronized VtkState getOrCreateForSDSState(long sdsStateId) {
		State state = stateDAO.findRequiredByPrimaryKey(sdsStateId);
		VtkState vtkState = vtkStateDAO.findByFIPSCode(state.getFipsCode());

		if (vtkState != null)
			return vtkState;

		VtkState s = new VtkState(state);
		s = vtkStateDAO.saveOrUpdate(s);
		return s;
	}

}
