package gov.va.vss.service.messaging.listeners;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jms.JMSException;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.batch.core.JobParameter;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jms.annotation.JmsListener;
import org.springframework.stereotype.Component;

import gov.va.vss.model.BinaryObject;
import gov.va.vss.service.BinaryObjectService;
import gov.va.vss.service.batch.BatchProcessService;
import gov.va.vss.util.SerializationUtil;

@Component
public class BatchProcessListener {
	private static final Logger log = LoggerFactory.getLogger(BatchProcessListener.class);

	public static final String JOB_NAME = "jobName";
	public static final String USER_ID = "userId";
	public static final String JOB_PARAM_MAP = "jobParamMap";
	public static final String EXAMPLE_SERIALIZED_OBJECT = "serializedObject";

	public static final String PERSISTENT_SERIALIZABLES_ID_LIST = "persistentSerializablesList";

	@Autowired
	private BatchProcessService batchProcessServiceImpl;
	@Autowired
	private BinaryObjectService binaryObjectService;

	@JmsListener(concurrency = "${jms.batchProcess.concurrency}", destination = "${jms.batchProcess.queueName}")
	public void handleMessage(Map<String, Serializable> messageMap) throws Exception {
		String jobName = (String) messageMap.get(JOB_NAME);
		log.debug("Executing job {}", jobName);

		/*
		 * Need to merge the separate JMS fields into a single map for the batch
		 * job framework - CPB
		 */
		Map<String, JobParameter> batchJobParameterMap = new HashMap<>();

		String userId = (String) messageMap.get(USER_ID);
		if (userId != null)
			batchJobParameterMap.put(USER_ID, new JobParameter(userId));

		byte[] serializedMap = (byte[]) messageMap.get(JOB_PARAM_MAP);
		@SuppressWarnings("unchecked")
		Map<String, Serializable> jobParamMap = (Map<String, Serializable>) SerializationUtil
				.deserializeObject(serializedMap);
		copyJobParamMap(jobParamMap, batchJobParameterMap);

		/*
		 * Binary objects we can't push into the batch job framework via
		 * JobParameters, so we save them to the DB and push in the
		 * comma-separated list of IDs instead - CPB
		 */
		List<Long> idList = new ArrayList<>();

		byte[] serializedTokenMap = (byte[]) messageMap.get(EXAMPLE_SERIALIZED_OBJECT);
		if (serializedTokenMap != null) {
			@SuppressWarnings("unchecked")
			Map<String, Serializable> serializableObject = (Map<String, Serializable>) SerializationUtil
					.deserializeObject(serializedTokenMap);
			long id = persistSerializable((Serializable) serializableObject);
			idList.add(id);
		}

		batchJobParameterMap.put(PERSISTENT_SERIALIZABLES_ID_LIST, new JobParameter(StringUtils.join(idList, ",")));

		batchJobParameterMap.put("Date", new JobParameter(new Date()));

		batchProcessServiceImpl.executeJob(jobName, batchJobParameterMap);
	}

	private long persistSerializable(Serializable s) throws ClassNotFoundException, IOException, JMSException {
		BinaryObject bo = new BinaryObject(s);
		bo = binaryObjectService.saveOrUpdate(bo);
		return bo.getId();
	}

	public void copyJobParamMap(Map<String, Serializable> jobParamMap, Map<String, JobParameter> jobParameterMap) {
		for (Map.Entry<String, ?> entry : jobParamMap.entrySet()) {
			Object val = entry.getValue();
			if (val instanceof String) {
				jobParameterMap.put(entry.getKey(), new JobParameter((String) val));
			} else if (val instanceof Long) {
				jobParameterMap.put(entry.getKey(), new JobParameter((Long) val));
			} else if (val instanceof Date) {
				jobParameterMap.put(entry.getKey(), new JobParameter((Date) val));
			} else if (val instanceof Double) {
				jobParameterMap.put(entry.getKey(), new JobParameter((Double) val));

			} else {
				throw new IllegalArgumentException(
						"The job parameter " + entry.getKey() + " has an unsupported value type of " + val.getClass()
								+ " (only String, Long, Date and Double are supported)");
			}
		}
	}

}
