package gov.va.vss.web;

import java.time.ZonedDateTime;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.web.authentication.logout.SecurityContextLogoutHandler;
import org.springframework.stereotype.Controller;
import org.springframework.ui.ModelMap;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;

import com.hazelcast.core.Hazelcast;
import com.hazelcast.core.HazelcastInstance;

import gov.va.shared.model.AppUser;
import gov.va.shared.model.lookup.sds.VAFacility;
import gov.va.shared.persistence.lookup.sds.VAFacilityDAO;
import gov.va.shared.util.SecurityUtil;
import gov.va.shared.util.SessionUtil;
import gov.va.shared.web.DatabaseDrivenPreAuthenticationFilter;
import gov.va.vss.config.SecurityConfig;
import gov.va.vss.model.lookup.TemplateType;
import gov.va.vss.service.VelocityService;
import gov.va.vss.service.cluster.ClusterInstanceName;
import gov.va.vss.web.breadcrumbs.Breadcrumb;

@Controller
public class CommonController extends AbstractAppController {
	private static final Logger log = LoggerFactory.getLogger(CommonController.class);

	private static final String VIEW_STATION_CHANGE = "stationChange";

	public static final String BREADCRUMB_HOME = "Home";
	public static final String URL_HOME = "/home.htm";
	public static final String VIEW_HOME = "home";

	@Autowired
	private VelocityService velocityService;
	@Value("${session.expirationSeconds}")
	private int sessionExpirationSeconds;

	@RequestMapping(value = "/processAuthorizationException.htm", method = RequestMethod.GET)
	public String authError() {
		return "authorizationException";
	}

	@RequestMapping(value = SecurityConfig.URI_LOGIN, method = RequestMethod.GET)
	public String loginPage() {
		return "login";
	}

	@RequestMapping(value = SecurityConfig.URI_LOGOUT, method = RequestMethod.GET)
	public String logoutPage(HttpServletRequest request, HttpServletResponse response) {
		Authentication auth = SecurityContextHolder.getContext().getAuthentication();
		if (auth != null) {
			/*
			 * If they manually clicked Logout, then let's disable our PreAuth
			 * filter for the remainder of this JVM, for testing - CPB
			 */
			DatabaseDrivenPreAuthenticationFilter.preauthDisabled = true;
			new SecurityContextLogoutHandler().logout(request, response, auth);
		}
		return "redirect:/login.htm?logout";
	}

	@RequestMapping("/index.htm")
	public String index(ModelMap modelMap, HttpSession session) {
		session.setMaxInactiveInterval(sessionExpirationSeconds);

		AppUser user = getCurrentUser();
		appUserService.logApplicationAccess(user.getUsername(), ZonedDateTime.now());
		resetSiteContextIfNecessary();
		VAFacility dutyStationToSet = getSiteContext();

		if (dutyStationToSet == null) {
			if (user.getFacilities().size() > 1) {
				VAFacility primaryFacility = appUserFacilityDAO.findPrimaryFacilityForUser(user.getId());
				if (primaryFacility == null) {
					return "redirect:/selectStation.htm";
				} else {
					setNewDutyStation(primaryFacility);
				}
			} else if (user.getFacilities().size() == 1) {
				VAFacility facility = user.getFacilities().iterator().next().getFacility();
				setNewDutyStation(facility);
			}
		}

		return "redirect:" + URL_HOME;
	}

	@RequestMapping(URL_HOME)
	@Breadcrumb(BREADCRUMB_HOME)
	public String home(ModelMap model) {
		if (getSiteContext() == null)
			return "redirect:/selectStation.htm";

		HazelcastInstance instance = Hazelcast.getHazelcastInstanceByName(ClusterInstanceName.UNATTENDED);
		model.put("unattendedServerRunning", !instance.getCluster().getMembers().isEmpty());

		model.put("homepageContent", velocityService.mergeTemplateIntoString(TemplateType.HOMEPAGE_CONTENT));
		model.put("homepageAnnouncement", velocityService.mergeTemplateIntoString(TemplateType.HOMEPAGE_ANNOUNCEMENT));

		// Already a station in session. Do nothing, redirect to sitemap.
		return VIEW_HOME;
	}

	/**
	 * Renew the session, so it will not timeout, ajax call
	 */
	@RequestMapping("/keepAlive")
	public @ResponseBody boolean keepAlive() {
		log.debug("Received request to keep-alive the session...");
		return true;
	}

	private void resetSiteContextIfNecessary() {
		VAFacility siteContext = getSiteContext();
		AppUser user = getCurrentUser();
		if (user == null || siteContext == null)
			return;

		if (!isFacilityAvailableToUser(user, siteContext)) {
			SessionUtil.setSiteContext(null);
		}
	}

	private boolean isFacilityAvailableToUser(AppUser user, VAFacility newFacility) {
		if (user.isNationalAdmin())
			return true;
		return user.getAppUserFacility(newFacility.getId()) != null;
	}

	public static void populateFacilityList(AppUser user, ModelMap model, VAFacilityDAO vaFacilityDAO) {
		if (user.isNationalAdmin()) {
			model.put("facilityList", vaFacilityDAO.findAll());
		} else {
			model.put("facilityList", user.getFacilities());
		}
	}

	@RequestMapping(value = "/selectStation.htm", method = RequestMethod.GET)
	public String selectStation(ModelMap model) {
		populateStationChangeModel(model);
		model.put("cancelAllowed", false);
		model.put("currentPage", "selectStation.htm");
		return VIEW_STATION_CHANGE;
	}

	@RequestMapping("/changeStation.htm")
	public String changeStation(ModelMap model) {
		populateStationChangeModel(model);
		model.put("cancelAllowed", true);
		model.put("currentPage", "changeStation.htm");
		return VIEW_STATION_CHANGE;
	}

	private void populateStationChangeModel(ModelMap model) {
		AppUser user = SecurityUtil.getCurrentUser();

		VAFacility currentFacility = user.getLastVisitedFacility();
		if (currentFacility != null) {
			model.put("currentStationName", currentFacility.getName());
			model.put("currentStationId", currentFacility.getId());
		}

		populateFacilityList(user, model, vaFacilityDAO);
		if (currentFacility != null) {
			VAFacility visn = currentFacility.getVisn();
			if (visn != null && visn.getId() != null) {
				VAFacility attachedVisn = vaFacilityDAO.findRequiredByPrimaryKey(visn.getId());
				model.put("currentStationVisnName", attachedVisn.getDisplayName());
			}

			VAFacility parent = currentFacility.getParent();
			if (parent != null && parent.getId() != null) {
				VAFacility attachedParent = vaFacilityDAO.findRequiredByPrimaryKey(parent.getId());
				model.put("currentStationParentName", attachedParent.getDisplayName());
			}
		}
	}

	private void setNewDutyStation(VAFacility newDutyStation) {
		AppUser user = SecurityUtil.getCurrentUser();
		user.setLastVisitedFacility(newDutyStation);
		appUserService.updateFieldsWithoutVersionCheck(user.getId(), false, newDutyStation.getId(), false, null, null);
		SessionUtil.setSiteContext(newDutyStation);
	}

	@RequestMapping(value = "/selectStation.htm", method = RequestMethod.POST)
	public String processSelectStation(@RequestParam VAFacility stationCode, @RequestParam String currentPage) {
		setNewDutyStation(stationCode);
		return "redirect:" + URL_HOME;
	}
}
