package gov.va.vss.web.interceptor;

import java.io.IOException;
import java.io.InputStream;
import java.time.ZoneId;
import java.time.format.TextStyle;
import java.util.Locale;
import java.util.Map;
import java.util.jar.Attributes;
import java.util.jar.Manifest;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.web.servlet.ModelAndView;
import org.springframework.web.servlet.view.RedirectView;

import gov.va.shared.model.AppUser;
import gov.va.shared.model.Permission;
import gov.va.shared.model.Role.RoleType;
import gov.va.shared.model.lookup.sds.VAFacility;
import gov.va.shared.model.lookup.sds.VAFacility.VAFacilityValue;
import gov.va.shared.persistence.AppUserDAO;
import gov.va.shared.persistence.lookup.sds.StateDAO;
import gov.va.shared.util.SecurityUtil;
import gov.va.shared.util.SessionUtil;
import gov.va.shared.util.WebUtil;
import gov.va.vss.config.WebConfig;
import gov.va.vss.model.VistaStationData;
import gov.va.vss.model.lookup.TemplateType;
import gov.va.vss.persistence.dao.VistaStationDataDAO;
import gov.va.vss.service.VelocityService;
import gov.va.vss.util.DateUtil;
import gov.va.vss.web.UserNotifier;

@Component
public class CommonReferenceDataInterceptor extends AbstractInterceptor {
	private static final Logger log = LoggerFactory.getLogger(CommonReferenceDataInterceptor.class);

	public static final String ONE_TIME_USER_NOTIFICATION = "oneTimeUserNotification";

	private static final String MANIFEST_PATH = "/META-INF/MANIFEST.MF";
	private static final String MANIFEST_ATTRIBUTE_VERSION = "Implementation-Version";

	@Autowired
	protected AppUserDAO appUserDAO;
	@Autowired
	protected StateDAO stateDAO;
	@Autowired
	protected VistaStationDataDAO vistaStationDataDAO;
	@Autowired
	protected VelocityService velocityService;

	@Value("${session.heartBeatTimeoutMillis}")
	private int sessionHeartBeatTimeoutMillis;
	@Value("${session.idleAfterSeconds}")
	private int sessionIdleAfterSeconds;
	@Value("${session.pollingIntervalSeconds}")
	private int sessionPollingIntervalSeconds;
	@Value("${session.failedRequestsCount}")
	private int sessionFailedRequestsCount;

	@Value("${solr.baseURL}")
	private String solrBaseURL;

	@Value("${web.version}")
	private String appVersion;

	@Override
	public void postHandle(HttpServletRequest request, HttpServletResponse response, Object handler,
			ModelAndView modelAndView) throws Exception {
		if (isAjax(request)) {
			log.debug("Request was an AJAX request, no need to populate reference data");
			return;
		}

		if (modelAndView == null) {
			log.debug("No modelAndView returned, the controller must have dealt with the response manually.");
			return;
		}

		Map<String, Object> model = modelAndView.getModel();
		if (model == null) {
			log.error(
					"No model in the ModelAndView, that is unexpected. Returning without adding common reference data.");
			return;
		}

		if (modelAndView.getView() instanceof RedirectView
				|| (modelAndView.isReference() && modelAndView.getViewName().startsWith("redirect:"))) {
			log.debug("View is null or a redirect, no need to populate reference data");
			return;
		}

		HttpSession session = request.getSession();
		ServletContext servletContext = session.getServletContext();

		if (session.getAttribute("sessionHeartBeatTimeoutMillis") == null) {
			session.setAttribute("sessionHeartBeatTimeoutMillis", sessionHeartBeatTimeoutMillis);
			session.setAttribute("sessionIdleAfterSeconds", sessionIdleAfterSeconds);
			session.setAttribute("sessionPollingIntervalSeconds", sessionPollingIntervalSeconds);
			session.setAttribute("sessionFailedRequestsCount", sessionFailedRequestsCount);
		}

		String val = UserNotifier.getAndRemoveAttributeFromSession(session);
		if (val != null)
			model.put(ONE_TIME_USER_NOTIFICATION, val);

		AppUser appUser = SecurityUtil.getCurrentUser();

		if (appUser != null) {
			model.put("currentUser", appUser.getDisplayName());
			model.put("currentUserID", appUser.getId());

			model.put("multipleStationsAssigned", appUser.getFacilities().size() > 1);

			/*
			 * Use of this flag should be discouraged since most things should
			 * be based on permissions - CPB
			 */
			boolean hasNationalAdmin = appUser.isNationalAdmin();
			model.put("hasNationalAdminRole", hasNationalAdmin);

			ZoneId timeZone = appUser.getTimeZone();

			if (timeZone != null) {
				model.put("userTimeZone", timeZone);
				model.put("userTimeZoneName", timeZone.getDisplayName(TextStyle.FULL, Locale.US));
			}
		}

		/*
		 * Add all roles and permissions to model for standard use in
		 * <security:authorize> JSP tags - CPB
		 */
		WebUtil.addClassConstantsToModel(Permission.class, model, true);
		WebUtil.addEnumToModel(RoleType.class, model);
			
		WebUtil.addEnumToModel(VAFacilityValue.class, model);
		
		/*
		 * Add all date util constants as they are defined (without adding
		 * "DATE_UTIL_" before their names) - CPB
		 */
		WebUtil.addClassConstantsToModel(DateUtil.class, model, false);

		/* Add constants from WebUtils for global error display */
		WebUtil.addClassConstantsToModel(WebUtil.class, model);

		model.put("AJAX_CONTEXT_PATH_PREFIX", WebConfig.AJAX_CONTEXT_PATH_PREFIX);
		model.put("solrBaseURL", solrBaseURL);

		VAFacility siteContext = SessionUtil.getSiteContext(session);
		model.put("siteContext", siteContext);
		if (siteContext != null) {
			VistaStationData vsd = vistaStationDataDAO.findByStationNumber(siteContext.getSta3n());
			model.put("siteContextStationData", vsd);
		}

		model.put("appVersionNumber", appVersion);

		/* Velocity caches these vals per the engine setup - CPB */
		model.put("referenceDataLinks", velocityService.mergeTemplateIntoString(TemplateType.REFERENCE_DATA_LINKS));
		model.put("footerContent", velocityService.mergeTemplateIntoString(TemplateType.FOOTER_CONTENT));
		model.put("systemNotification",
				StringUtils.trim(velocityService.mergeTemplateIntoString(TemplateType.SYSTEM_NOTIFICATION)));
	}

}
