package gov.va.vss.web.volunteer;

import java.util.Arrays;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.servlet.http.HttpServletRequest;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Controller;
import org.springframework.ui.ModelMap;
import org.springframework.validation.BindingResult;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.SessionAttributes;
import org.springframework.web.bind.support.SessionStatus;

import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.lookup.sds.Gender;
import gov.va.shared.model.lookup.sds.Gender.GenderType;
import gov.va.shared.model.lookup.sds.State;
import gov.va.shared.model.lookup.sds.VAFacility;
import gov.va.shared.util.StringUtil;
import gov.va.vss.model.NextOfKin;
import gov.va.vss.model.Volunteer;
import gov.va.vss.model.Volunteer.VolunteerView;
import gov.va.vss.model.VolunteerContactInfo;
import gov.va.vss.model.VolunteerStationProfile;
import gov.va.vss.model.lookup.VtkInstitution;
import gov.va.vss.model.lookup.VtkState;
import gov.va.vss.service.validation.ServiceValidationException;
import gov.va.vss.web.AbstractAppController;
import gov.va.vss.web.ValidationException;
import gov.va.vss.web.breadcrumbs.Breadcrumb;

@Controller
@SessionAttributes(value = { AbstractAppController.DEFAULT_COMMAND_NAME })
public class VolunteerController extends AbstractAppController {

	@Autowired
	private VolunteerValidator volunteerValidator;

	@RequestMapping("/findVolunteers")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView({ VolunteerView.Search.class })
	public @ResponseBody SortedSet<Volunteer> findVolunteers(@RequestParam(required = false) String name,
			@RequestParam(required = false) String code, @RequestParam(required = false) String address,
			@RequestParam(required = false) String location, @RequestParam(required = false) String phone,
			@RequestParam(required = false) String email, @RequestParam(required = false) String scope) {
		if (StringUtil.allBlank(name, code, address, location, phone, email))
			throw new IllegalArgumentException("Please specify at least one piece of search criteria");

		String sta3nRestriction = null;
		if (StringUtils.isBlank(scope) || !"national".equals(scope))
			sta3nRestriction = getSiteContext().getSta3n();

		String firstName = null;
		String lastName = null;
		String middleName = null;
		if (StringUtils.isNotBlank(name)) {
			String[] nameComponents = StringUtil.parseNameComponents(name);
			lastName = nameComponents[0];
			firstName = nameComponents[1];
			middleName = nameComponents[2];
		}

		String city = null;
		String state = null;
		String zip = null;
		if (StringUtils.isNotBlank(location)) {
			String[] locationComponents = StringUtil.parseLocationComponents(location);
			city = locationComponents[0];
			state = locationComponents[1];
			zip = locationComponents[2];
		}

		SortedSet<Volunteer> volunteers = new TreeSet<>(
				volunteerDAO.findByCriteria(firstName, middleName, lastName, code, address, city, state, zip, phone,
						email, null, sta3nRestriction != null ? Arrays.asList(sta3nRestriction) : null));

		/*
		 * Populate a map if tertiary information (outside of the Volunteer
		 * object itself) needs to be included in the response; otherwise, just
		 * return the Collection<Volunteer> above directly and Jackson will
		 * serialize it into JSON (per the annotations in the Volunteer class).
		 * CPB
		 */
		// List<Map<String, Object>> results = new ArrayList<>();
		//
		// for (Volunteer volunteer : volunteers) {
		// Map<String, Object> result = new HashMap<>();
		// result.put("volunteerId", volunteer.getId());
		// result.put("displayName", volunteer.getDisplayName());
		// result.put("contactInfo",
		// volunteer.getStationProfileForSta3n(sta3n).getContactInfo());
		// results.add(result);
		// }
		//
		// return results;

		return volunteers;
	}

	@RequestMapping("/volunteerCreate.htm")
	@Breadcrumb("Create Volunteer")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_EDIT + "')")
	public String createVolunteer(ModelMap model, HttpServletRequest request) {
		Volunteer volunteer = new Volunteer();
		VtkInstitution vtkInstitution = vtkService.getOrCreateInstitutionForSDSFacility(getSiteContext().getId());
		VolunteerStationProfile stationProfile = new VolunteerStationProfile(volunteer, vtkInstitution);
		VolunteerCommand command = new VolunteerCommand(volunteer, stationProfile, null, null);
		model.addAttribute(DEFAULT_COMMAND_NAME, command);
		createReferenceData(model);
		return "createVolunteer";
	}

	@RequestMapping("/volunteerEdit.htm")
	@Breadcrumb("Edit Volunteer")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_EDIT + "')")
	public String editVolunteer(@RequestParam long id, ModelMap model, HttpServletRequest request) {
		VAFacility currentSite = getSiteContext();
		Volunteer volunteer = volunteerDAO.findRequiredByPrimaryKey(id);

		VolunteerStationProfile stationProfile = volunteer.getStationProfileForSta3n(currentSite.getSta3n());
		if (stationProfile == null) {
			VtkInstitution vtkInstitution = vtkService.getOrCreateInstitutionForSDSFacility(getSiteContext().getId());
			stationProfile = new VolunteerStationProfile(volunteer, vtkInstitution);
		}

		String gender = volunteer.getSexMF();
		VtkState vtkState = stationProfile.getContactInfo().getState();
		VolunteerCommand command = new VolunteerCommand(volunteer, stationProfile,
				gender == null ? null : genderDAO.findByCode(gender),
				vtkState == null ? null : stateDAO.findStateByPostalCode(vtkState.getAbbreviation()));
		model.addAttribute(DEFAULT_COMMAND_NAME, command);
		createReferenceData(model);

		return "editVolunteer";
	}

	private void createReferenceData(ModelMap model) {
		model.put("allGenders", // genderDAO.findAllSorted()
				Arrays.asList( //
						genderDAO.findByLookup(GenderType.MALE), //
						genderDAO.findByLookup(GenderType.FEMALE)));
		model.put("allStates", stateDAO.findListOfStatesInUSA());
		model.put("allLanguages", vtkLanguageDAO.findAllSorted());
	}

	@RequestMapping("/volunteerSubmit.htm")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_EDIT + "')")
	public String submitVolunteer(@ModelAttribute(DEFAULT_COMMAND_NAME) VolunteerCommand command, BindingResult result,
			SessionStatus status, ModelMap model, HttpServletRequest request) throws ValidationException {
		Volunteer volunteer = command.getVolunteer();
		VolunteerStationProfile stationProfile = command.getStationProfile();
		VolunteerContactInfo contactInfo = stationProfile.getContactInfo();

		boolean isEdit = volunteer.isPersistent();

		/*
		 * Perform any remaining binding steps from Command obj -> Volunteer
		 * here
		 */
		Gender gender = command.getGender();
		volunteer.setSexMF(gender != null ? gender.getCode() : null);

		State state = command.getState();
		contactInfo.setState(state != null ? vtkService.getOrCreateForSDSState(state.getId()) : null);

		NextOfKin nextOfKin = command.getNextOfKin();
		if (StringUtils.isNotBlank(nextOfKin.getName())) {
			stationProfile.setNextOfKin(nextOfKin);
		} else {
			stationProfile.setNextOfKin(null);
		}

		if (!volunteer.getStationProfiles().contains(stationProfile))
			volunteer.getStationProfiles().add(stationProfile);

		/* Validation step (JSR303, other custom logic in the validator) */
		volunteerValidator.validate(command, result, false, "volunteer");
		boolean hasErrors = result.hasErrors();

		if (!hasErrors) {
			try {
				volunteer = volunteerService.saveOrUpdate(volunteer);
				userNotifier.notifyUserOnceWithMessage(request,
						getMessage(isEdit ? "volunteer.update.success" : "volunteer.create.success"));
			} catch (ServiceValidationException e) {
				webValidationService.handle(e, result);
				hasErrors = true;
			}
		}

		if (hasErrors) {
			createReferenceData(model);
			return isEdit ? "editVolunteer" : "createVolunteer";
		} else {
			status.setComplete();
			return "redirect:/volunteerEdit.htm?id=" + volunteer.getId();
		}
	}

}
