package gov.va.vss.model;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Transient;
import javax.validation.constraints.NotNull;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.hibernate.annotations.BatchSize;
import org.hibernate.validator.constraints.NotEmpty;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnore;

import gov.va.shared.model.core.AbstractAuditedVersionedPersistent;
import gov.va.vss.model.lookup.VtkState;

@Entity
@Table(name = "VtkVolunteerContactInformation", schema = "dbo")
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public class VolunteerContactInfo extends AbstractAuditedVersionedPersistent<VolunteerContactInfo> {
	private static final long serialVersionUID = -8641562840185748448L;

	// -------------------------------------- Fields

	@NotNull
	private Volunteer volunteer;
	/*
	 * Only contains one item, lazy-loaded; mapping like this instead
	 * of @OneToOne so it can be optional and still lazy-load - CPB
	 */
	@NotEmpty
	private List<VolunteerStationProfile> stationProfileList;
	@NotNull
	private String addressLine1;
	private String addressLine2;
	@NotNull
	private String city;
	@NotNull
	private VtkState vtkState;
	private String zip;
	private String phone;
	private String phoneAlternate;
	private String phoneAlternate2;
	private String email;

	// -------------------------------------- Constructors

	public VolunteerContactInfo() {
	}

	/**
	 * Constructor for new VolunteerContactInfo - populates parent reference and
	 * creates new empty @NotNull children
	 * 
	 * @param volunteer
	 */
	public VolunteerContactInfo(Volunteer volunteer) {
		this.volunteer = volunteer;
	}

	// -------------------------------------- Business Methods

	@Transient
	public String getAddressMultilineDisplay() {
		List<String> items = getAddressComponents();
		if (items.isEmpty())
			return "";
		return StringUtils.join(items, "\n");
	}

	@Transient
	@JsonIgnore
	private List<String> getAddressComponents() {
		List<String> items = new ArrayList<>(6);
		if (StringUtils.isNotBlank(getAddressLine1()))
			items.add(getAddressLine1());
		if (StringUtils.isNotBlank(getAddressLine2()))
			items.add(getAddressLine2());

		boolean hasCity = StringUtils.isNotBlank(getCity());
		boolean hasState = getState() != null;
		if (hasCity || hasState) {
			if (hasCity && hasState) {
				items.add(city + ", " + getState().getAbbreviation());
			} else if (hasCity) {
				items.add(city);
			} else {
				items.add(getState().getAbbreviation());
			}
		}
		if (StringUtils.isNotBlank(getZip()))
			items.add(getZip());
		return items;
	}

	@Transient
	@JsonIgnore
	public VolunteerStationProfile getStationProfile() {
		List<VolunteerStationProfile> l = getStationProfileList();
		return l.isEmpty() ? null : l.get(0);
	}

	public void setStationProfile(VolunteerStationProfile stationProfile) {
		getStationProfileList().clear();
		if (stationProfile != null) {
			getStationProfileList().add(stationProfile);
			stationProfile.setContactInfo(this);
		}
	}

	// -------------------------------------- Common Methods

	@Override
	protected boolean requiredEquals(VolunteerContactInfo oo) {
		/*
		 * Because the DB model doesn't have a backreference to the
		 * StationProfile I don't see a better way to do this yet - CPB
		 */
		return new EqualsBuilder().append(nullSafeGetId(volunteer), nullSafeGetId(oo.getVolunteer()))
				.append(addressLine1, oo.getAddressLine1()).append(addressLine2, oo.getAddressLine2())
				.append(city, oo.getCity()).append(vtkState, oo.getState()).append(zip, oo.getZip()).isEquals();
	}

	@Override
	protected int requiredHashCode() {
		return new HashCodeBuilder().append(nullSafeGetId(volunteer)).append(addressLine1).append(addressLine2)
				.append(city).append(vtkState).append(zip).toHashCode();
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder(getAddressMultilineDisplay());
		if (StringUtils.isNotBlank(phone))
			sb.append("\nPhone: ").append(phone);
		if (StringUtils.isNotBlank(email))
			sb.append("\nEmail: ").append(email);
		return sb.toString();
	}

	// -------------------------------------- Accessor Methods

	@Column(name = "StreetAddress1", length = 35, nullable = false)
	public String getAddressLine1() {
		return addressLine1;
	}

	public void setAddressLine1(String addressLine1) {
		this.addressLine1 = addressLine1;
	}

	@Column(name = "StreetAddress2", length = 35)
	public String getAddressLine2() {
		return addressLine2;
	}

	public void setAddressLine2(String addressLine2) {
		this.addressLine2 = addressLine2;
	}

	@Column(length = 30, nullable = false)
	public String getCity() {
		return city;
	}

	public void setCity(String city) {
		this.city = city;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "StateFK", nullable = false)
	public VtkState getState() {
		return vtkState;
	}

	public void setState(VtkState vtkState) {
		this.vtkState = vtkState;
	}

	@Column(name = "ZipCode", length = 10)
	public String getZip() {
		return zip;
	}

	public void setZip(String zip) {
		this.zip = zip;
	}

	@Column(name = "Telephone", length = 30)
	public String getPhone() {
		return phone;
	}

	public void setPhone(String phone) {
		this.phone = phone;
	}

	@Column(name = "AlternateTelephone", length = 30)
	public String getPhoneAlternate() {
		return phoneAlternate;
	}

	public void setPhoneAlternate(String phoneAlternate) {
		this.phoneAlternate = phoneAlternate;
	}

	@Column(name = "AlternateTelephone2", length = 30)
	public String getPhoneAlternate2() {
		return phoneAlternate2;
	}

	public void setPhoneAlternate2(String phoneAlternate2) {
		this.phoneAlternate2 = phoneAlternate2;
	}

	@Column(name = "EMailAddress", length = 50)
	public String getEmail() {
		return email;
	}

	public void setEmail(String email) {
		this.email = email;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "VolunteerFK", nullable = false)
	@JsonIgnore
	public Volunteer getVolunteer() {
		return volunteer;
	}

	public void setVolunteer(Volunteer volunteer) {
		this.volunteer = volunteer;
	}

	@OneToMany(mappedBy = "contactInfo", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	protected List<VolunteerStationProfile> getStationProfileList() {
		return stationProfileList;
	}

	protected void setStationProfileList(List<VolunteerStationProfile> stationProfileList) {
		this.stationProfileList = stationProfileList;
	}

}
