package gov.va.vss.persistence.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.persistence.Query;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Repository;

import gov.va.vss.model.Organization;
import gov.va.vss.model.OrganizationScopeType;
import gov.va.vss.persistence.dao.OrganizationDAO;
import gov.va.vss.persistence.queryCustomization.QueryCustomization;

@Repository
public class OrganizationDAOImpl extends GenericHibernateSortedDAOImpl<Organization> implements OrganizationDAO {
	@SuppressWarnings("unused")
	private static final Logger log = LoggerFactory.getLogger(OrganizationDAOImpl.class);

	@SuppressWarnings("unchecked")
	@Override
	public List<Organization> findByCriteria(String name, boolean includeLocalFacility, List<Long> localFacilityIds,
			boolean includeNational, QueryCustomization... customization) {
		if (!includeLocalFacility && !includeNational)
			return new ArrayList<>();

		if (includeLocalFacility && localFacilityIds == null)
			throw new IllegalArgumentException(
					"Parameter 'localFacilityIds' is required if includeLocalFacility is true");

		StringBuilder sb = new StringBuilder("select o from ").append(Organization.class.getName()).append(" o");
		sb.append(" left join o.vtkInstitution.vaFacility f");

		QueryCustomization cust = ArrayUtils.isEmpty(customization) ? new QueryCustomization() : customization[0];
		cust.appendRemainingJoins(sb, "o");

		if (cust.getOrderBy() == null)
			cust.setOrderBy("o.scope desc, f.stationNumber, o.name");

		List<String> whereClauseItems = new ArrayList<>();
		Map<String, Object> params = new HashMap<>();

		if (StringUtils.isNotBlank(name)) {
			whereClauseItems.add("LOWER(o.name) like :name1");
			params.put("name1", "%" + name.toLowerCase() + "%");
		}

		List<OrganizationScopeType> validScopeTypes = new ArrayList<>();
		if (includeLocalFacility) {
			validScopeTypes.add(OrganizationScopeType.LOCAL);

			whereClauseItems.add("o.scope <> :localScope or COALESCE(f.id, -1) in ( :localFacilityIds )");
			params.put("localScope", OrganizationScopeType.LOCAL);
			params.put("localFacilityIds", localFacilityIds);
		}

		if (includeNational)
			validScopeTypes.add(OrganizationScopeType.NATIONAL);

		whereClauseItems.add("o.scope in (:validScopes)");
		params.put("validScopes", validScopeTypes);

		Query q = constructQuery(em, sb, whereClauseItems, params, null, cust);

		return q.getResultList();
	}

	@Override
	public
	/*
	 * Should probably synchronize across HazelCast but just doing in-JVM for
	 * now - CPB
	 */
	synchronized int findNextAvailableCode(OrganizationScopeType scope, Long localFacilityId) {
		Number result = null;

		if (scope == OrganizationScopeType.LOCAL) {
			if (localFacilityId == null)
				throw new IllegalArgumentException("Must specify the localFacilityId if scope = Local");

			result = (Number) query("select max(o.code) + 1 from " + Organization.class.getName()
					+ " o where o.scope = :localScope and o.vtkInstitution.vaFacility.id = :facilityId")
							.setParameter("localScope", scope).setParameter("facilityId", localFacilityId)
							.getSingleResult();
			
			/* If scope = Local; then it will be assigned with the next available number in the sequence 900-9999. */
			
			if (result == null || result.intValue() == 1)
				result = 900;
			
			if (result.intValue() == 10000) {
				throw new AssertionError("Unable assign local code, maximum allocation reached at 9999");
			}
			
		} else if (scope == OrganizationScopeType.NATIONAL) {
			if (localFacilityId != null)
				throw new IllegalArgumentException("Must not specify a localFacilityId if scope = National");

			result = (Number) query(
					"select max(o.code) + 1 from " + Organization.class.getName() + " o where o.scope = :nationalScope")
							.setParameter("nationalScope", scope).getSingleResult();

			/*
			 * If we run out of nationals from 1-899, jump to 10000 and allocate
			 * from there to 99999 - CPB
			 */
			if (result != null && result.intValue() == 900)
				result = 10000;
		} else {
			throw new AssertionError("Unexpected scope value " + scope);
		}

		return result == null ? 1 : result.intValue();
	}

}
