package gov.va.vss.persistence.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.persistence.Query;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Repository;

import gov.va.vss.model.Volunteer;
import gov.va.vss.persistence.dao.VolunteerDAO;
import gov.va.vss.persistence.queryCustomization.QueryCustomization;

@Repository
public class VolunteerDAOImpl extends GenericHibernateSortedDAOImpl<Volunteer> implements VolunteerDAO {
	@SuppressWarnings("unused")
	private static final Logger log = LoggerFactory.getLogger(VolunteerDAOImpl.class);

	@Value("${volunteerSearch.maxResults}")
	private int maxResults;

	@SuppressWarnings("unchecked")
	@Override
	public List<Volunteer> findByCriteria(String firstName, String middleName, String lastName, String code,
			String addressStreet, String city, String state, String zip, String phone, String email,
			Collection<Long> vaFacilityIds, Collection<String> vaFacilitySta3ns, QueryCustomization... customization) {
		StringBuilder sb = new StringBuilder("select v from ").append(Volunteer.class.getName())
				.append(" v");
		sb.append(" left join fetch v.stationProfiles sp");
		sb.append(" left join fetch sp.vtkInstitution i");
		sb.append(" left join fetch sp.contactInfo ci");

		/* Don't bother with this yet - CPB */
		QueryCustomization cust = new QueryCustomization();

		List<String> whereClauseItems = new ArrayList<>();
		Map<String, Object> params = new HashMap<>();

		if (StringUtils.isNotBlank(firstName)) {
			whereClauseItems.add("LOWER(v.firstName) like :firstName");
			params.put("firstName", "%" + firstName.toLowerCase() + "%");
		}

		if (StringUtils.isNotBlank(lastName)) {
			whereClauseItems.add("LOWER(v.lastName) like :lastName");
			params.put("lastName", "%" + lastName.toLowerCase() + "%");
		}

		if (StringUtils.isNotBlank(code)) {
			whereClauseItems.add("LOWER(v.identifyingCode) like :identifyingCode");
			params.put("identifyingCode", "%" + code.toLowerCase() + "%");
		}

		if (vaFacilityIds != null) {
			whereClauseItems.add("i.vaFacility.id in (:vaFacilityIds)");
			params.put("vaFacilityIds", vaFacilityIds);
		}

		if (CollectionUtils.isNotEmpty(vaFacilitySta3ns)) {
			whereClauseItems.add("i.stationNumber in (:vaFacilitySta3ns)");
			params.put("vaFacilitySta3ns", vaFacilitySta3ns);
		}

		if (StringUtils.isNotBlank(addressStreet)) {
			whereClauseItems.add(
					"LOWER(ci.addressLine1) like :addressStreet" + " or LOWER(ci.addressLine2) like :addressStreet");
			params.put("addressStreet", "%" + addressStreet.toLowerCase() + "%");
		}

		if (StringUtils.isNotBlank(city)) {
			whereClauseItems.add("LOWER(ci.city) like :city");
			params.put("city", "%" + city.toLowerCase() + "%");
		}

		/*
		 * Match on State abbreviation if specified value has only two letters;
		 * otherwise, wildcard match on state name
		 */
		if (StringUtils.isNotBlank(state)) {
			state = state.trim();
			if (state.trim().length() == 2) {
				whereClauseItems.add("LOWER(ci.state.abbreviation) = :state");
				params.put("state", state.toLowerCase());
			} else {
				whereClauseItems.add("LOWER(ci.state.name) like :state");
				params.put("state", "%" + state.toLowerCase() + "%");
			}
		}

		/* Match only on the first five digits */
		if (StringUtils.isNotBlank(zip)) {
			whereClauseItems.add("SUBSTRING(ci.zip, 1, 5) = :zip");
			if (zip.length() > 5)
				zip = zip.substring(0, 5);
			params.put("zip", zip);
		}

		if (StringUtils.isNotBlank(phone)) {
			StringBuilder sb2 = new StringBuilder();
			for (String phoneField : new String[] { "phone", "phoneAlternate", "phoneAlternate2" })
				sb2.append(sb.length() > 0 ? " OR " : "") //
						.append("CONCAT(SUBSTRING(ci.").append(phoneField).append(",1,3), ") //
						.append("SUBSTRING(ci.").append(phoneField).append(", 5,3), ") //
						.append("SUBSTRING(ci.").append(phoneField).append(", 10,4)) like :phone");
			whereClauseItems.add(sb2.toString());
			params.put("phone", "%" + phone.replaceAll("\\D", "") + "%");
		}

		if (StringUtils.isNotBlank(email)) {
			whereClauseItems.add("LOWER(ci.email) like :email");
			params.put("email", "%" + email.toLowerCase() + "%");
		}

		Query q = constructQuery(em, sb, whereClauseItems, params, null, cust);

		int maxResults = this.maxResults;
		if (cust.getRowLimitation() != null)
			maxResults = cust.getRowLimitation();

		return q.setMaxResults(maxResults).getResultList();
	}

}
