package gov.va.vss.web.organization;

import java.time.LocalDate;
import java.time.Month;
import java.util.ArrayList;
import java.util.List;

import javax.servlet.http.HttpServletRequest;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataIntegrityViolationException;
import org.springframework.format.annotation.DateTimeFormat;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.ModelMap;
import org.springframework.validation.BindingResult;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.SessionAttributes;
import org.springframework.web.bind.support.SessionStatus;

import gov.va.shared.model.Permission;
import gov.va.shared.model.Permission.PermissionType;
import gov.va.shared.model.lookup.sds.State;
import gov.va.shared.model.lookup.sds.VAFacility.VAFacilityValue;
import gov.va.shared.util.SecurityUtil;
import gov.va.vss.model.NationalOfficial;
import gov.va.vss.model.Organization;
import gov.va.vss.model.OrganizationScopeType;
import gov.va.vss.model.lookup.StdVAVSTitle;
import gov.va.vss.model.lookup.VtkInstitution;
import gov.va.vss.service.validation.ServiceValidationException;
import gov.va.vss.util.DateUtil;
import gov.va.vss.web.AbstractAppController;
import gov.va.vss.web.ValidationException;
import gov.va.vss.web.breadcrumbs.Breadcrumb;

@Controller
@SessionAttributes(value = { AbstractAppController.DEFAULT_COMMAND_NAME })
public class OrganizationController extends AbstractAppController {

	@Autowired
	private OrganizationValidator organizationValidator;

	@RequestMapping("/organizationCreate.htm")
	@Breadcrumb("Create Organizations")
	@PreAuthorize("hasAnyAuthority('" + Permission.ORG_CODE_NATIONAL_CREATE + ", " + Permission.ORG_CODE_LOCAL_CREATE
			+ "')")
	public String createOrganization(ModelMap model, HttpServletRequest request) {
		Organization org = new Organization();

		if (SecurityUtil.hasAllPermissionsAtCurrentStation(PermissionType.ORG_CODE_NATIONAL_CREATE)) {
			if (isUserWorkingFacilityCO()) {
				org.setScope(OrganizationScopeType.NATIONAL);
			} else {
				setOrgAsLocal(org);
			}
		} else if (SecurityUtil.hasAllPermissionsAtCurrentStation(PermissionType.ORG_CODE_LOCAL_CREATE)) {
			setOrgAsLocal(org);
		} else {
			throw new SecurityException("The user does not have permission to create organizations");
		}

		OrganizationCommand command = new OrganizationCommand(org);
		// set default status to active
		org.setInactive(false);

		if (org.getScope() == OrganizationScopeType.LOCAL) {
			org.setOnNationalAdvisoryCommittee(false);
		}

		model.addAttribute(DEFAULT_COMMAND_NAME, command);

		createReferenceData(model);

		return "createOrganization";
	}

	private boolean isUserWorkingFacilityCO() {
		return VAFacilityValue.CENTRAL_OFFICE.getSta3n().equals(getSiteContext().getSta3n());
	}

	private void setOrgAsLocal(Organization org) {
		org.setScope(OrganizationScopeType.LOCAL);
		VtkInstitution vtkInstitution = vtkService.getOrCreateInstitutionForSDSFacility(getSiteContext().getId());
		org.setVtkInstitution(vtkInstitution);
	}

	@RequestMapping("/organizationEdit.htm")
	@Breadcrumb("Edit Organizations")
	@PreAuthorize("hasAnyAuthority('" + Permission.ORG_CODE_NATIONAL_READ + ", " + Permission.ORG_CODE_LOCAL_READ
			+ "')")
	public String editOrganization(@RequestParam long id, ModelMap model, HttpServletRequest request) {
		Organization org = organizationDAO.findRequiredByPrimaryKey(id);
		OrganizationCommand command = new OrganizationCommand(org);
		model.addAttribute(DEFAULT_COMMAND_NAME, command);
		createReferenceData(model);

		if (org.getScope() == OrganizationScopeType.NATIONAL) {
			if (isUserWorkingFacilityCO()) {
				updateModelForPermissionsIfNecessary(model, PermissionType.ORG_CODE_NATIONAL_CREATE);
			} else {
				SecurityUtil.ensureAllPermissionsAtCurrentFacility(PermissionType.ORG_CODE_NATIONAL_READ);
				setFormAsReadOnly(model, true);
			}
		} else {
			if (org.getVtkInstitution() != null && getSiteContext().equals(org.getVtkInstitution().getVaFacility())) {
				updateModelForPermissionsIfNecessary(model, PermissionType.ORG_CODE_LOCAL_CREATE);
			} else {
				SecurityUtil.ensureAllPermissionsAtCurrentFacility(PermissionType.ORG_CODE_LOCAL_READ);
				setFormAsReadOnly(model, true);
			}
		}

		return "editOrganization";
	}

	@RequestMapping("/organizationList.htm")
	@Breadcrumb("List Organizations")
	@PreAuthorize("hasAnyAuthority('" + Permission.ORG_CODE_NATIONAL_READ + ", " + Permission.ORG_CODE_LOCAL_READ
			+ "')")
	public String listOrganization(ModelMap model, HttpServletRequest request) {
		Long staId = getSiteContext().getId();

		boolean includeNational = SecurityUtil.hasAllPermissionsAtCurrentStation(PermissionType.ORG_CODE_NATIONAL_READ);
		boolean includeLocal = SecurityUtil.hasAllPermissionsAtCurrentStation(PermissionType.ORG_CODE_LOCAL_READ);
		List<Long> stationsToInclude = new ArrayList<Long>();
		stationsToInclude.add(staId);
		List<Organization> organizations = organizationDAO.findByCriteria(null, includeLocal, stationsToInclude,
				includeNational);

		OrganizationListCommand command = new OrganizationListCommand(organizations);
		model.addAttribute(DEFAULT_COMMAND_NAME, command);
		return "organizationList";
	}

	private void createReferenceData(ModelMap model) {
		model.put("allTypes", vtkOrganizationTypeDAO.findAllSorted());
		model.put("allNacStatuses", nacStatusDAO.findAll());
		model.put("allMonths", Month.values());
		model.put("allStates", stateDAO.findListOfStatesInUSA());
		model.put("allOfficialVAVSTitles", stdVAVSTitleDAO.findAll());
	}

	@RequestMapping("/organizationSubmit.htm")
	@PreAuthorize("hasAnyAuthority('" + Permission.ORG_CODE_NATIONAL_CREATE + ", " + Permission.ORG_CODE_LOCAL_CREATE
			+ "')")
	public String submitOrganization(@ModelAttribute(DEFAULT_COMMAND_NAME) OrganizationCommand command,
			BindingResult result, SessionStatus status, ModelMap model, HttpServletRequest request)
					throws ValidationException {
		Organization org = command.getOrganization();

		boolean isEdit = org.isPersistent();

		/* Validation step (JSR303, other custom logic in the validator) */
		organizationValidator.validate(command, result, false, "organization");
		boolean hasErrors = result.hasErrors();

		if (!hasErrors) {
			try {
				org = organizationService.saveOrUpdate(org, getCurrentUser(), getSiteContext().getId(), isEdit);
				userNotifier.notifyUserOnceWithMessage(request,
						getMessage(isEdit ? "organization.update.success" : "organization.create.success"));
			} catch (DataIntegrityViolationException e) {
				hasErrors = true;
				webValidationService.handle(e, result, "organization.create.error.dataIntegrityError");
			} catch (ServiceValidationException e) {
				webValidationService.handle(e, result);
				hasErrors = true;
			}
		}

		if (hasErrors) {
			createReferenceData(model);
			return isEdit ? "editOrganization" : "createOrganization";
		} else {
			status.setComplete();
			return "redirect:/organizationEdit.htm?id=" + org.getId();
		}
	}

	// ======================= National Official ==============================
	@RequestMapping("/nationalOfficialList")
	@PreAuthorize("hasAnyAuthority('" + Permission.ORG_CODE_NATIONAL_READ + ", " + Permission.ORG_CODE_LOCAL_READ
			+ "')")
	public @ResponseBody List<NationalOfficial> getOfficialForOrganization(@RequestParam long orgId) {
		List<NationalOfficial> officials = nationalOfficialDAO.findByCriteria(orgId, null);
		return officials;
	}

	@RequestMapping("/nationalOfficalCreateOrUpdate")
	@PreAuthorize("hasAnyAuthority('" + Permission.ORG_CODE_NATIONAL_CREATE + ", " + Permission.ORG_CODE_LOCAL_CREATE
			+ "')")
	public @ResponseBody NationalOfficial createOrUpdateNationalOfficial(@RequestParam Long organizationId,
			@RequestParam Long nationalOfficialId, @RequestParam String lastName, @RequestParam String firstName,
			@RequestParam String middleName, @RequestParam String suffix, @RequestParam String prefix,
			@RequestParam String title, @RequestParam boolean certifyingOfficial, @RequestParam String email,
			@RequestParam String streetAddress, @RequestParam String city, @RequestParam State state,
			@RequestParam String zip, @RequestParam String phone, @RequestParam StdVAVSTitle stdVAVSTitle,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam LocalDate vavsStartDate,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam LocalDate vavsEndDate,
			@RequestParam boolean nationalCommitteeMember,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam LocalDate nacStartDate,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam LocalDate nacEndDate)
					throws ServiceValidationException {

		if (organizationId == null) {
			throw new IllegalArgumentException("Organization Id cannot be null");
		}

		NationalOfficial official = null;
		if (nationalOfficialId == null) {
			official = new NationalOfficial();
			Organization org = organizationDAO.findByPrimaryKey(organizationId);
			if (org == null) {
				throw new IllegalArgumentException("Organization not found for id " + organizationId);
			}
			official.setOrganization(org);
		} else {
			official = nationalOfficialDAO.findRequiredByPrimaryKey(nationalOfficialId);
		}

		official.setLastName(lastName);
		official.setFirstName(firstName);
		official.setMiddleName(middleName);
		official.setSuffix(suffix);
		official.setPrefix(prefix);
		official.setTitle(title);
		official.setCertifyingOfficial(certifyingOfficial);
		official.setEmail(email);
		official.setStreetAddress(streetAddress);
		official.setCity(city);
		official.setZip(zip);
		official.setPhone(phone);
		official.setStdVAVSTitle(stdVAVSTitle);
		official.setVavsStartDate(vavsStartDate);
		official.setVavsEndDate(vavsEndDate);
		official.setNationalCommitteeMember(nationalCommitteeMember);
		official.setNacStartDate(nacStartDate);
		official.setNacEndDate(nacEndDate);
		official.setState(state != null ? vtkService.getOrCreateForSDSState(state.getId()) : null);

		official = nationalOfficialService.saveOrUpdate(official);

		return official;
	}

	@RequestMapping("/nationalOfficialDelete")
	@PreAuthorize("hasAnyAuthority('" + Permission.ORG_CODE_NATIONAL_CREATE + ", " + Permission.ORG_CODE_LOCAL_CREATE
			+ "')")
	public @ResponseBody boolean deleteNationalOfficial(@RequestParam long officialId)
			throws ServiceValidationException {
		nationalOfficialService.delete(officialId);
		return true;
	}

}
