package gov.va.vss.web.serviceParameters;

import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.util.List;

import javax.servlet.http.HttpServletRequest;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.format.annotation.DateTimeFormat;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.ModelMap;
import org.springframework.validation.BindingResult;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.SessionAttributes;
import org.springframework.web.bind.support.SessionStatus;

import gov.va.shared.model.Permission;
import gov.va.shared.model.Permission.PermissionType;
import gov.va.vss.model.VoluntaryServiceParameters;
import gov.va.vss.model.VoluntaryServiceStaff;
import gov.va.vss.model.VtkStaffTitle;
import gov.va.vss.model.lookup.VtkInstitution;
import gov.va.vss.model.lookup.VtkVoluntaryServiceType;
import gov.va.vss.model.lookup.VtkVoluntaryServiceType.VtkVoluntaryServiceTypeValue;
import gov.va.vss.service.validation.ServiceValidationException;
import gov.va.vss.util.DateUtil;
import gov.va.vss.web.AbstractAppController;
import gov.va.vss.web.ValidationException;
import gov.va.vss.web.breadcrumbs.Breadcrumb;

@Controller
@SessionAttributes(value = { AbstractAppController.DEFAULT_COMMAND_NAME })
public class ServiceParametersController extends AbstractAppController {

	@Autowired
	private ServiceParametersValidator serviceParametersValidator;

	@RequestMapping("/editVoluntaryService.htm")
	@Breadcrumb("Edit Voluntary Service")
	@PreAuthorize("hasAuthority('" + Permission.PERM_CODE_SERVICE_READ + "')")
	public String editVoluntaryService(ModelMap model, HttpServletRequest request) {
		VtkInstitution vtkInstitution = vtkService.getOrCreateInstitutionForSDSFacility(getSiteContext().getId());
		VoluntaryServiceParameters serviceParameters = vtkInstitution.getVoluntaryServiceParameters();
		if (serviceParameters.getVtkVoluntaryServiceType() == null) {
			VtkVoluntaryServiceType vtkVoluntaryServiceType = vtkVoluntaryServiceTypeDAO
					.findByLookup(VtkVoluntaryServiceTypeValue.SERVICE);
			serviceParameters.setVtkVoluntaryServiceType(vtkVoluntaryServiceType);
		}
		ServiceParametersCommand command = new ServiceParametersCommand(serviceParameters);
		model.addAttribute(DEFAULT_COMMAND_NAME, command);
		createReferenceData(model);
		updateModelForPermissionsIfNecessary(model, PermissionType.PERM_CODE_SERVICE_CREATE);
		return "editVoluntaryService";
	}

	@RequestMapping("/voluntaryServiceStaff")
	// @PreAuthorize("hasAuthority('" + Permission.PERM_CODE_SERVICE_READ +
	// "')")
	public @ResponseBody List<VoluntaryServiceStaff> getStaffForCurrentStation() {
		List<VoluntaryServiceStaff> staff = voluntaryServiceStaffDAO.findByCriteria(getSiteContext().getSta3n(), null,
				null);
		return staff;
	}

	@RequestMapping("/voluntaryServiceStaffCreateOrUpdate")
	// @PreAuthorize("hasAuthority('" + Permission.PERM_CODE_SERVICE_READ +
	// "')")
	public @ResponseBody VoluntaryServiceStaff createOrUpdateStaff(@RequestParam Long serviceStaffId,
			@RequestParam String newStaffUserName, @RequestParam String nickName, @RequestParam String staffNamePrefix,
			@RequestParam VtkStaffTitle staffVavsRole, @RequestParam String staffGrade,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam LocalDate staffVavsStartDate,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam LocalDate staffVavsEndDate,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam LocalDate staffRetirementEligibleDate,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam LocalDate staffRetirementEstimateDate,
			@RequestParam boolean staffIsVavsLeadership, @RequestParam String staffComment)
					throws ServiceValidationException {

		VoluntaryServiceStaff serviceStaff = null;
		if (serviceStaffId == null) {
			serviceStaff = voluntaryServiceStaffService.createOrRetrieveServiceStaff(newStaffUserName,
					getSiteContext().getId());
		} else {
			serviceStaff = voluntaryServiceStaffDAO.findRequiredByPrimaryKey(serviceStaffId);
		}

		serviceStaff.set    Name(nickName);
		serviceStaff.setNamePrefix(staffNamePrefix);
		serviceStaff.setStaffTitle(staffVavsRole);
		serviceStaff.setGrade(staffGrade);
		serviceStaff.setVavsStartDate(staffVavsStartDate);
		serviceStaff.setVavsEndDate(staffVavsEndDate);
		serviceStaff.setRetirementEligibleDate(staffRetirementEligibleDate);
		serviceStaff.setRetirementEstimateDate(staffRetirementEstimateDate);
		serviceStaff.setVavsLeadership(staffIsVavsLeadership);
		serviceStaff.setComment(staffComment);
		serviceStaff = voluntaryServiceStaffService.saveOrUpdate(serviceStaff);

		return serviceStaff;
	}

	@RequestMapping("/deleteServiceStaff")
	@PreAuthorize("hasAuthority('" + Permission.VOL_SVC_STAFF_DELETE + "')")
	public @ResponseBody boolean deleteStaff(@RequestParam long serviceStaffId) throws ServiceValidationException {
		voluntaryServiceStaffService.delete(serviceStaffId);
		return true;
	}

	private void createReferenceData(ModelMap model) {
		model.put("allServiceTypes", vtkVoluntaryServiceTypeDAO.findAllSorted());
		model.put("allStaffTitles", vtkStaffTitleDAO.findAllSorted());
	}

	@RequestMapping("/serviceParametersSubmit.htm")
	@PreAuthorize("hasAuthority('" + Permission.PERM_CODE_SERVICE_CREATE + "')")
	public String submitServiceParameters(@ModelAttribute(DEFAULT_COMMAND_NAME) ServiceParametersCommand command,
			BindingResult result, SessionStatus status, ModelMap model, HttpServletRequest request)
					throws ValidationException {
		VoluntaryServiceParameters serviceParameters = command.getServiceParameters();

		/*
		 * Perform any remaining binding steps from Command obj -> Voluntary
		 * Service Parameters here
		 */

		/* Validation step (JSR303, other custom logic in the validator) */
		serviceParametersValidator.validate(command, result, false, "serviceParameters");
		boolean hasErrors = result.hasErrors();

		if (!hasErrors) {
			try {
				if (command.isChangeLastReviewedDate())
					serviceParameters.setLastUpdated(ZonedDateTime.now());
				serviceParameters = voluntaryServiceParametersService.saveOrUpdate(serviceParameters);
				userNotifier.notifyUserOnceWithMessage(request, getMessage("serviceParameters.update.success"));
			} catch (ServiceValidationException e) {
				webValidationService.handle(e, result);
				hasErrors = true;
			}
		}

		if (hasErrors) {
			createReferenceData(model);
			return "editVoluntaryService";
		} else {
			status.setComplete();
			return "redirect:/editVoluntaryService.htm";
		}
	}

}
