package gov.va.shared.util;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.commons.lang3.StringUtils;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;

import gov.va.shared.model.AppUser;
import gov.va.shared.model.AppUserFacility;
import gov.va.shared.model.Permission.PermissionType;
import gov.va.shared.model.Role;
import gov.va.shared.model.lookup.sds.VAFacility;

public class SecurityUtil {

	public static void resetUserAuthoritiesForFacility(long facilityId) {
		AppUser user = getCurrentUser();
		Collection<? extends GrantedAuthority> authoritiesForFacility = getAuthoritiesAtFacility(user, facilityId);
		SecurityContext context = SecurityContextHolder.getContext();
		Authentication auth = context.getAuthentication();
		context.setAuthentication(new UsernamePasswordAuthenticationToken(auth.getPrincipal(), auth.getCredentials(),
				authoritiesForFacility));
	}

	private static Set<? extends GrantedAuthority> getAuthoritiesAtFacility(AppUser appUser, long facilityId) {
		AppUserFacility auf = appUser.getAppUserFacility(facilityId);
		if (auf == null)
			return new HashSet<>();

		if (auf.isRolesCustomizedForFacility())
			return getUnmodifiableCombinedAuthorities(auf.getRoles());

		return getUnmodifiableCombinedAuthorities(appUser.getBasicGlobalRoles());
	}

	private static Set<? extends GrantedAuthority> getUnmodifiableCombinedAuthorities(Collection<Role> roles) {
		SortedSet<GrantedAuthority> results = new TreeSet<>();
		for (Role role : roles) {
			if (role.isUsedAsPermission())
				results.add(role);
			results.addAll(role.getPermissions());
		}
		return Collections.unmodifiableSortedSet(results);
	}

	public static void assertPermission(PermissionType permissionType) {
		if (permissionType != null && !hasAnyPermissionAtCurrentFacility(permissionType))
			throw new AccessDeniedException("Permission " + permissionType.getName() + " needed for this operation");
	}

	public static boolean isCurrentUserAssignedToFacility(long facilityId) {
		return isUserAssignedToFacility(getCurrentUser(), facilityId);
	}

	public static boolean isUserAssignedToFacility(AppUser user, long facilityId) {
		return user.getAppUserFacility(facilityId) != null;
	}

	/**
	 * Returns whether the specified user is assigned to the specified station.
	 * If the user holds the NATIONAL_ADMIN role and the
	 * considerNationalAdminsImplicitlyAssigned is true, true is returned
	 * regardless of the passed-in station or the user's explicitly assigned
	 * facilities.
	 */
	public static boolean isUserAssignedToFacility(AppUser user, long facilityId,
			boolean considerNationalAdminsImplicitlyAssigned) {
		if (considerNationalAdminsImplicitlyAssigned && user.isNationalAdmin())
			return true;
		return isUserAssignedToFacility(user, facilityId);
	}

	/**
	 * Returns true iff the currently authenticated user has any of the
	 * specified permissions (or roles with the usedAsPermission flag set to
	 * true)
	 */
	public static boolean hasAnyPermissionAtCurrentFacility(PermissionType... permissions) {
		return hasAnyPermission(false, null, permissions);
	}

	/**
	 * Returns true iff the currently authenticated user has any of the
	 * specified permissions (or roles with the usedAsPermission flag set to
	 * true) at the facility with the specified vaFacilityId
	 */
	public static boolean hasAnyPermissionAtFacility(long vaFacilityId, PermissionType... permissions) {
		return hasAnyPermission(false, vaFacilityId, permissions);
	}

	/**
	 * Throws an AccessDeniedException if the currently authenticated user does
	 * not have any of the specified permissions (or roles with the
	 * usedAsPermission flag set to true)
	 */
	public static void ensureAnyPermissionAtCurrentFacility(PermissionType... permissions) {
		hasAnyPermission(true, null, permissions);
	}

	/**
	 * Throws an AccessDeniedException if the currently authenticated user does
	 * not have any of the specified permissions (or roles with the
	 * usedAsPermission flag set to true) at the facility with the specified
	 * vaFacilityId
	 */
	public static void ensureAnyPermissionAtFacility(long vaFacilityId, PermissionType... permissions) {
		hasAnyPermission(true, vaFacilityId, permissions);
	}

	private static boolean hasAnyPermission(boolean throwExceptionOnFailure, Long vaFacilityId,
			PermissionType... permissionsRequired) {
		Collection<? extends GrantedAuthority> permissionsAssigned = null;
		if (vaFacilityId == null) {
			permissionsAssigned = SecurityContextHolder.getContext().getAuthentication().getAuthorities();
		} else {
			permissionsAssigned = getAuthoritiesAtFacility(getCurrentUser(), vaFacilityId);
		}

		for (GrantedAuthority authority : permissionsAssigned) {
			String a = authority.getAuthority();
			for (PermissionType permission : permissionsRequired) {
				if (permission.getName().equals(a)) {
					return true;
				}
			}
		}

		if (throwExceptionOnFailure) {
			throw new AccessDeniedException(
					"The user is not assigned any of the permissions {" + StringUtils.join(permissionsRequired, ", ")
							+ "} at " + (vaFacilityId == null ? "the current station" : "station ID " + vaFacilityId));
		} else {
			return false;
		}
	}

	/**
	 * Returns true iff the currently authenticated user has all of the
	 * specified permissions (or roles with the usedAsPermission flag set to
	 * true)
	 */
	public static boolean hasAllPermissionsAtCurrentFacility(PermissionType... permissions) {
		return hasAllPermissions(false, null, permissions);
	}

	/**
	 * Returns true iff the currently authenticated user has all of the
	 * specified permissions (or roles with the usedAsPermission flag set to
	 * true) at the facility with the specified vaFacilityId
	 */
	public static boolean hasAllPermissionsAtFacility(long vaFacilityId, PermissionType... permissions) {
		return hasAllPermissions(false, vaFacilityId, permissions);
	}

	/**
	 * Throws an AccessDeniedException if the currently authenticated user does
	 * not have all of the specified permissions (or roles with the
	 * usedAsPermission flag set to true) at the facility with the specified
	 * vaFacilityId
	 */
	public static void ensureAllPermissionsAtFacility(long vaFacilityId, PermissionType... permissions) {
		hasAllPermissions(true, vaFacilityId, permissions);
	}

	/**
	 * Throws an AccessDeniedException if the currently authenticated user does
	 * not have all of the specified permissions (or roles with the
	 * usedAsPermission flag set to true)
	 */
	public static void ensureAllPermissionsAtCurrentFacility(PermissionType... permissions) {
		hasAllPermissions(true, null, permissions);
	}

	private static boolean hasAllPermissions(boolean throwExceptionOnFailure, Long vaFacilityId,
			PermissionType... permissionsRequired) {

		Collection<? extends GrantedAuthority> permissionsAssigned = null;
		if (vaFacilityId == null) {
			permissionsAssigned = SecurityContextHolder.getContext().getAuthentication().getAuthorities();
		} else {
			permissionsAssigned = getAuthoritiesAtFacility(getCurrentUser(), vaFacilityId);
		}

		outer: //
		for (PermissionType permission : permissionsRequired) {
			for (GrantedAuthority authority : permissionsAssigned) {
				String a = authority.getAuthority();
				if (permission.getName().equals(a)) {
					continue outer;
				}
			}
			if (throwExceptionOnFailure) {
				throw new AccessDeniedException("The user is not assigned the permission '" + permission.getName()
						+ " at " + (vaFacilityId == null ? "the current station" : "station ID " + vaFacilityId));
			} else {
				return false;
			}
		}

		return true;
	}

	public static Set<VAFacility> getAllFacilitiesWhereUserHasPermissions(PermissionType... permissionsRequired) {
		AppUser appUser = getCurrentUser();
		Set<VAFacility> results = new HashSet<VAFacility>();
		for (AppUserFacility auf : appUser.getFacilities()) {
			if (hasAllPermissionsAtFacility(auf.getFacility().getId(), permissionsRequired))
				results.add(auf.getFacility());
		}
		return results;
	}

	public static AppUser getCurrentUser() {
		Authentication authentication = SecurityContextHolder.getContext().getAuthentication();
		if (authentication == null)
			return null;

		Object principal = authentication.getPrincipal();
		if (principal instanceof AppUser)
			return (AppUser) principal;

		return null;
	}

	public static String getCurrentUserName() {
		AppUser currentUser = getCurrentUser();
		return currentUser != null ? currentUser.getUsername() : null;
	}

}