package gov.va.shared.util;

import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.google.common.base.Predicate;

import gov.va.shared.model.timeline.TimelineEvent;
import gov.va.shared.model.timeline.TimelineEventWithStatus;
import gov.va.shared.model.timeline.TimelineInterval;

public class TimelineUtil {

	/**
	 * Returns the most recent TimelineEvent in the specified SortedSet,
	 * relative to the specified ZonedDateTime. The returned TimelineEvent will
	 * have an effectiveDate on or before the specified ZonedDateTime.
	 * <p>
	 * If the specified SortedSet is null or empty, or all TimelineEvents in the
	 * SortedSet have effectiveDates after the specified ZonedDateTime, then
	 * null is returned.
	 * <p>
	 * This method assumes the natural ordering of all items in the specified
	 * SortedSet follow normal chronological order.
	 * 
	 * @param c
	 *            The SortedSet to search
	 * @return See description
	 */
	public static <T extends TimelineEvent> T getTimelineEventEffectiveAsOfDate(SortedSet<T> c, ZonedDateTime dt) {
		if (c == null)
			return null;

		T lastItem = null;

		for (T e : c) {
			if (e.getEffectiveDate().isAfter(dt))
				return lastItem;
			lastItem = e;
		}

		return lastItem;
	}

	/**
	 * Returns all TimelineEvents in the specified timeline whose effective
	 * intervals (defined as the time between the event's effective date and the
	 * subsequent TimelineEvent's effective date) overlap the specified
	 * TimelineInterval.
	 * 
	 * @param c
	 * @param ti
	 * @return
	 */
	public static <T extends TimelineEvent> SortedSet<T> getTimelineEventsActiveOverInterval(SortedSet<T> c,
			TimelineInterval ti) {
		SortedSet<T> results = new TreeSet<>();
		if (c == null || c.isEmpty())
			return results;

		SortedMap<T, TimelineInterval> eventIntervalMap = translateEventsToIntervals(c);

		boolean matchedOne = false;

		for (Map.Entry<T, TimelineInterval> entry : eventIntervalMap.entrySet()) {
			T event = entry.getKey();
			TimelineInterval interval = entry.getValue();

			if (intervalsOverlap(interval.getStartDate(), interval.getEndDate(), ti.getStartDate(), ti.getEndDate())) {
				results.add(event);
				matchedOne = true;
			} else if (matchedOne) {
				/*
				 * We can safely bail out here since we must be after the last
				 * match chronologically
				 */
				break;
			}
		}

		return results;
	}

	/**
	 * Returns a set of Intervals representing the time spans between the
	 * specified TimelineEvents. The keys in the returned map are the original
	 * TimelineEvents and the values are Intervals starting with that event's
	 * effective date and ending with the next sequential event's effective
	 * date.
	 * <p>
	 * Note that the last item in the returned Map will always have a null end
	 * date.
	 * 
	 * @param events
	 * @return
	 */
	public static <T extends TimelineEvent> SortedMap<T, TimelineInterval> translateEventsToIntervals(
			SortedSet<T> events) {
		SortedMap<T, TimelineInterval> results = new TreeMap<>();
		if (events == null || events.isEmpty())
			return results;

		T previous = null;

		for (T event : events) {
			if (previous == null) {
				previous = event;
				continue;
			}
			results.put(previous, new SimpleTimelineInterval(previous.getEffectiveDate(), event.getEffectiveDate()));
			previous = event;
		}

		if (previous != null)
			results.put(previous, new SimpleTimelineInterval(previous.getEffectiveDate(), null));

		return results;
	}

	public static TimelineInterval getOverlap(TimelineInterval a, TimelineInterval b) {
		return getOverlap(a.getStartDate(), a.getEndDate(), b.getStartDate(), b.getEndDate());
	}

	public static TimelineInterval getOverlap(ZonedDateTime startDate, ZonedDateTime endDate, ZonedDateTime startDate2,
			ZonedDateTime endDate2) {
		if (!intervalsOverlap(startDate, endDate, startDate2, endDate2))
			return null;

		ZonedDateTime start = null;
		if (startDate == null) {
			start = startDate2;
		} else if (startDate2 == null) {
			start = startDate;
		} else {
			start = startDate.isAfter(startDate2) ? startDate : startDate2;
		}

		ZonedDateTime end = null;
		if (endDate == null) {
			end = endDate2;
		} else if (endDate2 == null) {
			end = endDate;
		} else {
			end = endDate.isAfter(endDate2) ? endDate2 : endDate;
		}

		return new SimpleTimelineInterval(start, end);
	}

	/**
	 * Returns true if and only if the interval represented by {myStart, myEnd}
	 * overlaps the interval represented by {aStart, aEnd}. Intervals are always
	 * inclusive of the start date and exclusive of the end date. Any of the
	 * values specified may be null, which signifies the first moment in time
	 * (for start dates) or the last moment in time (for end dates).
	 * 
	 * @param myStart
	 * @param myEnd
	 * @param aStart
	 * @param aEnd
	 * @return
	 */
	public static boolean intervalsOverlap(ZonedDateTime myStart, ZonedDateTime myEnd, ZonedDateTime aStart,
			ZonedDateTime aEnd) {
		if (myStart == null && myEnd == null)
			return true;
		if (aStart == null && aEnd == null)
			return true;

		if (myEnd == null && aEnd == null)
			return true;
		if (myStart == null && aStart == null)
			return true;

		if (myEnd == null)
			return aEnd.isAfter(myStart);
		if (aEnd == null)
			return myEnd.isAfter(aStart);

		if (myStart == null)
			return aStart.isBefore(myEnd);
		if (aStart == null)
			return myStart.isBefore(aEnd);

		return myStart.isBefore(aEnd) && myEnd.isAfter(aStart);
	}

	/**
	 * Returns true if and only if the interval represented by {myStart, myEnd}
	 * completely contains the interval represented by {aStart, aEnd}. Intervals
	 * are always inclusive of the start date and exclusive of the end date. Any
	 * of the values specified may be null, which signifies the first moment in
	 * time (for start dates) or the last moment in time (for end dates).
	 * 
	 * @param myStart
	 * @param myEnd
	 * @param aStart
	 * @param aEnd
	 * @return
	 */
	public static boolean intervalIsEnclosedBy(ZonedDateTime myStart, ZonedDateTime myEnd, ZonedDateTime aStart,
			ZonedDateTime aEnd) {
		if (myStart == null && myEnd == null)
			return true;
		if (myStart == null)
			return aEnd != null && !aEnd.isAfter(myEnd);
		if (myEnd == null)
			return aStart != null && !aStart.isBefore(myStart);

		if (aStart == null || aEnd == null)
			return false;
		return !aStart.isBefore(myStart) && !aEnd.isAfter(myEnd);
	}

	public static boolean intervalsOverlap(TimelineInterval t1, TimelineInterval t2) {
		return intervalsOverlap(t1.getStartDate(), t1.getEndDate(), t2.getStartDate(), t2.getEndDate());
	}

	/**
	 * Returns an unmodifiable list of length 3:
	 * <ol>
	 * <li>Item 1: The TimelineInterval that comes immediately prior to the
	 * specified date, but whose date range doesn't contain the date
	 * <li>Item 2: The TimelineInterval whose date range overlaps the specified
	 * date (if more than one, then the last matching one is used)
	 * <li>Item 3: The TimelineInterval that comes immediately after the
	 * specified date, but whose date range doesn't contain the date.
	 * </ol>
	 * A TimelineInterval "contains" the date if the date is >= the start date
	 * and < the end date.
	 * 
	 * @param assignments
	 * @param date
	 * @return
	 */
	public static <T extends TimelineInterval> List<T> getBeforeDuringAndAfterIntervals(SortedSet<T> assignments,
			ZonedDateTime date) {
		T before = null;
		T during = null;
		T after = null;

		for (T assignment : assignments) {
			if (assignment.getEndDate() != null && !date.isBefore(assignment.getEndDate()))
				before = assignment;
			if (isActiveAsOfDate(assignment, date))
				during = assignment;
			if (assignment.getStartDate() != null && assignment.getStartDate().isAfter(date)) {
				after = assignment;
				break;
			}
		}
		return Arrays.asList(before, during, after);
	}

	/**
	 * Returns an unmodifiable List of length 3:
	 * <ol>
	 * <li>Item 1: A SortedSet containing all TimelineIntervals that come prior
	 * to the specified date, but whose date range doesn't contain the date
	 * <li>Item 2: A SortedSet containing all TimelineIntervals whose date range
	 * overlaps the specified date
	 * <li>Item 3: A SortedSet containing all TimelineIntervals that come after
	 * the specified date, but whose date range doesn't contain the date.
	 * </ol>
	 * A TimelineInterval "contains" the date if the date is >= the start date
	 * and < the end date. The returned List may contain empty SortedSets, but
	 * never null for any of the three items.
	 * 
	 * @param assignments
	 * @param date
	 * @return
	 */
	public static <T extends TimelineInterval> List<SortedSet<T>> getBeforeDuringAndAfterIntervalSets(
			SortedSet<T> assignments, ZonedDateTime date) {
		SortedSet<T> before = new TreeSet<>();
		SortedSet<T> during = new TreeSet<>();
		SortedSet<T> after = new TreeSet<>();

		for (T assignment : assignments) {
			if (assignment.getEndDate() != null && !date.isBefore(assignment.getEndDate()))
				before.add(assignment);
			if (isActiveAsOfDate(assignment, date))
				during.add(assignment);
			if (assignment.getStartDate() != null && assignment.getStartDate().isAfter(date)) {
				after.add(assignment);
			}
		}
		return Arrays.asList(before, during, after);
	}

	public static enum DateComparisonType {
		BEFORE, DURING, AFTER
	};

	public static <T extends TimelineInterval> T selectSingleApplicableIntervalRelativeTo(SortedSet<T> items,
			ZonedDateTime date, DateComparisonType... comparisons) {
		List<T> bdas = getBeforeDuringAndAfterIntervals(items, date);
		for (DateComparisonType type : comparisons) {
			T selection = null;
			if (type == DateComparisonType.BEFORE)
				selection = bdas.get(0);
			if (type == DateComparisonType.DURING)
				selection = bdas.get(1);
			if (type == DateComparisonType.AFTER)
				selection = bdas.get(2);
			if (selection != null)
				return selection;
		}
		return null;
	}

	public static <T extends TimelineInterval> SortedSet<T> selectApplicableIntervalSetRelativeTo(SortedSet<T> items,
			ZonedDateTime date, DateComparisonType... comparisons) {
		List<SortedSet<T>> bdas = getBeforeDuringAndAfterIntervalSets(items, date);
		for (DateComparisonType type : comparisons) {
			SortedSet<T> selection = null;
			if (type == DateComparisonType.BEFORE) {
				selection = bdas.get(0);
			} else if (type == DateComparisonType.DURING) {
				selection = bdas.get(1);
			} else if (type == DateComparisonType.AFTER) {
				selection = bdas.get(2);
			} else {
				throw new AssertionError("Unexpected value " + type);
			}
			
			if (selection != null && !selection.isEmpty())
				return selection;
		}
		return null;
	}

	/**
	 * Returns true if the next event after the passed in timestamp is active.
	 * 
	 * @param c
	 * @param date
	 * @return
	 */
	public static <T extends TimelineEventWithStatus> boolean isReactivationDueInFuture(SortedSet<T> c,
			ZonedDateTime date) {
		boolean reactivationDueInFuture = false;
		List<T> beforeAfterEvents = getBeforeAndAfterEvents(c, date);

		T afterEvent = beforeAfterEvents.get(1);
		if (afterEvent != null) {
			reactivationDueInFuture = afterEvent.getStatus().isActive();
		}

		return reactivationDueInFuture;
	}

	/**
	 * Returns a list of length 2:
	 * <ol>
	 * <li>Item 1: The TimelineEvent that comes immediately prior to the
	 * specified date, and which determines the status on the specified date
	 * (possibly null)
	 * <li>Item 2: The TimelineEvent that comes immediately after the specified
	 * date, signifying the last time the status on the specified date is valid
	 * (possibly null)
	 * </ol>
	 * 
	 * @param events
	 * @param date
	 * @return
	 */
	public static <T extends TimelineEvent> List<T> getBeforeAndAfterEvents(SortedSet<T> events, ZonedDateTime date) {
		T before = null;
		T after = null;

		T previousEvent = null;
		for (T event : events) {
			if (event.getEffectiveDate().isAfter(date)) {
				after = event;
				break;
			} else {
				before = previousEvent;
			}
			previousEvent = event;
		}

		return Arrays.asList(before, after);
	}

	/**
	 * Returns a list of active assignments as of a specific date
	 * 
	 * @param assignments
	 * @param date
	 * @return
	 */

	public static <T extends TimelineInterval> List<T> getActiveAssignments(SortedSet<T> assignments,
			ZonedDateTime date) {
		List<T> activeAssignments = new ArrayList<>();

		for (T assignment : assignments)
			if (isActiveAsOfDate(assignment, date))
				activeAssignments.add(assignment);

		return activeAssignments;
	}

	/**
	 * Returns a list of inactive (in the past or yet to be activated)
	 * assignments as of a specific date
	 * 
	 * @param assignments
	 * @param date
	 * @return
	 */

	public static <T extends TimelineInterval> List<T> getInactiveIntervals(SortedSet<T> assignments,
			ZonedDateTime date) {
		List<T> inactiveAssignments = new ArrayList<>();

		for (T assignment : assignments)
			if (!isActiveAsOfDate(assignment, date))
				inactiveAssignments.add(assignment);

		return inactiveAssignments;
	}

	/**
	 * Returns true if the specified effectiveDate is greater than or equal to
	 * the specified TimelineInterval's start date, and either the
	 * TimelineInterval doesn't have an end date or its end date is after the
	 * effectiveDate.
	 * 
	 * @param ti
	 * @param effectiveDate
	 * @return
	 */
	public static boolean isActiveAsOfDate(TimelineInterval ti, ZonedDateTime effectiveDate) {
		if (ti.getStartDate() == null && ti.getEndDate() == null)
			return true;
		if (ti.getStartDate() == null)
			return effectiveDate.isBefore(ti.getEndDate());
		if (ti.getEndDate() == null)
			return !effectiveDate.isBefore(ti.getStartDate());
		return !effectiveDate.isBefore(ti.getStartDate()) && ti.getEndDate().isAfter(effectiveDate);
	}

	/**
	 * Checks all combinations of the items in intervalsToConsider1 and
	 * intervalsToConsider2 and returns either null (if none of the intervals
	 * overlap with any other) or a List of size two, containing an example of
	 * two items that overlap. Note that multiple overlapping intervals may
	 * exist, but only the first discovered instance is returned.
	 * 
	 * @param intervalsToConsider1
	 * @param intervalsToConsider2
	 * @return
	 */
	public static <T extends TimelineInterval> List<T> getOverlappingIntervals(Collection<T> intervalsToConsider1,
			Collection<T> intervalsToConsider2) {
		for (T t1 : intervalsToConsider1) {
			for (T t2 : intervalsToConsider2) {
				if (intervalsOverlap(t1, t2))
					return Arrays.asList(t1, t2);
			}
		}
		return null;
	}

	/**
	 * Checks all combinations of the items in intervalsToConsider and returns
	 * either null (if none of the intervals overlap with any other) or a List
	 * of size two, containing an example of two items that overlap. Note that
	 * multiple overlapping intervals may exist, but only the first discovered
	 * instance is returned.
	 * 
	 * @param intervalsToConsider
	 * @return
	 */
	public static <T extends TimelineInterval> List<TimelineInterval[]> getOverlappingIntervals(
			Collection<T> intervalsToConsider) {
		List<T> intervalsToConsiderList = new ArrayList<>(intervalsToConsider);
		List<TimelineInterval[]> results = new ArrayList<>();

		for (int i = 0; i < intervalsToConsiderList.size() - 1; i++) {
			for (int j = i + 1; j < intervalsToConsiderList.size(); j++) {
				T t1 = intervalsToConsiderList.get(i);
				T t2 = intervalsToConsiderList.get(j);
				if (intervalsOverlap(t1, t2))
					results.add(new TimelineInterval[] { t1, t2 });
			}
		}
		return results.isEmpty() ? null : results;
	}

	/**
	 * Returns those items in the specified wrappedIntervals collection that are
	 * not fully enclosed by at least one of the items in the wrappingIntervals
	 * collection.
	 * 
	 * @param wrappingIntervals
	 * @param wrappedIntervals
	 * @return A list of non-wrapped intervals (never null)
	 */
	public static <T extends TimelineInterval> List<T> getNonWrappedIntervals(Collection<T> wrappingIntervals,
			Collection<T> wrappedIntervals) {
		List<T> results = new ArrayList<>(wrappedIntervals.size());
		for (T wrappedInterval : wrappedIntervals) {
			boolean enclosed = false;
			for (T wrappingInterval : wrappingIntervals) {
				if (intervalIsEnclosedBy(wrappingInterval.getStartDate(), wrappingInterval.getEndDate(),
						wrappedInterval.getStartDate(), wrappedInterval.getEndDate())) {
					enclosed = true;
					break;
				}
			}
			if (!enclosed)
				results.add(wrappedInterval);
		}
		return results;
	}

	public static class SimpleTimelineInterval implements TimelineInterval, Comparable<TimelineInterval> {
		private ZonedDateTime startDate;
		private ZonedDateTime endDate;

		public SimpleTimelineInterval() {
			super();
		}

		public SimpleTimelineInterval(ZonedDateTime startDate, ZonedDateTime endDate) {
			this.startDate = startDate;
			this.endDate = endDate;
		}

		@Override
		public ZonedDateTime getStartDate() {
			return startDate;
		}

		public void setStartDate(ZonedDateTime startDate) {
			this.startDate = startDate;
		}

		@Override
		public ZonedDateTime getEndDate() {
			return endDate;
		}

		public void setEndDate(ZonedDateTime endDate) {
			this.endDate = endDate;
		}

		public String toString() {
			return "{" + startDate + " - " + endDate + "}";
		}

		public boolean equals(Object o) {
			if (o == null)
				return false;
			if (this == o)
				return true;
			if (!(o instanceof TimelineInterval))
				return false;
			TimelineInterval o1 = (TimelineInterval) o;
			return new EqualsBuilder().append(startDate, o1.getStartDate()).append(endDate, o1.getEndDate()).isEquals();
		}

		public int hashCode() {
			return new HashCodeBuilder().append(startDate).append(endDate).toHashCode();
		}

		@Override
		public int compareTo(TimelineInterval o) {
			if (equals(o))
				return 0;
			if (startDate == null && o.getStartDate() != null)
				return -1;
			if (startDate != null && o.getStartDate() == null)
				return 1;
			if (startDate != null && o.getStartDate() != null)
				return startDate.compareTo(o.getStartDate());
			// both start days are null
			if (endDate == null && o.getEndDate() != null)
				return 1;
			if (endDate != null && o.getEndDate() == null)
				return -1;
			if (endDate != null && o.getEndDate() != null)
				return endDate.compareTo(o.getEndDate());
			return 0;
		}
	}

	/**
	 * A comparator which returns the natural ordering of two objects.
	 */
	@SuppressWarnings("rawtypes")
	public static final Comparator<Comparable> NATURAL_ORDERING_COMPARATOR = new Comparator<Comparable>() {
		@SuppressWarnings("unchecked")
		@Override
		public int compare(Comparable o1, Comparable o2) {
			return o1.compareTo(o2);
		}
	};

	/**
	 * A comparator which returns 0 if the status of each
	 * TimelineEventWithStatus object matches, and another number otherwise.
	 */
	public static final Comparator<TimelineEventWithStatus> STATUS_COMPARATOR = new Comparator<TimelineEventWithStatus>() {
		@Override
		public int compare(TimelineEventWithStatus o1, TimelineEventWithStatus o2) {
			return o1.getStatus().equals(o2.getStatus()) ? 0 : -1;
		}
	};

	/**
	 * A Predicate which only accepts TimelineEventWithStatus items if their
	 * status is ACTIVE.
	 */
	public static final Predicate<TimelineEventWithStatus> ACTIVE_TIMELINE_EVENT_WITH_STATUS_PREDICATE = new Predicate<TimelineEventWithStatus>() {
		@Override
		public boolean apply(TimelineEventWithStatus e) {
			return e.getStatus().isActive();
		}
	};

	/**
	 * A Predicate which only accepts TimelineEventWithStatus items if their
	 * status is INACTIVE.
	 */
	public static final Predicate<TimelineEventWithStatus> INACTIVE_TIMELINE_EVENT_WITH_STATUS_PREDICATE = new Predicate<TimelineEventWithStatus>() {
		@Override
		public boolean apply(TimelineEventWithStatus e) {
			return !e.getStatus().isActive();
		}
	};

	public static void main(String[] args) {
		ZonedDateTime s1 = ZonedDateTime.of(2000, 1, 1, 0, 0, 0, 0, ZoneOffset.UTC);
		ZonedDateTime s2 = ZonedDateTime.of(2000, 2, 1, 0, 0, 0, 0, ZoneOffset.UTC);
		ZonedDateTime s3 = ZonedDateTime.of(2000, 3, 1, 0, 0, 0, 0, ZoneOffset.UTC);
		ZonedDateTime s4 = ZonedDateTime.of(2000, 4, 1, 0, 0, 0, 0, ZoneOffset.UTC);

		Object[][] tests = new Object[][] {
				// six normal tests
				{ s1, s2, s3, s4, false }, //
				{ s1, s3, s2, s4, true }, //
				{ s2, s3, s1, s4, true }, //
				{ s1, s4, s2, s3, true }, //
				{ s2, s4, s1, s3, true }, //
				{ s3, s4, s1, s2, false }, //

				// null tests
				{ s1, s2, s3, null, false }, //
				{ s1, s3, s2, null, true }, //
				{ s3, s4, s2, null, true }, //

				{ s1, s2, null, s3, true }, //
				{ s1, s3, null, s2, true }, //
				{ s2, s3, null, s1, false }, //

				{ s1, s2, null, null, true }, //

				{ s1, null, s2, s3, true }, //
				{ s2, null, s1, s3, true }, //
				{ s3, null, s1, s2, false }, //

				{ s1, null, s2, null, true }, //
				{ s2, null, s1, null, true }, //

				{ s1, null, null, s2, true }, //
				{ s2, null, null, s1, false }, //

				{ s1, null, null, null, true }, //

				{ null, s1, s2, s3, false }, //
				{ null, s2, s1, s3, true }, //
				{ null, s3, s1, s2, true }, //

				{ null, s1, s2, null, false }, //
				{ null, s2, s1, null, true }, //

				{ null, s1, null, s2, true }, //
				{ null, s2, null, s1, true }, //

				{ null, s1, null, null, true }, //

				{ null, null, null, s1, true }, //

				{ null, null, null, null, true }, //
		};

		for (Object[] test : tests) {
			boolean real = intervalsOverlap((ZonedDateTime) test[0], (ZonedDateTime) test[1], (ZonedDateTime) test[2],
					(ZonedDateTime) test[3]);
			boolean expected = (Boolean) test[4];
			if (real != expected)
				System.out.println(
						"failed: " + test[0] + ", " + test[1] + ", " + test[2] + ", " + test[3] + ", " + test[4]);
		}

		// --------------------- overlap tests

		TimelineInterval overlap = getOverlap(new SimpleTimelineInterval(s1, s3), new SimpleTimelineInterval(s2, s4));
		System.out.println(overlap);

		overlap = getOverlap(new SimpleTimelineInterval(s1, s2), new SimpleTimelineInterval(s3, s4));
		System.out.println(overlap);

		overlap = getOverlap(new SimpleTimelineInterval(s1, s3), new SimpleTimelineInterval(s2, null));
		System.out.println(overlap);
	}

}
