package gov.va.vss.model;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import java.time.Month;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import javax.persistence.Column;
import javax.persistence.Convert;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Transient;
import javax.validation.constraints.Max;
import javax.validation.constraints.Min;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Pattern;

import org.apache.commons.lang3.builder.CompareToBuilder;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.hibernate.annotations.BatchSize;
import org.hibernate.validator.constraints.Email;
import org.hibernate.validator.constraints.Length;
import org.hibernate.validator.constraints.NotBlank;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.core.AbstractAuditedVersionedPersistent;
import gov.va.vss.model.lookup.NACStatus;
import gov.va.vss.model.lookup.OrganizationType;
import gov.va.vss.model.lookup.VtkInstitution;
import gov.va.vss.model.lookup.VtkState;
import gov.va.vss.persistence.conversion.ScopeTypeConverter;
import gov.va.vss.util.ValidationUtil;

@Entity
@Table(name = "VssOrganizations", schema = "dbo")
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public class Organization extends AbstractAuditedVersionedPersistent<Organization>
		implements Comparable<Organization>, ObjectScopedToStationNumbers {
	private static final long serialVersionUID = 6904844123870655771L;

	/**
	 * See @JsonView documentation - CPB
	 */
	public static class OrganizationView {
		public interface Basic {
		}

		public interface Search extends Basic {
		}

		public interface Extended extends Search {
		}
	}

	// -------------------------------------- Fields

	@NotNull
	private ScopeType scope;

	private int code;
	@NotBlank(message = "Display Name is required.")
	@Length(max = 35)
	private String name;
	@NotBlank(message = "Full Name is required.")
	@Length(max = 250)
	private String fullName;
	@Length(max = 6)
	private String abbreviation;
	private boolean inactive;
	private VtkInstitution vtkInstitution;
	private boolean onNationalAdvisoryCommittee;
	private boolean nacExecutiveMember;
	private NACStatus nacMembershipStatus;
	private OrganizationType type;
	@Length(max = 35)
	private String addressLine1;
	@Length(max = 35)
	private String addressLine2;
	@Length(max = 30)
	private String city;
	private VtkState vtkState;
	@Length(max = 10)
	private String zip;
	@Length(max = 250)
	@Email(message = "Please enter a valid email in the format 'user@domain.tld'.")
	private String email;
	@Length(max = 30)
	@Pattern(regexp = ValidationUtil.PHONE_REGEX, message = "Please enter a valid phone number.")
	private String phone;

	@Min(1)
	@Max(12)
	private Integer annualJointReviewMonth;

	private List<VolunteerOrganization> volunteerOrganizations;
	private List<OrganizationBranch> branches;

	private List<WorkEntry> workEntries;
	
	// -------------------------------------- Business Methods

	@Transient
	@JsonIgnore
	public Collection<String> getScopedToStationNumbers() {
		VtkInstitution institution = getVtkInstitution();
		if (institution == null)
			return new ArrayList<>();
		return Arrays.asList(institution.getStationNumber());
	}

	@Transient
	@JsonIgnore
	public String getAJRMonthName() {
		if (this.annualJointReviewMonth != null) {
			return Month.of(annualJointReviewMonth.intValue()).name();
		} else {
			return "";
		}
	}

	@Transient
	@JsonView(OrganizationView.Extended.class)
	public boolean isActive() {
		return !inactive;
	}

	@Transient
	@JsonView(OrganizationView.Basic.class)
	public String getDisplayName() {
		return getName();
	}

	// -------------------------------------- Common Methods

	@Override
	protected boolean requiredEquals(Organization oo) {
		return new EqualsBuilder().append(code, oo.getCode())
				.append(nullSafeGetId(vtkInstitution), nullSafeGetId(oo.getVtkInstitution())).isEquals();
	}

	@Override
	protected int requiredHashCode() {
		return new HashCodeBuilder().append(code).append(nullSafeGetId(vtkInstitution)).toHashCode();
	}

	@Override
	public int compareTo(Organization o) {
		if (equals(o))
			return 0;

		return new CompareToBuilder().append(nullSafeLowercase(name), nullSafeLowercase(o.getName()))
				.append(code, o.getCode()).append(nullSafeGetId(vtkInstitution), nullSafeGetId(o.getVtkInstitution()))
				.toComparison() > 0 ? 1 : -1;
	}

	public String toString() {
		return getName() + " (code " + getCode() + ", station ID " + nullSafeGetId(vtkInstitution) + ")";
	}

	// -------------------------------------- Accessor Methods

	@Column(name = "CodeNUM", length = 5)
	@JsonView(OrganizationView.Search.class)
	public int getCode() {
		return code;
	}

	public void setCode(int code) {
		this.code = code;
	}

	@Column(length = 6)
	@JsonView(OrganizationView.Search.class)
	public String getAbbreviation() {
		return abbreviation;
	}

	public void setAbbreviation(String abbreviation) {
		this.abbreviation = abbreviation;
	}

	@Column(name = "OrganizationName", length = 35, nullable = false)
	@JsonView(OrganizationView.Search.class)
	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	@Column(name = "FullName", length = 250, nullable = false)
	@JsonView(OrganizationView.Extended.class)
	public String getFullName() {
		return fullName;
	}

	public void setFullName(String fullName) {
		this.fullName = fullName;
	}

	@Column(name = "IsInactive", nullable = false)
	@JsonView(OrganizationView.Extended.class)
	public boolean isInactive() {
		return inactive;
	}

	public void setInactive(boolean inactive) {
		this.inactive = inactive;
	}

	@Column(nullable = false)
	@Convert(converter = ScopeTypeConverter.class)
	@JsonView(OrganizationView.Extended.class)
	public ScopeType getScope() {
		return scope;
	}

	public void setScope(ScopeType scopeType) {
		this.scope = scopeType;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "VSS_STD_VolunteerOrganizationTypesFK")
	@JsonView(OrganizationView.Extended.class)
	public OrganizationType getType() {
		return type;
	}

	public void setType(OrganizationType type) {
		this.type = type;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "StationFK")
	@JsonView(OrganizationView.Search.class)
	public VtkInstitution getVtkInstitution() {
		return vtkInstitution;
	}

	public void setVtkInstitution(VtkInstitution station) {
		this.vtkInstitution = station;
	}

	@Column(name = "IsOnNationalAdvisoryCommittee", nullable = false)
	@JsonView(OrganizationView.Extended.class)
	public boolean isOnNationalAdvisoryCommittee() {
		return onNationalAdvisoryCommittee;
	}

	public void setOnNationalAdvisoryCommittee(boolean onNationalAdvisoryCommittee) {
		this.onNationalAdvisoryCommittee = onNationalAdvisoryCommittee;
	}

	@Column(name = "IsNACExecutiveMember", nullable = false)
	@JsonView(OrganizationView.Extended.class)
	public boolean isNacExecutiveMember() {
		return nacExecutiveMember;
	}

	public void setNacExecutiveMember(boolean nacExecutiveMember) {
		this.nacExecutiveMember = nacExecutiveMember;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "NACMembershipStatusFK")
	@JsonView(OrganizationView.Extended.class)
	public NACStatus getNacMembershipStatus() {
		return nacMembershipStatus;
	}

	public void setNacMembershipStatus(NACStatus nacMembershipStatus) {
		this.nacMembershipStatus = nacMembershipStatus;
	}

	@JsonView(OrganizationView.Extended.class)
	public Integer getAnnualJointReviewMonth() {
		return annualJointReviewMonth;
	}

	public void setAnnualJointReviewMonth(Integer annualJointReviewMonth) {
		this.annualJointReviewMonth = annualJointReviewMonth;
	}

	@OneToMany(mappedBy = "organization", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	public List<VolunteerOrganization> getVolunteerOrganizations() {
		if (volunteerOrganizations == null)
			volunteerOrganizations = new ArrayList<>();
		return volunteerOrganizations;
	}

	public void setVolunteerOrganizations(List<VolunteerOrganization> volunteerOrganizations) {
		this.volunteerOrganizations = volunteerOrganizations;
	}

	@OneToMany(mappedBy = "organization", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	public List<OrganizationBranch> getBranches() {
		if (branches == null)
			branches = new ArrayList<>();
		return branches;
	}

	public void setBranches(List<OrganizationBranch> branches) {
		this.branches = branches;
	}

	@Column(name = "Address1", length = 35)
	@JsonView(OrganizationView.Extended.class)
	public String getAddressLine1() {
		return addressLine1;
	}

	public void setAddressLine1(String address1) {
		this.addressLine1 = address1;
	}

	@Column(name = "Address2", length = 35)
	@JsonView(OrganizationView.Extended.class)
	public String getAddressLine2() {
		return addressLine2;
	}

	public void setAddressLine2(String address2) {
		this.addressLine2 = address2;
	}

	@Column(length = 30)
	@JsonView(OrganizationView.Extended.class)
	public String getCity() {
		return city;
	}

	public void setCity(String city) {
		this.city = city;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "STD_StateFK", unique = true)
	@JsonView(OrganizationView.Extended.class)
	public VtkState getVtkState() {
		return vtkState;
	}

	public void setVtkState(VtkState vtkState) {
		this.vtkState = vtkState;
	}

	@Column(length = 10)
	@JsonView(OrganizationView.Extended.class)
	public String getZip() {
		return zip;
	}

	public void setZip(String zip) {
		this.zip = zip;
	}

	@Column(length = 250)
	@JsonView(OrganizationView.Extended.class)
	public String getEmail() {
		return email;
	}

	public void setEmail(String email) {
		this.email = email;
	}

	@Column(length = 30)
	@JsonView(OrganizationView.Extended.class)
	public String getPhone() {
		return phone;
	}

	public void setPhone(String phone) {
		this.phone = phone;
	}
	
	@OneToMany(mappedBy = "organization", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	public List<WorkEntry> getWorkEntries() {
		if (workEntries == null)
			workEntries = new ArrayList<>();
		return workEntries;
	}

	public void setWorkEntries(List<WorkEntry> workEntries) {
		this.workEntries = workEntries;
	}

}
