package gov.va.vss.model.lookup;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import javax.persistence.Column;
import javax.persistence.FetchType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.MappedSuperclass;
import javax.persistence.Transient;
import javax.validation.constraints.NotNull;

import org.apache.commons.lang3.builder.CompareToBuilder;
import org.hibernate.annotations.BatchSize;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.core.AbstractAuditedVersionedPersistent;
import gov.va.shared.model.lookup.sds.VAFacility;

@MappedSuperclass
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public abstract class AbstractBasicInstitution<T extends AbstractBasicInstitution<T>>
		extends AbstractAuditedVersionedPersistent<T> implements Comparable<T> {
	private static final long serialVersionUID = 3810180355498406255L;

	/**
	 * See @JsonView documentation - CPB
	 */
	public static class VtkInstitutionView {
		public interface Basic {
		}

		public interface Extended extends Basic {
		}
	}

	// ---------------------------------------- Fields

	@NotNull
	private String name;
	@NotNull
	private String stationNumber;
	private VtkInstitution parent;
	private VtkAdministrativeUnit administrativeParent;
	private String addressLine1;
	private String addressLine2;
	private String city;
	@NotNull
	private VtkState state;
	private String zip;
	private VtkInstitutionType type;
	private Boolean mealPrintOCX;

	private VAFacility vaFacility;

	// ---------------------------------------- Constructors

	protected AbstractBasicInstitution() {
	}

	protected AbstractBasicInstitution(VAFacility facility, VtkState state, VtkInstitutionType type,
			VtkInstitution parent) {
		this.name = facility.getName();
		this.stationNumber = facility.getStationNumber();
		this.parent = parent;
		if (parent != null)
			this.administrativeParent = parent.getAdministrativeParent();
		this.addressLine1 = facility.getAddressLine1();
		this.addressLine2 = facility.getAddressLine2();
		this.city = facility.getCity();
		this.state = state;
		this.zip = facility.getZip();
		this.type = type;
		this.vaFacility = facility;
	}

	// ---------------------------------------- Business Methods

	@Transient
	public String getDisplayName() {
		return getVaFacility() != null ? getVaFacility().getDisplayName() : getName() + " (" + getStationNumber() + ")";
	}

	// ---------------------------------------- Common Methods

	@Override
	public int compareTo(T o) {
		if (equals(o))
			return 0;

		return new CompareToBuilder().append(getDisplayName(), o.getDisplayName()).toComparison() > 0 ? 1 : -1;
	}

	// ---------------------------------------- Accessor Methods

	@Column(name = "NameOfInstitution", length = 80, nullable = false)
	@JsonView(VtkInstitutionView.Extended.class)
	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	@Column(length = 7, nullable = false)
	@JsonView(VtkInstitutionView.Basic.class)
	public String getStationNumber() {
		return stationNumber;
	}

	public void setStationNumber(String stationNumber) {
		this.stationNumber = stationNumber;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "ParentInstitutionFK")
	@BatchSize(size = 500)
	@JsonIgnore
	public VtkInstitution getParent() {
		return parent;
	}

	public void setParent(VtkInstitution parent) {
		this.parent = parent;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "AdministrativeParentFK")
	@BatchSize(size = 500)
	@JsonIgnore
	public VtkAdministrativeUnit getAdministrativeParent() {
		return administrativeParent;
	}

	public void setAdministrativeParent(VtkAdministrativeUnit administrativeParent) {
		this.administrativeParent = administrativeParent;
	}

	@Column(name = "StreetAddress1", length = 35)
	@JsonView(VtkInstitutionView.Extended.class)
	public String getAddressLine1() {
		return addressLine1;
	}

	public void setAddressLine1(String addressLine1) {
		this.addressLine1 = addressLine1;
	}

	@Column(name = "StreetAddress2", length = 35)
	@JsonView(VtkInstitutionView.Extended.class)
	public String getAddressLine2() {
		return addressLine2;
	}

	public void setAddressLine2(String addressLine2) {
		this.addressLine2 = addressLine2;
	}

	@Column(length = 30)
	@JsonView(VtkInstitutionView.Extended.class)
	public String getCity() {
		return city;
	}

	public void setCity(String city) {
		this.city = city;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "StateFK", nullable = false)
	@BatchSize(size = 500)
	@JsonView(VtkInstitutionView.Extended.class)
	public VtkState getState() {
		return state;
	}

	public void setState(VtkState state) {
		this.state = state;
	}

	@Column(name = "ZipCode", length = 10)
	@JsonView(VtkInstitutionView.Extended.class)
	public String getZip() {
		return zip;
	}

	public void setZip(String zip) {
		this.zip = zip;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "TypeFK", nullable = false)
	@BatchSize(size = 500)
	@JsonView(VtkInstitutionView.Extended.class)
	public VtkInstitutionType getType() {
		return type;
	}

	public void setType(VtkInstitutionType type) {
		this.type = type;
	}

	@JsonIgnore
	public Boolean getMealPrintOCX() {
		return mealPrintOCX;
	}

	public void setMealPrintOCX(Boolean mealPrintOCX) {
		this.mealPrintOCX = mealPrintOCX;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "SDS_INSTITUTION_ID")
	@BatchSize(size = 500)
	@JsonIgnore
	public VAFacility getVaFacility() {
		return vaFacility;
	}

	public void setVaFacility(VAFacility facility) {
		this.vaFacility = facility;
	}
}
