package gov.va.vss.util;

import java.time.LocalDate;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Date;

import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import gov.va.shared.model.AppUser;

/**
 * Contains utility methods for manipulating dates as well as helpful formatting
 * constants. All "AM" / "PM" designations are in uppercase by convention. In
 * addition to static methods, this class is also a stateful component which can
 * be injected for processing things like Fiscal year calculations, which depend
 * on properties file entries.
 * 
 * @author             
 * 
 */
@Component
public final class DateUtil {

	@Value("${donationFiscalYearStartMMDD}")
	private String fiscalYearStartMMDD;
	@Value("${donationFiscalYearEndMMDD}")
	private String fiscalYearEndMMDD;
	@Value("${donationGracePeriod}")
	private int gracePeriod;
	
	public LocalDate getCurrentFiscalYearStartDate() {
		return getFiscalYearStartDateForDate(LocalDate.now());
	}

	public LocalDate getFiscalYearStartDateForDate(LocalDate d) {
		String[] tokens = fiscalYearStartMMDD.split("/");
		int month = Integer.parseInt(tokens[0]);
		int day = Integer.parseInt(tokens[1]);

		LocalDate fiscalYearStartDate = d.withMonth(month).withDayOfMonth(day).plusDays(gracePeriod);
		if (d.isBefore(fiscalYearStartDate))
			fiscalYearStartDate = fiscalYearStartDate.minusYears(1);

		return fiscalYearStartDate;
	}
	
	public LocalDate getCurrentFiscalYearEndDate() {
		return getFiscalYearEndDateForDate(LocalDate.now());
	}

	public LocalDate getFiscalYearEndDateForDate(LocalDate d) {
		String[] tokens = fiscalYearEndMMDD.split("/");
		int month = Integer.parseInt(tokens[0]);
		int day = Integer.parseInt(tokens[1]);

		LocalDate fiscalYearEndDate = d.withMonth(month).withDayOfMonth(day);
		if (d.isAfter(fiscalYearEndDate))
			fiscalYearEndDate = fiscalYearEndDate.plusYears(1);

		return fiscalYearEndDate;
	}
	
	public static final String DATE_ONLY = "M/d/yyyy";
	public static final DateTimeFormatter DATE_ONLY_FORMAT = DateTimeFormatter.ofPattern(DATE_ONLY);

	public static final String TWO_DIGIT_DATE_ONLY = "MM/dd/yyyy";
	public static final DateTimeFormatter TWO_DIGIT_DATE_ONLY_FORMAT = DateTimeFormatter.ofPattern(TWO_DIGIT_DATE_ONLY);

	public static final String MILITARY_DATE_TIME = "M/d/yyyy HH:mm";
	public static final DateTimeFormatter MILITARY_DATE_TIME_FORMAT = DateTimeFormatter.ofPattern(MILITARY_DATE_TIME);

	public static final String MILITARY_DATE_TIME_WITH_ZONE = MILITARY_DATE_TIME + " z";
	public static final DateTimeFormatter MILITARY_DATE_TIME_WITH_ZONE_FORMAT = DateTimeFormatter
			.ofPattern(MILITARY_DATE_TIME_WITH_ZONE);

	public static final String TWO_DIGIT_MILITARY_DATE_TIME = "MM/dd/yyyy HH:mm";
	public static final DateTimeFormatter TWO_DIGIT_MILITARY_DATE_TIME_FORMAT = DateTimeFormatter
			.ofPattern(TWO_DIGIT_MILITARY_DATE_TIME);

	/* Useful when sorting */
	public static final String MSD_TO_LSD = "yyyyMMddHHmm";
	public static final DateTimeFormatter MSD_TO_LSD_FORMAT = DateTimeFormatter.ofPattern(MSD_TO_LSD);

	public static final String TWO_DIGIT_DATE_MASK = "99/99/9999";
	public static final String TWO_DIGIT_DATE_TIME_MASK = "99/99/9999 99:99";

	public static final ZoneId UTC = ZoneId.of("Z");

	public static ZonedDateTime currentMinute() {
		return ZonedDateTime.now().withSecond(0).withNano(0);
	}

	public static String localFormat(AppUser user, ZonedDateTime date) {
		return localFormat(user.getTimeZone(), date);
	}

	public static String localFormat(ZoneId userZone, ZonedDateTime date) {
		return date == null ? "(unspecified)"
				: date.withZoneSameInstant(userZone).format(DateTimeFormatter.ofPattern(MILITARY_DATE_TIME_WITH_ZONE));
	}

	public static String getDateRangeDescription(ZoneId userZone, ZonedDateTime startDate, ZonedDateTime endDate) {
		if (startDate != null && endDate != null) {
			return "between " + localFormat(userZone, startDate) + " and " + localFormat(userZone, endDate);
		} else if (startDate != null) {
			return "from " + localFormat(userZone, startDate) + " onward";
		} else {
			return "before " + localFormat(userZone, endDate);
		}
	}

	/**
	 * Helper method to reduce boilerplate code - JDK runs in UTC so this will
	 * initially be set to UTC zone.
	 * 
	 * @param d
	 * @return
	 */
	public static ZonedDateTime getZonedDateTime(Date d) {
		if (d == null)
			return null;
		return ZonedDateTime.ofInstant(d.toInstant(), ZoneId.systemDefault());
	}

	public static LocalDate getLocalDate(Date d) {
		if (d == null)
			return null;
		return d.toInstant().atZone(ZoneId.systemDefault()).toLocalDate();
	}

}
