package gov.va.vss.web.donation;

import java.time.LocalDate;
import java.util.Arrays;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.servlet.http.HttpServletRequest;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataIntegrityViolationException;
import org.springframework.format.annotation.DateTimeFormat;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.ModelMap;
import org.springframework.validation.BindingResult;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.SessionAttributes;
import org.springframework.web.bind.support.SessionStatus;

import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.Permission;
import gov.va.shared.model.Permission.PermissionType;
import gov.va.shared.model.lookup.sds.State;
import gov.va.shared.util.StringUtil;
import gov.va.vss.model.Donor;
import gov.va.vss.model.Donor.DonorView;
import gov.va.vss.model.Volunteer;
import gov.va.vss.model.Volunteer.VolunteerView;
import gov.va.vss.model.lookup.VtkState;
import gov.va.vss.service.validation.ServiceValidationException;
import gov.va.vss.util.DateUtil;
import gov.va.vss.web.AbstractAppController;
import gov.va.vss.web.ValidationException;
import gov.va.vss.web.breadcrumbs.Breadcrumb;
import gov.va.vss.web.volunteer.VolunteerCommand;

@Controller
@SessionAttributes(value = { AbstractAppController.DEFAULT_COMMAND_NAME })
public class DonorController extends AbstractAppController {

	@Autowired
	private DonorValidator donorValidator;

	@RequestMapping("/donorCreate.htm")
	@Breadcrumb("Create Donor")
	@PreAuthorize("hasAuthority('" + Permission.DONATION_CREATE + "')")
	
	
	public String createDonor(ModelMap model, HttpServletRequest request) {
		Donor donor = new Donor();

		DonorCommand command = new DonorCommand(donor, null);

		model.addAttribute(DEFAULT_COMMAND_NAME, command);

		createReferenceData(model);

		return "createDonor";
	}

	@RequestMapping("/donorEdit.htm")
	@Breadcrumb("Edit Donor")
	@PreAuthorize("hasAuthority('" + Permission.DONATION_READ + "')")
	public String editDonor(@RequestParam long id, ModelMap model, HttpServletRequest request) {
		Donor donor = donorDAO.findRequiredByPrimaryKey(id);
		
		VtkState vtkState = donor.getVtkState();
		DonorCommand command = new DonorCommand(donor,
				vtkState == null ? null : stateDAO.findStateByPostalCode(vtkState.getAbbreviation()));		
		
		model.addAttribute(DEFAULT_COMMAND_NAME, command);
		createReferenceData(model);
		updateModelForPermissionsIfNecessary(model, PermissionType.DONATION_CREATE);
		
		return "editDonor";
	}

	private void createReferenceData(ModelMap model) {
		model.put("allStates", stateDAO.findListOfStatesInUSA());
	}

	@RequestMapping("/donorSubmit.htm")
	@PreAuthorize("hasAuthority('" + Permission.DONATION_CREATE + "')")
	public String submitDonor(@ModelAttribute(DEFAULT_COMMAND_NAME) DonorCommand command,
			BindingResult result, SessionStatus status, ModelMap model, HttpServletRequest request)
					throws ValidationException {
		Donor donor = command.getDonor();

		boolean isEdit = donor.isPersistent();

		/* Validation step (JSR303, other custom logic in the validator) */
		donorValidator.validate(command, result, false, "donor");
		boolean hasErrors = result.hasErrors();

		if (!hasErrors) {
			try {
				State state = command.getState();
				donor.setVtkState(state != null ? vtkService.getOrCreateForSDSState(state.getId()) : null);

				donor = donorService.saveOrUpdate(donor, isEdit);
				userNotifier.notifyUserOnceWithMessage(request,
						getMessage(isEdit ? "donor.update.success" : "donor.create.success"));
			} catch (ServiceValidationException e) {
				webValidationService.handle(e, result);
				hasErrors = true;
			}
		}

		if (hasErrors) {
			createReferenceData(model);
			return isEdit ? "editDonor" : "createDonor";
		} else {
			status.setComplete();
			return "redirect:/donorEdit.htm?id=" + donor.getId();
		}
	}
	
	@RequestMapping("/findDonors")
	@PreAuthorize("hasAuthority('" + Permission.DONATION_READ + "')")
	@JsonView(DonorView.Basic.class)
	public @ResponseBody SortedSet<Donor> findDonorss(@RequestParam(required = false) String name,
			 @RequestParam(required = false) String email) {
		if (StringUtil.allBlank(name, email))
			throw new IllegalArgumentException("Please specify at least one piece of search criteria");

		String firstName = null;
		String lastName = null;
		String middleName = null;
		if (StringUtils.isNotBlank(name)) {
			String[] nameComponents = StringUtil.parseNameComponents(name);
			lastName = nameComponents[0];
			firstName = nameComponents[1];
			middleName = nameComponents[2];
		}

		SortedSet<Donor> donors = new TreeSet<>(donorDAO.findByCriteria(firstName, middleName, lastName, email));

		/*
		 * Populate a map if tertiary information (outside of the Volunteer
		 * object itself) needs to be included in the response; otherwise, just
		 * return the Collection<Volunteer> above directly and Jackson will
		 * serialize it into JSON (per the annotations in the Volunteer class).
		 * CPB
		 */
		// List<Map<String, Object>> results = new ArrayList<>();
		//
		// for (Volunteer volunteer : volunteers) {
		// Map<String, Object> result = new HashMap<>();
		// result.put("volunteerId", volunteer.getId());
		// result.put("displayName", volunteer.getDisplayName());
		// result.put("contactInfo",
		// volunteer.getStationProfileForSta3n(sta3n).getContactInfo());
		// results.add(result);
		// }
		//
		// return results;

		return donors;
	}

	//@RequestMapping("/volunteerDuplicateCheck")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	/*@JsonView(VolunteerView.Search.class)
	public @ResponseBody SortedSet<Volunteer> findDuplicateVolunteers(@RequestParam String firstName,
			@RequestParam String lastName,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam LocalDate dob) {
		if (StringUtils.isBlank(lastName) || StringUtils.isBlank(firstName))
			throw new IllegalArgumentException("Both last name and first name are required.");

		SortedSet<Volunteer> volunteers = new TreeSet<>();
		List<Volunteer> firstNameMatches = volunteerDAO.findByCriteria(firstName, null, null, null, dob, null, null,
				null, null, null, null, null, null, null, null);
		volunteers.addAll(firstNameMatches);

		List<Volunteer> lastNameMatches = volunteerDAO.findByCriteria(null, null, lastName, null, dob, null, null, null,
				null, null, null, null, null, null, null);
		volunteers.addAll(lastNameMatches);

		return volunteers;
	}*/

}
