package gov.va.vss.web.volunteer;

import static gov.va.shared.model.Permission.PermissionType.VOLUNTEER_CREATE;
import static gov.va.shared.util.SecurityUtil.hasAllPermissionsAtCurrentFacility;

import java.time.LocalDate;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.servlet.http.HttpServletRequest;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.format.annotation.DateTimeFormat;
import org.springframework.stereotype.Controller;
import org.springframework.ui.ModelMap;
import org.springframework.validation.BindingResult;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.SessionAttributes;
import org.springframework.web.bind.support.SessionStatus;

import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.lookup.sds.State;
import gov.va.shared.util.StringUtil;
import gov.va.vss.model.ParkingSticker;
import gov.va.vss.model.ParkingSticker.ParkingStickerView;
import gov.va.vss.model.Uniform;
import gov.va.vss.model.Volunteer;
import gov.va.vss.model.Volunteer.VolunteerView;
import gov.va.vss.model.VolunteerFacilityAssignment;
import gov.va.vss.model.VolunteerFacilityAssignment.CompareByFacilityAssignment;
import gov.va.vss.model.VolunteerFacilityAssignment.VolunteerFacilityAssignmentView.SearchForVolunteerFacilityAssignments;
import gov.va.vss.model.VolunteerOrganization;
import gov.va.vss.model.VolunteerOrganization.CompareByOrganization;
import gov.va.vss.model.VolunteerOrganization.VolunteerOrganizationView;
import gov.va.vss.model.lookup.VtkInstitution;
import gov.va.vss.model.lookup.VtkState;
import gov.va.vss.persistence.dao.VolunteerDAO.TimeSummary;
import gov.va.vss.persistence.queryCustomization.QueryCustomization;
import gov.va.vss.persistence.queryCustomization.fieldTypes.VolunteerFacilityAssignmentFieldType;
import gov.va.vss.service.validation.ServiceValidationException;
import gov.va.vss.util.DateUtil;
import gov.va.vss.web.AbstractAppController;
import gov.va.vss.web.ValidationException;
import gov.va.vss.web.breadcrumbs.Breadcrumb;

@Controller
@SessionAttributes(value = { AbstractAppController.DEFAULT_COMMAND_NAME })
public class VolunteerController extends AbstractAppController {

	@Autowired
	private VolunteerValidator volunteerValidator;

	@RequestMapping("/getVolunteerSearchStations")
	public @ResponseBody SortedSet<VtkInstitution> getStationsForVolunteerSearch() {
		// Set<VAFacility> validFacilities =
		// SecurityUtil.getAllFacilitiesWhereUserHasPermissions(VOLUNTEER_READ);
		// return new TreeSet<>(validFacilities);
		return vtkInstitutionDAO.findAllSorted(); // findVAFacilitiesWithLinkToVtkInstitution();
	}

	@RequestMapping("/volunteerOrganizations")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(VolunteerOrganizationView.SearchForOrganizations.class)
	public @ResponseBody Map<String, Object> findOrgsForVolunteer(@RequestParam long volunteerId) {
		Map<String, Object> results = new HashMap<>();
		Volunteer v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);

		SortedSet<VolunteerOrganization> r = new TreeSet<>(new CompareByOrganization());
		r.addAll(v.getVolunteerOrganizations());
		results.put("organizations", r);

		results.put("primaryOrganization", v.getPrimaryOrganization());
		return results;
	}

	@RequestMapping(value = "/volunteer/setPrimaryOrganization", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean processSetPrimaryOrganization(@RequestParam long volunteerId,
			@RequestParam long organizationId) {
		volunteerService.setPrimaryOrganization(volunteerId, organizationId);
		return true;
	}

	@RequestMapping(value = "/volunteer/setPrimaryFacility", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean processSetPrimaryFacility(@RequestParam long volunteerId,
			@RequestParam long facilityId) {
		volunteerService.setPrimaryFacility(volunteerId, facilityId);
		return true;
	}

	@RequestMapping("/volunteerAssignments")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(SearchForVolunteerFacilityAssignments.class)
	public @ResponseBody Map<String, Object> findAssignmentsForVolunteer(@RequestParam long volunteerId) {
		Map<String, Object> results = new HashMap<>();
		Volunteer v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);

		List<VolunteerFacilityAssignment> resultList = volunteerFacilityAssignmentDAO
				.findByCriteria(Arrays.asList(volunteerId), true,
						new QueryCustomization()
								.prefetchField(VolunteerFacilityAssignmentFieldType.FACILITY_ASSIGNMENT, "fa")
								.addExtraJoin(" left join fetch fa.benefitingService"));
		SortedSet<VolunteerFacilityAssignment> r = new TreeSet<>(new CompareByFacilityAssignment());
		r.addAll(resultList);
		results.put("assignments", r);

		results.put("primaryFacility", v.getPrimaryFacility());
		return results;
	}

	@RequestMapping("/volunteerParkingStickers")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(ParkingStickerView.Extended.class)
	public @ResponseBody SortedSet<ParkingSticker> findParkingStickersForVolunteer(@RequestParam long volunteerId) {
		Volunteer v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);
		SortedSet<ParkingSticker> r = new TreeSet<>();
		r.addAll(v.getParkingStickers());
		return r;
	}

	@RequestMapping("/volunteerUniforms")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(ParkingStickerView.Extended.class)
	public @ResponseBody SortedSet<Uniform> findUniformsForVolunteer(@RequestParam long volunteerId) {
		Volunteer v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);
		SortedSet<Uniform> r = new TreeSet<>();
		r.addAll(v.getUniforms());
		return r;
	}

	@RequestMapping("/findVolunteers")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(VolunteerView.Search.class)
	public @ResponseBody SortedSet<Volunteer> findVolunteers(@RequestParam(required = false) String name,
			@RequestParam(required = false) String code, @RequestParam(required = false) String email,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam(required = false) LocalDate dob,
			@RequestParam String scope, @RequestParam(required = false) Long facilityId,
			@RequestParam boolean includeInactive) {
		if (StringUtil.allBlank(name, code, email) && dob == null)
			throw new IllegalArgumentException("Please specify at least one piece of search criteria");

		if ("National".equals(scope))
			facilityId = null;

		String firstName = null;
		String lastName = null;
		String middleName = null;
		if (StringUtils.isNotBlank(name)) {
			String[] nameComponents = StringUtil.parseNameComponents(name);
			lastName = nameComponents[0];
			firstName = nameComponents[1];
			middleName = nameComponents[2];
		}

		SortedSet<Volunteer> volunteers = new TreeSet<>(volunteerDAO.findByCriteria(firstName, middleName, lastName,
				code, dob, null, null, null, null, null, email, includeInactive ? null : true, null, null,
				facilityId != null ? Arrays.asList(facilityId) : null));

		/*
		 * Populate a map if tertiary information (outside of the Volunteer
		 * object itself) needs to be included in the response; otherwise, just
		 * return the Collection<Volunteer> above directly and Jackson will
		 * serialize it into JSON (per the annotations in the Volunteer class).
		 * CPB
		 */
		// List<Map<String, Object>> results = new ArrayList<>();
		//
		// for (Volunteer volunteer : volunteers) {
		// Map<String, Object> result = new HashMap<>();
		// result.put("volunteerId", volunteer.getId());
		// result.put("displayName", volunteer.getDisplayName());
		// result.put("contactInfo",
		// volunteer.getStationProfileForSta3n(sta3n).getContactInfo());
		// results.add(result);
		// }
		//
		// return results;

		return volunteers;
	}

	@RequestMapping("/volunteerDuplicateCheck")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(VolunteerView.Search.class)
	public @ResponseBody SortedSet<Volunteer> findDuplicateVolunteers(@RequestParam String firstName,
			@RequestParam String lastName,
			@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam LocalDate dob) {
		if (StringUtils.isBlank(lastName) || StringUtils.isBlank(firstName))
			throw new IllegalArgumentException("Both last name and first name are required.");

		SortedSet<Volunteer> volunteers = new TreeSet<>();
		List<Volunteer> firstNameMatches = volunteerDAO.findByCriteria(firstName, null, null, null, dob, null, null,
				null, null, null, null, null, null, null, null);
		volunteers.addAll(firstNameMatches);

		List<Volunteer> lastNameMatches = volunteerDAO.findByCriteria(null, null, lastName, null, dob, null, null, null,
				null, null, null, null, null, null, null);
		volunteers.addAll(lastNameMatches);

		return volunteers;
	}

	@RequestMapping("/volunteerCreate.htm")
	// Don't want a Breadcrumb here since we want to force them to search first
	// every time - CPB
	// @Breadcrumb("Create Volunteer")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_EDIT + "')")
	public String createVolunteer(ModelMap model, HttpServletRequest request) {
		Volunteer volunteer = new Volunteer();
		volunteer.setMealsEligible(1);

		vtkService.getOrCreateInstitutionForSDSFacility(getSiteContext().getId());
		VolunteerCommand command = new VolunteerCommand(volunteer, null, null);
		model.addAttribute(DEFAULT_COMMAND_NAME, command);
		createReferenceData(model);
		return "createVolunteer";
	}

	@RequestMapping("/volunteerEdit.htm")
	@Breadcrumb("Edit Volunteer")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_EDIT + "')")
	public String editVolunteer(@RequestParam long id, ModelMap model, HttpServletRequest request) {
		Volunteer volunteer = volunteerDAO.findRequiredByPrimaryKey(id);

		TimeSummary summary = volunteerDAO.getTimeSummary(id);
		VtkState vtkState = volunteer.getState();
		VolunteerCommand command = new VolunteerCommand(volunteer, summary,
				vtkState == null ? null : stateDAO.findStateByPostalCode(vtkState.getAbbreviation()));
		model.addAttribute(DEFAULT_COMMAND_NAME, command);

		// TODO query DB once we figure out how to join - CPB
		model.addAttribute("backgroundStatus", "No Issue Reported");
		model.addAttribute("allVolunteerFacilities", volunteerDAO.findFacilitiesForVolunteer(id));
		model.addAttribute("yearsVolunteering", volunteerDAO.getYearsVolunteering(id));

		createReferenceData(model);

		setFormAsReadOnly(model, !hasAllPermissionsAtCurrentFacility(VOLUNTEER_CREATE));

		return "editVolunteer";
	}

	private void createReferenceData(ModelMap model) {
		model.put("allGenders", genderDAO.findAllSorted());
		model.put("allStates", stateDAO.findListOfStatesInUSA());
		model.put("allLanguages", vtkLanguageDAO.findAllSorted());
		model.put("allTransportationMethods", transportationMethodDAO.findAllSorted());
		model.put("allAwards", awardDAO.findAllSorted());
	}

	@RequestMapping("/volunteerSubmit.htm")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_EDIT + "')")
	public String submitVolunteer(@ModelAttribute(DEFAULT_COMMAND_NAME) VolunteerCommand command, BindingResult result,
			SessionStatus status, ModelMap model, HttpServletRequest request) throws ValidationException {
		Volunteer volunteer = command.getVolunteer();

		boolean isEdit = volunteer.isPersistent();

		/*
		 * Perform any remaining binding steps from Command obj -> Volunteer
		 * here
		 */

		State state = command.getState();
		volunteer.setState(state != null ? vtkService.getOrCreateForSDSState(state.getId()) : null);

		/* Validation step (JSR303, other custom logic in the validator) */
		volunteerValidator.validate(command, result, false, "volunteer");
		boolean hasErrors = result.hasErrors();

		if (!hasErrors) {
			try {
				volunteer = volunteerService.saveOrUpdate(volunteer);
				userNotifier.notifyUserOnceWithMessage(request,
						getMessage(isEdit ? "volunteer.update.success" : "volunteer.create.success"));
			} catch (ServiceValidationException e) {
				webValidationService.handle(e, result);
				hasErrors = true;
			}
		}

		if (hasErrors) {
			createReferenceData(model);
			return isEdit ? "editVolunteer" : "createVolunteer";
		} else {
			status.setComplete();
			return "redirect:/volunteerEdit.htm?id=" + volunteer.getId();
		}
	}

}
