package gov.va.vss.model;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.Column;
import javax.persistence.Convert;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Transient;
import javax.validation.constraints.NotNull;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.builder.CompareToBuilder;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.hibernate.annotations.BatchSize;
import org.hibernate.validator.constraints.Length;
import org.hibernate.validator.constraints.NotBlank;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.core.AbstractAuditedVersionedPersistent;
import gov.va.vss.model.facility.Facility;
import gov.va.vss.persistence.conversion.ScopeTypeConverter;

@Entity
@Table(name = "VssBenefitingServices", schema = "dbo")
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public class BenefitingService extends AbstractAuditedVersionedPersistent<BenefitingService>
		implements Comparable<BenefitingService> {
	private static final long serialVersionUID = 6904844123870655771L;

	/**
	 * See @JsonView documentation - CPB
	 */
	public static class BenefitingServiceView {
		public interface Basic {
		}

		public interface Extended extends Basic {
		}
	}

	// -------------------------------------- Fields

	@NotNull
	@Length(max = 4)
	private String code;
	@NotBlank
	@Length(max = 35)
	private String name;
	@NotBlank
	@Length(max = 30)
	private String subdivision;
	private String abbreviation;
	@NotNull
	private ScopeType scope;

	private Facility facility;

	private List<BenefitingServiceRole> benefitingServiceRoles;

	// -------------------------------------- Business Methods

	@Transient
	@JsonView(BenefitingServiceView.Basic.class)
	public String getDisplayName() {
		return getName() + (StringUtils.isBlank(subdivision) ? "" : " - " + subdivision) + " (" + code + ")";
	}

	// -------------------------------------- Common Methods

	@Override
	protected boolean requiredEquals(BenefitingService oo) {
		/*
		 * doubt all three of these are necessary but don't have data to confirm
		 * yet - CPB
		 */
		return new EqualsBuilder().append(getCode(), oo.getCode())
				.append(nullSafeGetId(getFacility()), nullSafeGetId(oo.getFacility())).isEquals();
	}

	@Override
	protected int requiredHashCode() {
		return new HashCodeBuilder().append(getCode()).append(nullSafeGetId(getFacility())).toHashCode();
	}

	@Override
	public int compareTo(BenefitingService o) {
		if (equals(o))
			return 0;

		return new CompareToBuilder().append(nullSafeLowercase(getDisplayName()), nullSafeLowercase(o.getDisplayName()))
				.toComparison() > 0 ? 1 : -1;
	}

	@Override
	public String toString() {
		return getDisplayName();
	}

	// -------------------------------------- Accessor Methods

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "FacilityFK", nullable = false)
	@JsonIgnore
	public Facility getFacility() {
		return facility;
	}

	public void setFacility(Facility facility) {
		this.facility = facility;
	}

	@Column(length = 4, nullable = false)
	@JsonView(BenefitingServiceView.Extended.class)
	public String getCode() {
		return code;
	}

	public void setCode(String code) {
		this.code = code;
	}

	@Column(name = "ServiceName", length = 35, nullable = false)
	@JsonView(BenefitingServiceView.Extended.class)
	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	@Column(length = 30, nullable = false)
	@JsonView(BenefitingServiceView.Extended.class)
	public String getSubdivision() {
		return subdivision;
	}

	public void setSubdivision(String subdivision) {
		this.subdivision = subdivision;
	}

	@Column(length = 7)
	@JsonView(BenefitingServiceView.Extended.class)
	public String getAbbreviation() {
		return abbreviation;
	}

	public void setAbbreviation(String abbreviation) {
		this.abbreviation = abbreviation;
	}

	@Convert(converter = ScopeTypeConverter.class)
	@JsonView(BenefitingServiceView.Extended.class)
	public ScopeType getScope() {
		return scope;
	}

	public void setScope(ScopeType scope) {
		this.scope = scope;
	}

	@OneToMany(mappedBy = "benefitingService", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	public List<BenefitingServiceRole> getBenefitingServiceRoles() {
		if (benefitingServiceRoles == null)
			benefitingServiceRoles = new ArrayList<>();
		return benefitingServiceRoles;
	}

	public void setBenefitingServiceRoles(List<BenefitingServiceRole> benefitingServiceRoles) {
		this.benefitingServiceRoles = benefitingServiceRoles;
	}

}
