package gov.va.vss.model.organization;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.Column;
import javax.persistence.Convert;
import javax.persistence.DiscriminatorColumn;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.Inheritance;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Transient;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Pattern;

import org.apache.commons.lang3.builder.CompareToBuilder;
import org.hibernate.annotations.BatchSize;
import org.hibernate.validator.constraints.Email;
import org.hibernate.validator.constraints.Length;
import org.hibernate.validator.constraints.NotBlank;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.core.AbstractAuditedVersionedPersistent;
import gov.va.shared.model.lookup.sds.State;
import gov.va.vss.model.ScopeType;
import gov.va.vss.model.VolunteerOrganization;
import gov.va.vss.model.WorkEntry;
import gov.va.vss.model.facility.Facility;
import gov.va.vss.persistence.conversion.ScopeTypeConverter;
import gov.va.vss.util.ValidationUtil;

@Entity
@Inheritance
@Table(name = "VssOrganizations", schema = "dbo")
@DiscriminatorColumn(name = "Type")
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public abstract class AbstractBasicOrganization extends AbstractAuditedVersionedPersistent<AbstractBasicOrganization>
		implements BasicOrganization {
	private static final long serialVersionUID = 3810180355498406255L;

	/**
	 * See @JsonView documentation - CPB
	 */
	public static class OrganizationView {
		public interface Basic {
		}

		public interface Search extends Basic {
		}

		public interface Extended extends Search {
		}
	}

	// ---------------------------------------- Fields

	@NotNull
	private ScopeType scope;
	private Facility facility;

	@NotBlank(message = "Display Name is required.")
	@Length(max = 35)
	protected String name;
	@NotBlank(message = "Full Name is required.")
	@Length(max = 250)
	protected String fullName;
	@Length(max = 6)
	protected String abbreviation;
	protected boolean inactive;

	@Length(max = 35)
	protected String addressLine1;
	@Length(max = 35)
	protected String addressLine2;
	@Length(max = 30)
	protected String city;
	protected State state;
	@Length(max = 10)
	protected String zip;
	@Length(max = 250)
	@Email(message = "Please enter a valid email in the format 'user@domain.tld'.")
	protected String email;
	@Length(max = 30)
	@Pattern(regexp = ValidationUtil.PHONE_REGEX, message = "Please enter a valid phone number.")
	protected String phone;

	protected List<VolunteerOrganization> volunteerOrganizations;
	protected List<WorkEntry> workEntries;

	// ---------------------------------------- Business Methods

	@Override
	@Transient
	@JsonView(OrganizationView.Extended.class)
	public boolean isActive() {
		return !inactive;
	}

	@Override
	@Transient
	@JsonView(OrganizationView.Basic.class)
	public abstract String getDisplayName();

	@Override
	@Transient
	public abstract String getScale();

	// ---------------------------------------- Common Methods

	@Override
	public int compareTo(BasicOrganization o) {
		if (equals(o))
			return 0;

		/*
		 * Sort by national first, then facility name, then main organization
		 * name, then branch name (if different) - CPB
		 */
		Facility facility = getRootOrganization().getFacility();
		String facilityName = facility == null ? "111111111" : facility.getDisplayName();
		Facility otherFacility = o.getRootOrganization().getFacility();
		String otherFacilityName = otherFacility == null ? "111111111" : otherFacility.getDisplayName();

		return new CompareToBuilder().append(facilityName, otherFacilityName)
				.append(getRootOrganization().getDisplayName(), o.getRootOrganization().getDisplayName())
				.append(getDisplayName(), o.getDisplayName()).toComparison() > 0 ? 1 : -1;
	}

	// ---------------------------------------- Accessor Methods

	@Override
	@Column(nullable = false)
	@Convert(converter = ScopeTypeConverter.class)
	@JsonView(OrganizationView.Extended.class)
	public ScopeType getScope() {
		return scope;
	}

	@Override
	public void setScope(ScopeType scopeType) {
		this.scope = scopeType;
	}

	@Override
	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "StationFK")
	@JsonView(OrganizationView.Search.class)
	public Facility getFacility() {
		return facility;
	}

	@Override
	public void setFacility(Facility facility) {
		this.facility = facility;
	}

	@Override
	@Column(length = 6)
	@JsonView(OrganizationView.Search.class)
	public String getAbbreviation() {
		return abbreviation;
	}

	@Override
	public void setAbbreviation(String abbreviation) {
		this.abbreviation = abbreviation;
	}

	@Override
	@Column(name = "OrganizationName", length = 35, nullable = false)
	@JsonView(OrganizationView.Search.class)
	public String getName() {
		return name;
	}

	@Override
	public void setName(String name) {
		this.name = name;
	}

	@Override
	@Column(name = "FullName", length = 250, nullable = false)
	@JsonView(OrganizationView.Extended.class)
	public String getFullName() {
		return fullName;
	}

	@Override
	public void setFullName(String fullName) {
		this.fullName = fullName;
	}

	@Override
	@Column(name = "IsInactive", nullable = false)
	@JsonView(OrganizationView.Extended.class)
	public boolean isInactive() {
		return inactive;
	}

	@Override
	public void setInactive(boolean inactive) {
		this.inactive = inactive;
	}

	@Override
	@OneToMany(mappedBy = "organization", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	public List<VolunteerOrganization> getVolunteerOrganizations() {
		if (volunteerOrganizations == null)
			volunteerOrganizations = new ArrayList<>();
		return volunteerOrganizations;
	}

	@Override
	public void setVolunteerOrganizations(List<VolunteerOrganization> volunteerOrganizations) {
		this.volunteerOrganizations = volunteerOrganizations;
	}

	@Override
	@Column(name = "Address1", length = 35)
	@JsonView(OrganizationView.Extended.class)
	public String getAddressLine1() {
		return addressLine1;
	}

	@Override
	public void setAddressLine1(String address1) {
		this.addressLine1 = address1;
	}

	@Override
	@Column(name = "Address2", length = 35)
	@JsonView(OrganizationView.Extended.class)
	public String getAddressLine2() {
		return addressLine2;
	}

	@Override
	public void setAddressLine2(String address2) {
		this.addressLine2 = address2;
	}

	@Override
	@Column(length = 30)
	@JsonView(OrganizationView.Extended.class)
	public String getCity() {
		return city;
	}

	@Override
	public void setCity(String city) {
		this.city = city;
	}

	@Override
	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "STD_StateFK", unique = true)
	@JsonView(OrganizationView.Extended.class)
	public State getState() {
		return state;
	}

	@Override
	public void setState(State state) {
		this.state = state;
	}

	@Override
	@Column(length = 10)
	@JsonView(OrganizationView.Extended.class)
	public String getZip() {
		return zip;
	}

	@Override
	public void setZip(String zip) {
		this.zip = zip;
	}

	@Override
	@Column(length = 250)
	@JsonView(OrganizationView.Extended.class)
	public String getEmail() {
		return email;
	}

	@Override
	public void setEmail(String email) {
		this.email = email;
	}

	@Override
	@Column(length = 30)
	@JsonView(OrganizationView.Extended.class)
	public String getPhone() {
		return phone;
	}

	@Override
	public void setPhone(String phone) {
		this.phone = phone;
	}

	@Override
	@OneToMany(mappedBy = "organization", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	public List<WorkEntry> getWorkEntries() {
		if (workEntries == null)
			workEntries = new ArrayList<>();
		return workEntries;
	}

	@Override
	public void setWorkEntries(List<WorkEntry> workEntries) {
		this.workEntries = workEntries;
	}

}
