package gov.va.vss.model.organization;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import java.time.Month;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import javax.persistence.Column;
import javax.persistence.DiscriminatorValue;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Transient;
import javax.validation.constraints.Max;
import javax.validation.constraints.Min;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.hibernate.annotations.BatchSize;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.util.PersistenceUtil;
import gov.va.vss.model.ObjectScopedToStationNumbers;
import gov.va.vss.model.facility.Facility;
import gov.va.vss.model.lookup.NACStatus;
import gov.va.vss.model.lookup.OrganizationType;

@Entity
@DiscriminatorValue("O")
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public class Organization extends AbstractBasicOrganization implements ObjectScopedToStationNumbers {
	private static final long serialVersionUID = 6904844123870655771L;

	// -------------------------------------- Fields

	private int code;

	private boolean onNationalAdvisoryCommittee;
	private boolean nacExecutiveMember;
	private NACStatus nacMembershipStatus;
	private OrganizationType type;

	@Min(1)
	@Max(12)
	private Integer annualJointReviewMonth;

	private List<OrganizationBranch> branches;

	// -------------------------------------- Business Methods

	@Override
	@Transient
	@JsonIgnore
	public Organization getRootOrganization() {
		return this;
	}

	@Transient
	@JsonIgnore
	public Collection<String> getScopedToStationNumbers() {
		Facility institution = getFacility();
		if (institution == null)
			return new ArrayList<>();
		return Arrays.asList(institution.getStationNumber());
	}

	@Transient
	@JsonIgnore
	public String getAJRMonthName() {
		if (this.annualJointReviewMonth != null) {
			return Month.of(annualJointReviewMonth.intValue()).name();
		} else {
			return "";
		}
	}

	@Override
	@Transient
	@JsonView(OrganizationView.Search.class)
	public String getScale() {
		return "Organization";
	}

	@Transient
	@JsonView(OrganizationView.Basic.class)
	public String getDisplayName() {
		return getName();
	}

	// -------------------------------------- Common Methods

	@Override
	protected boolean requiredEquals(AbstractBasicOrganization oo) {
		Organization o = (Organization) PersistenceUtil.initializeAndUnproxy(oo);
		return new EqualsBuilder().append(getCode(), o.getCode())
				.append(nullSafeGetId(getFacility()), nullSafeGetId(oo.getFacility())).isEquals();
	}

	@Override
	protected int requiredHashCode() {
		return new HashCodeBuilder().append(getCode()).append(nullSafeGetId(getFacility())).toHashCode();
	}

	public String toString() {
		return getName() + " (code " + getCode() + ", station ID " + nullSafeGetId(getFacility()) + ")";
	}

	// -------------------------------------- Accessor Methods

	@Column(name = "CodeNUM", length = 5)
	@JsonView(OrganizationView.Search.class)
	public int getCode() {
		return code;
	}

	public void setCode(int code) {
		this.code = code;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "VSS_STD_VolunteerOrganizationTypesFK")
	@JsonView(OrganizationView.Extended.class)
	public OrganizationType getType() {
		return type;
	}

	public void setType(OrganizationType type) {
		this.type = type;
	}

	@Column(name = "IsOnNationalAdvisoryCommittee", nullable = false)
	@JsonView(OrganizationView.Extended.class)
	public boolean isOnNationalAdvisoryCommittee() {
		return onNationalAdvisoryCommittee;
	}

	public void setOnNationalAdvisoryCommittee(boolean onNationalAdvisoryCommittee) {
		this.onNationalAdvisoryCommittee = onNationalAdvisoryCommittee;
	}

	@Column(name = "IsNACExecutiveMember", nullable = false)
	@JsonView(OrganizationView.Extended.class)
	public boolean isNacExecutiveMember() {
		return nacExecutiveMember;
	}

	public void setNacExecutiveMember(boolean nacExecutiveMember) {
		this.nacExecutiveMember = nacExecutiveMember;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "NACMembershipStatusFK")
	@JsonView(OrganizationView.Extended.class)
	public NACStatus getNacMembershipStatus() {
		return nacMembershipStatus;
	}

	public void setNacMembershipStatus(NACStatus nacMembershipStatus) {
		this.nacMembershipStatus = nacMembershipStatus;
	}

	@JsonView(OrganizationView.Extended.class)
	public Integer getAnnualJointReviewMonth() {
		return annualJointReviewMonth;
	}

	public void setAnnualJointReviewMonth(Integer annualJointReviewMonth) {
		this.annualJointReviewMonth = annualJointReviewMonth;
	}

	@OneToMany(mappedBy = "organization", fetch = FetchType.LAZY)
	@BatchSize(size = 500)
	@JsonIgnore
	public List<OrganizationBranch> getBranches() {
		if (branches == null)
			branches = new ArrayList<>();
		return branches;
	}

	public void setBranches(List<OrganizationBranch> branches) {
		this.branches = branches;
	}

}
