package gov.va.vss.persistence.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.persistence.Query;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import gov.va.shared.model.lookup.sds.VAFacility;
import gov.va.shared.persistence.lookup.sds.VAFacilityDAO;
import gov.va.shared.util.PersistenceUtil;
import gov.va.vss.model.facility.AbstractLocation;
import gov.va.vss.model.facility.Facility;
import gov.va.vss.persistence.dao.FacilityDAO;
import gov.va.vss.persistence.queryCustomization.QueryCustomization;

@Repository
public class FacilityDAOImpl extends GenericHibernateSortedDAOImpl<Facility> implements FacilityDAO {

	@Autowired
	private VAFacilityDAO vaFacilityDAO;

	// @Cacheable(value = CacheNames.QUERIES_FACILITY_DAO)
	public Facility findByStationNumber(String stationNumber) {
		@SuppressWarnings("unchecked")
		List<Facility> results = query("from " + Facility.class.getName() + " where stationNumber = :stationNumber")
				.setParameter("stationNumber", stationNumber).getResultList();
		return results.isEmpty() ? null : results.get(0);
	}

	@Override
	public Map<Long, Facility> findByVAFacilityIds(Collection<Long> vaFacilityIds) {
		Map<Long, Facility> results = new HashMap<>();

		if (CollectionUtils.isEmpty(vaFacilityIds))
			return results;

		List<Long> vaFacilityIdList = new ArrayList<>(vaFacilityIds);

		for (int i = 0; i < vaFacilityIdList.size(); i += 2000) {
			List<Long> batchChunk = vaFacilityIdList.subList(i, Math.min(vaFacilityIdList.size(), i + 2000));

			@SuppressWarnings("unchecked")
			List<Facility> queryResults = query("select i from " + Facility.class.getName()
					+ " i left join fetch i.vaFacility f where f.id in (:vaFacilityIds)")
							.setParameter("vaFacilityIds", batchChunk).getResultList();
			for (Facility r : queryResults)
				results.put(r.getVaFacility().getId(), r);
		}
		return results;
	}

	@Override
	public Map<Long, Facility> findByVAFacilities(Collection<VAFacility> vaFacilities) {
		return findByVAFacilityIds(PersistenceUtil.translateObjectsToIds(vaFacilities));
	}

	@Override
	// @Cacheable(value = CacheNames.QUERIES_FACILITY_DAO)
	public SortedSet<VAFacility> findVAFacilitiesWithLinkToFacility() {
		@SuppressWarnings("unchecked")
		List<VAFacility> results = query(
				"select i.vaFacility from " + Facility.class.getName() + " i where i.vaFacility is not null")
						.getResultList();
		return new TreeSet<>(results);
	}

	@Override
	// @Cacheable(value = CacheNames.QUERIES_FACILITY_DAO)
	public SortedSet<Facility> findWithLinkToVAFacility() {
		@SuppressWarnings("unchecked")
		List<Facility> results = query("select i from " + Facility.class.getName() + " i where vaFacility is not null")
				.getResultList();
		return new TreeSet<>(results);
	}

	@Override
	// @Cacheable(value = CacheNames.QUERIES_FACILITY_DAO)
	public Facility findByVAFacility(long vaFacilityId) {
		return findByVAFacilityIds(Arrays.asList(vaFacilityId)).get(vaFacilityId);
	}

	@Override
	// @CacheEvict(value = CacheNames.QUERIES_FACILITY_DAO,
	// inefficient but we shouldn't need to save these very often - CPB
	// allEntries = true)
	public Facility saveOrUpdate(Facility item) {
		return super.saveOrUpdate(item);
	}

	@Override
	public SortedSet<Facility> findRootFacilities() {
		@SuppressWarnings("unchecked")
		List<Facility> results = query(
				"select i from " + Facility.class.getName() + " i where parent is null or parent = i").getResultList();
		return new TreeSet<>(results);
	}

	@Override
	public void updateFieldsWithoutVersionIncrement(long facilityId, boolean setVAFacilityId, Long vaFacilityId) {
		if (!setVAFacilityId)
			throw new IllegalArgumentException("No update parameter was specified");

		List<String> updates = new ArrayList<>();
		Map<String, Object> params = new HashMap<>();

		if (setVAFacilityId) {
			VAFacility f = vaFacilityId != null ? vaFacilityDAO.findRequiredByPrimaryKey(vaFacilityId) : null;
			updates.add("vaFacility = :vaFacility");
			params.put("vaFacility", f);
		}

		StringBuilder sb = new StringBuilder("update " + Facility.class.getName() + " set ");
		sb.append(StringUtils.join(updates, ", "));
		sb.append(" where id = :facilityId");
		params.put("facilityId", facilityId);

		Query q = em.createQuery(sb.toString());
		for (Entry<String, Object> paramEntry : params.entrySet())
			q.setParameter(paramEntry.getKey(), paramEntry.getValue());
		int numUpdated = q.executeUpdate();

		if (numUpdated == 0)
			throw new IllegalStateException("No facility with ID " + facilityId + " found.");
	}

	@Override
	public <T extends AbstractLocation> SortedSet<T> findByCriteria(Long facilityOrLocationId, Long parentFacilityId, Class<T> requiredClassType) {
		if (parentFacilityId == null)
			throw new IllegalArgumentException("No criteria specified");

		StringBuilder sb = new StringBuilder("select v from ").append(requiredClassType.getName()).append(" v");

		/* Don't bother with this yet - CPB */
		QueryCustomization cust = new QueryCustomization();

		List<String> whereClauseItems = new ArrayList<>();
		Map<String, Object> params = new HashMap<>();

		if (parentFacilityId != null) {
			whereClauseItems.add("v.parent.id = :parentId");
			params.put("parentId", parentFacilityId);
		}

		Query q = constructQuery(em, sb, whereClauseItems, params, null, cust);

		int maxResults = -1;
		if (cust.getRowLimitation() != null)
			maxResults = cust.getRowLimitation();

		if (maxResults > 0)
			q.setMaxResults(maxResults);

		@SuppressWarnings("unchecked")
		List<T> resultList = q.getResultList();
		return new TreeSet<>(resultList);
	}

}
