package gov.va.vss.persistence.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.persistence.Query;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Repository;

import gov.va.vss.model.ScopeType;
import gov.va.vss.model.organization.AbstractBasicOrganization;
import gov.va.vss.model.organization.BasicOrganization;
import gov.va.vss.model.organization.Organization;
import gov.va.vss.model.organization.OrganizationBranch;
import gov.va.vss.persistence.dao.OrganizationDAO;
import gov.va.vss.persistence.queryCustomization.QueryCustomization;
import gov.va.vss.web.reports.OrgQuickSearchResult;

@Repository
public class OrganizationDAOImpl extends GenericHibernateSortedDAOImpl<AbstractBasicOrganization> implements OrganizationDAO {
	@SuppressWarnings("unused")
	private static final Logger log = LoggerFactory.getLogger(OrganizationDAOImpl.class);

	@Value("${organizationSearch.maxResults}")
	private int maxResults;

	@SuppressWarnings("unchecked")
	@Override
	public List<BasicOrganization> findByCriteria(String name, boolean includeNational, boolean includeLocalFacility,
			List<Long> localFacilityIds, Boolean activeStatus, Integer code, String abbreviation,
			QueryCustomization... customization) {
		if (!includeLocalFacility && !includeNational)
			return new ArrayList<>();

		List<String> whereClauseItems = new ArrayList<>();
		Map<String, Object> params = new HashMap<>();

		String[] nameTerms = new String[0];
		if (StringUtils.isNotBlank(name))
			nameTerms = name.split("\\s+");
		for (int i = 0; i < nameTerms.length; i++) {
			params.put("nameComp" + i, "%" + nameTerms[i].toLowerCase() + "%");
		}

		StringBuilder sb = new StringBuilder("select o from ").append(AbstractBasicOrganization.class.getName())
				.append(" o left join fetch o.organization parent").append(" left join o.facility facility");

		StringBuilder orgFrag = new StringBuilder("TYPE(o) = :organizationClass");
		params.put("organizationClass", Organization.class);

		for (int i = 0; i < nameTerms.length; i++)
			orgFrag.append(" and LOWER(o.name) like :nameComp" + i);

		if (code != null) {
			orgFrag.append(" and o.code = :code");
			params.put("code", code);
		}

		StringBuilder branchFrag = new StringBuilder("TYPE(o) = :branchClass");
		params.put("branchClass", OrganizationBranch.class);
		for (int i = 0; i < nameTerms.length; i++)
			branchFrag.append(" and CONCAT(parent.name, '-', o.name) like :nameComp" + i);

		QueryCustomization cust = ArrayUtils.isEmpty(customization) ? new QueryCustomization() : customization[0];
		cust.appendRemainingJoins(sb, "o");

		whereClauseItems.add("(" + orgFrag + ") or (" + branchFrag + ")");

		List<ScopeType> validScopeTypes = new ArrayList<>();
		if (includeLocalFacility) {
			validScopeTypes.add(ScopeType.LOCAL);
			if (localFacilityIds != null) {
				whereClauseItems.add("o.scope <> :localScope or COALESCE(facility.id, -1) in ( :localFacilityIds )");
				params.put("localFacilityIds", localFacilityIds);
				params.put("localScope", ScopeType.LOCAL);
			}
		}

		if (includeNational)
			validScopeTypes.add(ScopeType.NATIONAL);

		if (activeStatus != null) {
			whereClauseItems.add("o.inactive = :inactiveStatus");
			params.put("inactiveStatus", !activeStatus);
		}

		if (StringUtils.isNotBlank(abbreviation)) {
			whereClauseItems.add("LOWER(o.abbreviation) like :abbreviation");
			params.put("abbreviation", "%" + abbreviation.toLowerCase().trim() + "%");
		}

		whereClauseItems.add("o.scope in (:validScopes)");
		params.put("validScopes", validScopeTypes);

		if (cust.getOrderBy() == null)
			cust.setOrderBy("o.scope desc, "
					+ "case when TYPE(o) = :organizationClass then o.name else parent.name end,"
					+ "case when TYPE(o) = :organizationClass then '1' else '2' end,"
					+ "o.name");

		int maxResults = this.maxResults;
		if (cust.getRowLimitation() != null)
			maxResults = cust.getRowLimitation();

		Query q = constructQuery(em, sb, whereClauseItems, params, null, cust);

		return q.setMaxResults(maxResults).getResultList();
	}

	@Override
	public synchronized int findNextAvailableCode(ScopeType scope, Long localFacilityId) {
		Number result = null;

		if (scope == ScopeType.LOCAL) {
			if (localFacilityId == null)
				throw new IllegalArgumentException("Must specify the localFacilityId if scope = Local");

			result = (Number) query("select max(o.code) + 1 from " + Organization.class.getName()
					+ " o where o.scope = :localScope and o.facility.id = :facilityId")
							.setParameter("localScope", scope).setParameter("facilityId", localFacilityId)
							.getSingleResult();

			/*
			 * If scope = Local; then it will be assigned with the next
			 * available number in the sequence 900-9999.
			 */

			if (result == null || result.intValue() == 1)
				result = 900;

			if (result.intValue() == 10000) {
				throw new AssertionError("Unable assign local code, maximum allocation reached at 9999");
			}

		} else if (scope == ScopeType.NATIONAL) {
			if (localFacilityId != null)
				throw new IllegalArgumentException("Must not specify a localFacilityId if scope = National");

			result = (Number) query(
					"select max(o.code) + 1 from " + Organization.class.getName() + " o where o.scope = :nationalScope")
							.setParameter("nationalScope", scope).getSingleResult();

			/*
			 * If we run out of nationals from 1-899, jump to 10000 and allocate
			 * from there to 99999 - CPB
			 */
			if (result != null && result.intValue() == 900)
				result = 10000;
		} else {
			throw new AssertionError("Unexpected scope value " + scope);
		}

		return result == null ? 1 : result.intValue();
	}

	@Override
	public SortedSet<OrgQuickSearchResult> quickSearch(String searchValue, int maxResults) {
		if (StringUtils.isBlank(searchValue))
			return new TreeSet<>();

		String[] tokens = searchValue.split("\\W");

		StringBuilder sb = new StringBuilder();
		sb.append("select t.id, t.scope, t.code, t.name, t.abbreviation, i.name from ");
		sb.append(AbstractBasicOrganization.class.getName());
		sb.append(" t left join t.facility i where 1=1");

		Map<String, String> params = new HashMap<>();

		for (int i = 0; i < tokens.length; i++) {
			String token = tokens[i];
			if (StringUtils.isBlank(token))
				continue;

			sb.append(" and (lower(t.name) like :text").append(i);
			sb.append(" or t.code like :text").append(i);
			sb.append(" or lower(i.name) like :text").append(i).append(")");
			params.put("text" + i, "%" + token.toLowerCase() + "%");
		}
		Query q = query(sb.toString());
		for (Entry<String, String> entry : params.entrySet())
			q.setParameter(entry.getKey(), entry.getValue());

		if (maxResults > 0)
			q.setMaxResults(maxResults);

		@SuppressWarnings("unchecked")
		List<Object[]> results = q.getResultList();

		SortedSet<OrgQuickSearchResult> returnResults = new TreeSet<>();
		for (Object[] result : results) {
			returnResults.add(new OrgQuickSearchResult(((Number) result[0]).longValue(), (ScopeType) result[1],
					String.valueOf(result[2]), (String) result[3], (String) result[4], (String) result[5]));
		}
		return returnResults;
	}

}
