package gov.va.vss.persistence.impl;

import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.persistence.NoResultException;
import javax.persistence.Query;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.stereotype.Repository;

import gov.va.shared.util.StationsUtil;
import gov.va.vss.model.VistaStationData;
import gov.va.vss.persistence.dao.VistaStationDataDAO;
import gov.va.vss.persistence.queryCustomization.QueryCustomization;
import gov.va.vss.util.cache.CacheNames;

@Repository
public class VistaStationDataDAOImpl extends GenericHibernateDAOImpl<VistaStationData>implements VistaStationDataDAO {

	@Cacheable(CacheNames.QUERIES_VISTA_STATION_DATA_DAO)
	@Override
	public VistaStationData findByStationNumber(String stationNumber) {
		stationNumber = StationsUtil.getThreeDigitStationNumber(stationNumber);
		try {
			return (VistaStationData) query(
					"from " + VistaStationData.class.getName() + " where stationNumber = :stationNumber")
							.setParameter("stationNumber", stationNumber).getSingleResult();
		} catch (NoResultException e) {
			return null;
		}
	}

	@CacheEvict(value = CacheNames.QUERIES_VISTA_STATION_DATA_DAO, allEntries = true)
	@Override
	public void clearCache() {
	}

	@CacheEvict(value = CacheNames.QUERIES_VISTA_STATION_DATA_DAO, allEntries = true)
	@Override
	public int updateWithFields(String stationNumber, Boolean cprsHeaderSyncProcessEnabled,
			ZonedDateTime cprsHeaderSyncLastExecutionDate, Boolean patientInactivationProcessEnabled,
			ZonedDateTime patientInactivationLastExecutionDate, Boolean staffUpdateProcessEnabled,
			ZonedDateTime staffUpdateLastExecutionDate, Boolean vistaCleanupProcessEnabled,
			ZonedDateTime vistaCleanupLastExecutionDate, Boolean vistaSyncProcessEnabled) {
		if (stationNumber == null)
			throw new NullPointerException("Required field stationNumber missing");

		return updateInternal(stationNumber, cprsHeaderSyncProcessEnabled, cprsHeaderSyncLastExecutionDate,
				patientInactivationProcessEnabled, patientInactivationLastExecutionDate, vistaSyncProcessEnabled,
				vistaCleanupProcessEnabled, vistaCleanupLastExecutionDate, staffUpdateProcessEnabled,
				staffUpdateLastExecutionDate);
	}

	/*
	 * Allows for a null stationNumber
	 */
	private int updateInternal(String stationNumber, Boolean cprsHeaderSyncProcessEnabled,
			ZonedDateTime cprsHeaderSyncLastExecutionDate, Boolean patientInactivationProcessEnabled,
			ZonedDateTime patientInactivationLastExecutionDate, Boolean vistaSyncProcessEnabled,
			Boolean vistaCleanupProcessEnabled, ZonedDateTime vistaCleanupLastExecutionDate,
			Boolean staffUpdateProcessEnabled, ZonedDateTime staffUpdateLastExecutionDate) {
		if (patientInactivationProcessEnabled == null && cprsHeaderSyncProcessEnabled == null
				&& cprsHeaderSyncLastExecutionDate == null && patientInactivationLastExecutionDate == null
				&& vistaSyncProcessEnabled == null && vistaCleanupProcessEnabled == null
				&& vistaCleanupLastExecutionDate == null && staffUpdateProcessEnabled == null
				&& staffUpdateLastExecutionDate == null)
			throw new IllegalArgumentException("No updates specified");

		List<String> updateItems = new ArrayList<>();
		Map<String, Object> params = new HashMap<>();

		if (cprsHeaderSyncProcessEnabled != null) {
			updateItems.add("cprsHeaderSyncProcessEnabled = :cprsHeaderSyncProcessEnabled");
			params.put("cprsHeaderSyncProcessEnabled", cprsHeaderSyncProcessEnabled);
		}

		if (patientInactivationProcessEnabled != null) {
			updateItems.add("patientInactivationProcessEnabled = :patientInactivationProcessEnabled");
			params.put("patientInactivationProcessEnabled", patientInactivationProcessEnabled);
		}

		if (cprsHeaderSyncLastExecutionDate != null) {
			updateItems.add("cprsHeaderSyncLastExecutionDate = :cprsHeaderSyncLastExecutionDate");
			params.put("cprsHeaderSyncLastExecutionDate", cprsHeaderSyncLastExecutionDate);
		}

		if (patientInactivationLastExecutionDate != null) {
			updateItems.add("patientInactivationLastExecutionDate = :patientInactivationLastExecutionDate");
			params.put("patientInactivationLastExecutionDate", patientInactivationLastExecutionDate);
		}

		if (vistaSyncProcessEnabled != null) {
			updateItems.add("vistaSyncProcessEnabled = :vistaSyncProcessEnabled");
			params.put("vistaSyncProcessEnabled", vistaSyncProcessEnabled);
		}

		if (vistaCleanupProcessEnabled != null) {
			updateItems.add("vistaCleanupProcessEnabled = :vistaCleanupProcessEnabled");
			params.put("vistaCleanupProcessEnabled", vistaCleanupProcessEnabled);
		}

		if (vistaCleanupLastExecutionDate != null) {
			updateItems.add("vistaCleanupLastExecutionDate = :vistaCleanupLastExecutionDate");
			params.put("vistaCleanupLastExecutionDate", vistaCleanupLastExecutionDate);
		}

		if (staffUpdateProcessEnabled != null) {
			updateItems.add("staffUpdateProcessEnabled = :staffUpdateProcessEnabled");
			params.put("staffUpdateProcessEnabled", staffUpdateProcessEnabled);
		}

		if (staffUpdateLastExecutionDate != null) {
			updateItems.add("staffUpdateLastExecutionDate = :staffUpdateLastExecutionDate");
			params.put("staffUpdateLastExecutionDate", staffUpdateLastExecutionDate);
		}

		StringBuilder sb = new StringBuilder();
		sb.append("update ").append(VistaStationData.class.getName()).append(" set ");
		sb.append(StringUtils.join(updateItems, ", "));
		if (stationNumber != null)
			sb.append(" where stationNumber = :stationNumber");

		Query q = em.createQuery(sb.toString());
		if (stationNumber != null)
			q.setParameter("stationNumber", stationNumber);
		for (Map.Entry<String, Object> param : params.entrySet())
			q.setParameter(param.getKey(), param.getValue());

		int changes = q.executeUpdate();
		flushIfDebug();
		return changes;
	}

	@Override
	public List<VistaStationData> findByCriteria(Boolean cprsHeaderSyncProcessEnabled,
			ZonedDateTime cprsHeaderSyncLastExecutionDateBefore, Boolean patientInactivationProcessEnabled,
			QueryCustomization... customization) {
		StringBuilder sb = new StringBuilder("from ").append(VistaStationData.class.getName()).append(" t");

		if (ArrayUtils.isNotEmpty(customization))
			customization[0].appendRemainingJoins(sb, "t");

		List<String> whereClauseItems = new ArrayList<>();
		Map<String, Object> params = new HashMap<>();

		if (cprsHeaderSyncProcessEnabled != null) {
			whereClauseItems.add("t.cprsHeaderSyncProcessEnabled = :cprsHeaderSyncProcessEnabled");
			params.put("cprsHeaderSyncProcessEnabled", cprsHeaderSyncProcessEnabled);
		}

		if (cprsHeaderSyncLastExecutionDateBefore != null) {
			whereClauseItems.add("(t.cprsHeaderSyncLastExecutionDate is null or"
					+ " t.cprsHeaderSyncLastExecutionDate < :cprsHeaderSyncLastExecutionDate)");
			params.put("cprsHeaderSyncLastExecutionDate", cprsHeaderSyncLastExecutionDateBefore);
		}

		if (patientInactivationProcessEnabled != null) {
			whereClauseItems.add("t.patientInactivationProcessEnabled = :patientInactivationProcessEnabled");
			params.put("patientInactivationProcessEnabled", patientInactivationProcessEnabled);
		}

		Query q = constructQuery(em, sb, whereClauseItems, params, null, customization);

		@SuppressWarnings("unchecked")
		List<VistaStationData> dataResults = q.getResultList();

		return dataResults;
	}

	@Override
	@CacheEvict(value = CacheNames.QUERIES_VISTA_STATION_DATA_DAO, allEntries = true)
	public int updateCPRSHeaderSyncLastExecutionDatesForEnabledStations(Collection<String> stationNumbers,
			ZonedDateTime executionDate) {
		StringBuilder sb = new StringBuilder();
		sb.append("update ").append(VistaStationData.class.getName())
				.append(" set cprsHeaderSyncLastExecutionDate = :cprsHeaderSyncLastExecutionDate");
		sb.append(" where stationNumber in (:stationNumbers)");
		sb.append(" and cprsHeaderSyncProcessEnabled = true");

		Query q = em.createQuery(sb.toString());
		q.setParameter("cprsHeaderSyncLastExecutionDate", executionDate);
		q.setParameter("stationNumbers", stationNumbers);

		int changes = q.executeUpdate();
		flushIfDebug();
		return changes;
	}

}
