package gov.va.vss.persistence.impl;

import java.time.LocalDate;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.persistence.Query;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Repository;

import gov.va.vss.model.TimeAdjustment;
import gov.va.vss.model.Volunteer;
import gov.va.vss.model.VolunteerFacilityAssignment;
import gov.va.vss.model.WorkEntry;
import gov.va.vss.model.facility.Facility;
import gov.va.vss.persistence.dao.VolunteerDAO;
import gov.va.vss.persistence.queryCustomization.QueryCustomization;

@Repository
public class VolunteerDAOImpl extends GenericHibernateSortedDAOImpl<Volunteer> implements VolunteerDAO {
	@SuppressWarnings("unused")
	private static final Logger log = LoggerFactory.getLogger(VolunteerDAOImpl.class);

	@Value("${volunteerSearch.maxResults}")
	private int maxResults;

	@SuppressWarnings("unchecked")
	@Override
	public List<Volunteer> findByCriteria(String firstName, String middleName, String lastName, String code,
			LocalDate dateOfBirth, String addressStreet, String city, String state, String zip, String phone,
			String email, Boolean activeStatus, Collection<Long> vaFacilityIds, Collection<String> vaFacilitySta3ns,
			Collection<Long> facilityIds, QueryCustomization... customization) {
		boolean hasFacilityIds = CollectionUtils.isNotEmpty(vaFacilityIds);
		boolean hasFacilitySta3ns = CollectionUtils.isNotEmpty(vaFacilitySta3ns);
		boolean hasInstitutionIds = CollectionUtils.isNotEmpty(facilityIds);

		StringBuilder sb = new StringBuilder("select v from ").append(Volunteer.class.getName()).append(" v");

		if (hasFacilityIds || hasFacilitySta3ns || hasInstitutionIds) {
			sb.append(" join v.volunteerFacilityAssignments vfa");
			sb.append(" join vfa.facilityAssignment fa");
		}

		/* Don't bother with this yet - CPB */
		QueryCustomization cust = new QueryCustomization();

		List<String> whereClauseItems = new ArrayList<>();
		Map<String, Object> params = new HashMap<>();

		if (StringUtils.isNotBlank(firstName)) {
			whereClauseItems.add("LOWER(v.firstName) like :firstName");
			params.put("firstName", "%" + firstName.toLowerCase() + "%");
		}

		if (StringUtils.isNotBlank(lastName)) {
			whereClauseItems.add("LOWER(v.lastName) like :lastName");
			params.put("lastName", "%" + lastName.toLowerCase() + "%");
		}

		if (StringUtils.isNotBlank(code)) {
			whereClauseItems.add("LOWER(v.identifyingCode) like :identifyingCode");
			params.put("identifyingCode", "%" + code.toLowerCase() + "%");
		}

		if (dateOfBirth != null) {
			whereClauseItems.add("v.dateOfBirth = :dob");
			params.put("dob", dateOfBirth);
		}

		if (hasFacilitySta3ns) {
			whereClauseItems.add("fa.facility.stationNumber in (:vaFacilitySta3ns)");
			params.put("vaFacilitySta3ns", vaFacilitySta3ns);
		}

		if (hasInstitutionIds) {
			whereClauseItems.add("fa.facility.id in (:facilityIds)");
			params.put("facilityIds", facilityIds);
		}

		if (hasFacilityIds) {
			whereClauseItems.add("fa.facility.vaFacility.id in (:vaFacilityIds)");
			params.put("vaFacilityIds", vaFacilityIds);
		}

		if (StringUtils.isNotBlank(addressStreet)) {
			whereClauseItems
					.add("LOWER(v.addressLine1) like :addressStreet" + " or LOWER(v.addressLine2) like :addressStreet");
			params.put("addressStreet", "%" + addressStreet.toLowerCase() + "%");
		}

		if (StringUtils.isNotBlank(city)) {
			whereClauseItems.add("LOWER(v.city) like :city");
			params.put("city", "%" + city.toLowerCase() + "%");
		}

		if (activeStatus != null) {
			whereClauseItems.add("v.terminationDate is " + (activeStatus ? "" : "not") + " null");
		}

		/*
		 * Match on State abbreviation if specified value has only two letters;
		 * otherwise, wildcard match on state name
		 */
		if (StringUtils.isNotBlank(state)) {
			state = state.trim();
			if (state.trim().length() == 2) {
				whereClauseItems.add("LOWER(v.state.abbreviation) = :state");
				params.put("state", state.toLowerCase());
			} else {
				whereClauseItems.add("LOWER(v.state.name) like :state");
				params.put("state", "%" + state.toLowerCase() + "%");
			}
		}

		/* Match only on the first five digits */
		if (StringUtils.isNotBlank(zip)) {
			whereClauseItems.add("SUBSTRING(v.zip, 1, 5) = :zip");
			if (zip.length() > 5)
				zip = zip.substring(0, 5);
			params.put("zip", zip);
		}

		if (StringUtils.isNotBlank(phone)) {
			StringBuilder sb2 = new StringBuilder();
			for (String phoneField : new String[] { "phone", "phoneAlt", "phoneAlt2" })
				sb2.append(sb.length() > 0 ? " OR " : "") //
						.append("CONCAT(SUBSTRING(v.").append(phoneField).append(",1,3), ") //
						.append("SUBSTRING(v.").append(phoneField).append(", 5,3), ") //
						.append("SUBSTRING(v.").append(phoneField).append(", 10,4)) like :phone");
			whereClauseItems.add(sb2.toString());
			params.put("phone", "%" + phone.replaceAll("\\D", "") + "%");
		}

		if (StringUtils.isNotBlank(email)) {
			whereClauseItems.add("LOWER(v.email) like :email");
			params.put("email", "%" + email.toLowerCase() + "%");
		}

		Query q = constructQuery(em, sb, whereClauseItems, params, null, cust);

		int maxResults = this.maxResults;
		if (cust.getRowLimitation() != null)
			maxResults = cust.getRowLimitation();

		List<Volunteer> resultList = q.setMaxResults(maxResults).getResultList();
		return resultList;
	}

	@Override
	public List<String> findZipCodesAtFacilities(List<Long> facilityIds, Boolean volunteerActive,
			Boolean facilityAssignmentActive) {
		String query = "select distinct SUBSTRING(v.zip, 1, 5) from " + Volunteer.class.getName()
				+ " v join v.volunteerFacilityAssignments vfa join vfa.facilityAssignment fa"
				+ " where fa.facility.id in (:facilityIds)";
		if (volunteerActive != null)
			query += " and v.terminationDate is" + (!volunteerActive ? " not" : "") + " null";
		if (facilityAssignmentActive != null)
			query += " and fa.inactive = " + !facilityAssignmentActive;
		query += " order by SUBSTRING(v.zip, 1, 5)";

		@SuppressWarnings("unchecked")
		List<String> results = query(query).setParameter("facilityIds", facilityIds).getResultList();
		return results;
	}

	@SuppressWarnings("unchecked")
	@Override
	public SortedSet<Facility> findFacilitiesForVolunteer(long volunteerId) {
		return new TreeSet<>(query("select fa.facility from " + VolunteerFacilityAssignment.class.getName()
				+ " vfa join vfa.facilityAssignment fa where vfa.volunteer.id = :id").setParameter("id", volunteerId)
						.getResultList());
	}

	@Override
	public void updateFieldsWithoutVersionIncrement(long volunteerId, boolean setPrimaryOrganization,
			Long primaryOrganizationId, boolean setPrimaryFacility, Long primaryFacilityId) {
		if (!setPrimaryOrganization && !setPrimaryFacility)
			throw new IllegalArgumentException("No update parameter was specified");

		List<String> updates = new ArrayList<>();
		Map<String, Object> params = new HashMap<>();

		if (setPrimaryOrganization) {
			updates.add("PrimaryOrganizationFK = :primaryOrganizationId");
			params.put("primaryOrganizationId", primaryOrganizationId);
		}

		if (setPrimaryFacility) {
			updates.add("PrimaryFacilityFK = :primaryFacilityId");
			params.put("primaryFacilityId", primaryFacilityId);
		}

		StringBuilder sb = new StringBuilder("update VssVolunteers set ");
		sb.append(StringUtils.join(updates, ", "));
		sb.append(" where id = :volunteerId");
		params.put("volunteerId", volunteerId);

		Query q = em.createNativeQuery(sb.toString());
		for (Entry<String, Object> paramEntry : params.entrySet())
			q.setParameter(paramEntry.getKey(), paramEntry.getValue());
		int numUpdated = q.executeUpdate();

		if (numUpdated == 0)
			throw new IllegalStateException("No volunteer ID " + volunteerId + " found.");
	}

	@Override
	public TimeSummary getTimeSummary(long volunteerId) {
		LocalDate currentFYStart = dateUtil.getCurrentFiscalYearStartDate();
		Object[] result = (Object[]) query("select max(we.dateWorked), " //
				+ "COALESCE(sum(case when we.dateWorked < :currentFYStart then we.hoursWorked else 0 end), 0), " //
				+ "COALESCE(sum(case when we.dateWorked >= :currentFYStart then we.hoursWorked else 0 end), 0) " //
				+ " from " + WorkEntry.class.getName() //
				+ " we where we.volunteerAssignment.volunteer.id = :volunteerId")
						.setParameter("currentFYStart", currentFYStart).setParameter("volunteerId", volunteerId)
						.getSingleResult();
		LocalDate mostRecentDateWorked = (LocalDate) result[0];
		double priorYearHours = ((Number) result[1]).doubleValue();
		double currentYearHours = ((Number) result[2]).doubleValue();

		Object result2 = query("select COALESCE(sum(ta.hours), 0) from " + TimeAdjustment.class.getName()
				+ " ta where ta.volunteer.id = :volunteerId").setParameter("volunteerId", volunteerId)
						.getSingleResult();
		double adjustedHours = ((Number) result2).doubleValue();

		TimeSummary ts = new TimeSummary(mostRecentDateWorked, priorYearHours, currentYearHours,
				priorYearHours + currentYearHours + adjustedHours, adjustedHours);
		return ts;
	}

}
