package gov.va.vss.persistence.impl.jobExecution;

import static gov.va.vss.persistence.queryCustomization.fieldTypes.JobExecutionResultAssociationFieldType.REFERENCE_INSTITUTION;

import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;

import javax.persistence.Query;

import org.apache.commons.lang3.ArrayUtils;
import org.springframework.stereotype.Repository;

import gov.va.vss.model.jobExecution.JobExecutionResult;
import gov.va.vss.model.lookup.JobExecutionType;
import gov.va.vss.model.lookup.JobExecutionType.JobExecutionTypeValue;
import gov.va.vss.persistence.dao.jobExecution.JobExecutionResultDAO;
import gov.va.vss.persistence.impl.GenericHibernateSortedDAOImpl;
import gov.va.vss.persistence.queryCustomization.QueryCustomization;

@Repository
public class JobExecutionResultDAOImpl extends GenericHibernateSortedDAOImpl<JobExecutionResult>
		implements JobExecutionResultDAO {

	@SuppressWarnings("unchecked")
	@Override
	public List<JobExecutionResult> findByCriteria(Long jobExecutionId, Set<Long> validStationIds,
			JobExecutionType jobExecutionType, ZonedDateTime startDate, ZonedDateTime endDate,
			QueryCustomization... customization) {
		QueryCustomization c = ArrayUtils.isNotEmpty(customization) ? customization[0] : new QueryCustomization();

		StringBuilder sb = new StringBuilder("select t from ").append(JobExecutionResult.class.getName()).append(" t");

		c.appendRequiredJoin(sb, true, "t", REFERENCE_INSTITUTION, "ri");
		c.appendRemainingJoins(sb, "t");

		List<String> whereClauseItems = new ArrayList<>();
		Map<String, Object> params = new HashMap<>();

		if (jobExecutionId != null) {
			whereClauseItems.add("t.referenceJobExecution.id = :jobExecutionId");
			params.put("jobExecutionId", jobExecutionId);
		}

		if (validStationIds != null) {
			Set<Long> newStations = new HashSet<>(validStationIds);
			boolean includeStationlessJobs = newStations.remove(null);
			boolean hasAtLeastOneSpecificStation = false;
			for (Iterator<Long> it = newStations.iterator(); !hasAtLeastOneSpecificStation && it.hasNext();) {
				hasAtLeastOneSpecificStation = it.next() != null;
			}

			if (includeStationlessJobs || hasAtLeastOneSpecificStation) {
				StringBuilder sb2 = new StringBuilder();
				if (includeStationlessJobs)
					sb2.append("ri is null");
				if (includeStationlessJobs && hasAtLeastOneSpecificStation)
					sb2.append(" or ");
				if (hasAtLeastOneSpecificStation) {
					sb2.append("ri.id in (:stationIds)");
					params.put("stationIds", newStations);
				}
				whereClauseItems.add("(" + sb2.toString() + ")");
			} else {
				/*
				 * We passed in a valid but empty set of stations; must return
				 * no results - CPB
				 */
				return new ArrayList<>();
			}
		}

		if (jobExecutionType != null) {
			whereClauseItems.add("t.type.id = :jobExecutionTypeId");
			params.put("jobExecutionTypeId", jobExecutionType.getId());
		}

		if (startDate != null) {
			whereClauseItems.add("t.startTime >= :startTime");
			params.put("startTime", startDate);
		}

		if (endDate != null) {
			whereClauseItems.add("t.startTime <= :endTime");
			params.put("endTime", endDate);
		}

		if (c.getOrderBy() == null)
			c.setOrderBy("t.startTime desc");

		Query q = constructQuery(em, sb, whereClauseItems, params, null, c);

		return q.getResultList();
	}

	@Override
	public int deleteByCriteria(Long jobExecutionResultId, ZonedDateTime olderThanDate, JobExecutionTypeValue type,
			Long referenceJobExecutionId, String initiatorAppUserName) {
		StringBuilder queryBuilder = new StringBuilder();
		queryBuilder.append("select id from " + JobExecutionResult.class.getName() + " a");

		List<String> whereClauseItems = new ArrayList<>();
		Map<String, Object> paramMap = new HashMap<>();

		if (jobExecutionResultId != null) {
			whereClauseItems.add("a.id = :jobExecutionResultId");
			paramMap.put("jobExecutionResultId", jobExecutionResultId);
		}

		if (olderThanDate != null) {
			whereClauseItems.add("a.startTime < :cutoff");
			paramMap.put("cutoff", olderThanDate);
		}

		if (type != null) {
			whereClauseItems.add("a.type.id = :typeId");
			paramMap.put("typeId", type.getId());
		}

		if (referenceJobExecutionId != null) {
			whereClauseItems.add("a.referenceJobExecution.id = :referenceJobExecutionId");
			paramMap.put("referenceJobExecutionId", referenceJobExecutionId);
		}

		if (initiatorAppUserName != null) {
			whereClauseItems.add("a.initiator.userCredentials.userID = :initiatorAppUserName");
			paramMap.put("initiatorAppUserName", initiatorAppUserName);
		}

		for (ListIterator<String> it = whereClauseItems.listIterator(); it.hasNext();) {
			queryBuilder.append(it.hasPrevious() ? " and " : " where ").append(it.next());
		}

		String queryStr = "delete from " + JobExecutionResult.class.getName() + " ac where ac.id in ("
				+ queryBuilder.toString() + ")";

		Query q = query(queryStr);
		for (Map.Entry<String, Object> paramEntry : paramMap.entrySet())
			q.setParameter(paramEntry.getKey(), paramEntry.getValue());
		/*
		 * Return the number of Alerts deleted, we don't care about the children
		 * count
		 */
		int currResults = q.executeUpdate();
		flushIfDebug();
		return currResults;
	}

	@Override
	public JobExecutionResult findRequiredByJobExecutionId(long jobExecutionId, QueryCustomization... customization) {
		List<JobExecutionResult> jobExecResultList = findByCriteria(jobExecutionId, null, null, null, null,
				customization);
		if (jobExecResultList.isEmpty())
			throw new IllegalArgumentException("No " + JobExecutionResult.class.getSimpleName()
					+ " was found for the job execution ID " + jobExecutionId);
		return jobExecResultList.get(0);
	}

}
