package gov.va.vss.service.impl;

import static gov.va.shared.util.StringUtil.normalizeLineBreaks;

import java.time.LocalDate;
import java.time.ZonedDateTime;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang.WordUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;

import gov.va.shared.model.AppUser;
import gov.va.shared.model.lookup.sds.VAFacility;
import gov.va.shared.util.SessionUtil;
import gov.va.vss.model.FacilityAssignment;
import gov.va.vss.model.Volunteer;
import gov.va.vss.model.VolunteerFacilityAssignment;
import gov.va.vss.model.VolunteerOrganization;
import gov.va.vss.model.WorkEntry;
import gov.va.vss.model.facility.Facility;
import gov.va.vss.model.organization.AbstractBasicOrganization;
import gov.va.vss.service.VolunteerService;
import gov.va.vss.service.validation.ServiceValidationException;
import gov.va.vss.util.DateUtil;

@Service
public class VolunteerServiceImpl extends AbstractServiceImpl implements VolunteerService {
	private static final Logger log = LoggerFactory.getLogger(VolunteerServiceImpl.class);

	@Override
	public Volunteer saveOrUpdate(Volunteer volunteer, AppUser userContext) throws ServiceValidationException {
		Volunteer existingVolunteer = volunteer.isPersistent() ? volunteerDAO.findByPrimaryKey(volunteer.getId())
				: null;

		boolean appendUserToMainRemarks = StringUtils.isNotBlank(volunteer.getRemarks()) && (existingVolunteer == null
				|| existingVolunteer.getRemarks() == null || !normalizeLineBreaks(existingVolunteer.getRemarks())
						.equals(normalizeLineBreaks(volunteer.getRemarks())));
		boolean appendUserToMealsRemarks = StringUtils.isNotBlank(volunteer.getMealRemarks())
				&& (existingVolunteer == null || existingVolunteer.getMealRemarks() == null
						|| !normalizeLineBreaks(existingVolunteer.getMealRemarks())
								.equals(normalizeLineBreaks(volunteer.getMealRemarks())));
		boolean appendUserToTerminationRemarks = StringUtils.isNotBlank(volunteer.getTerminationRemarks())
				&& (existingVolunteer == null || existingVolunteer.getTerminationRemarks() == null
						|| !normalizeLineBreaks(existingVolunteer.getTerminationRemarks())
								.equals(normalizeLineBreaks(volunteer.getTerminationRemarks())));

		if (!volunteer.isPersistent()) {
			volunteer.setEntryDate(LocalDate.now());

			VAFacility f = SessionUtil.getSiteContext();
			if (f != null) {
				Facility i = facilityDAO.findByVAFacility(f.getId());
				volunteer.setPrimaryFacility(i);
			}
		}
		volunteer.setFirstName(WordUtils.capitalizeFully(volunteer.getFirstName()));
		volunteer.setMiddleName(WordUtils.capitalizeFully(volunteer.getMiddleName()));
		volunteer.setLastName(WordUtils.capitalizeFully(volunteer.getLastName()));
		volunteer.setSuffix(WordUtils.capitalizeFully(volunteer.getSuffix()));

		if (appendUserToMainRemarks)
			volunteer.setRemarks(volunteer.getRemarks() + getAuditText(userContext));
		if (appendUserToMealsRemarks)
			volunteer.setMealRemarks(volunteer.getMealRemarks() + getAuditText(userContext));
		if (appendUserToTerminationRemarks)
			volunteer.setTerminationRemarks(volunteer.getTerminationRemarks() + getAuditText(userContext));

		return volunteerDAO.saveOrUpdate(volunteer);
	}

	private String getAuditText(AppUser appUser) {
		return "\n - " + appUser.getDisplayName() + " [" + appUser.getUsername() + "] @"
				+ ZonedDateTime.now(appUser.getTimeZone()).format(DateUtil.MILITARY_DATE_TIME_WITH_ZONE_FORMAT);
	}

	@Override
	public void delete(long volunteerId) {
		volunteerDAO.delete(volunteerId);
	}

	@Override
	public void setPrimaryOrganization(long volunteerId, long organizationId) {
		volunteerDAO.updateFieldsWithoutVersionIncrement(volunteerId, true, organizationId, false, null);
	}

	@Override
	public void setPrimaryFacility(long volunteerId, long facilityId) {
		volunteerDAO.updateFieldsWithoutVersionIncrement(volunteerId, false, null, true, facilityId);
	}

	@Override
	public void addOrganization(long volunteerId, long organizationId) throws ServiceValidationException {
		List<VolunteerOrganization> list = volunteerOrganizationDAO.findByCriteria(volunteerId, organizationId);
		VolunteerOrganization vo = null;
		Volunteer v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);

		if (list.isEmpty()) {
			AbstractBasicOrganization o = organizationDAO.findRequiredByPrimaryKey(organizationId);
			if (o.isInactive())
				throw new ServiceValidationException("volunteer.error.organization.addInactiveOrganization");
			vo = new VolunteerOrganization(v, o);
		} else {
			vo = list.get(0);
			AbstractBasicOrganization o = vo.getOrganization();
			if (o.isInactive())
				throw new ServiceValidationException("volunteer.error.organization.reactivateInactiveOrganization");
			vo.setInactive(false);
		}

		vo = volunteerOrganizationDAO.saveOrUpdate(vo);

		if (v.getPrimaryOrganization() == null)
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), true, organizationId, false, null);
	}

	@Override
	public boolean addFacilityAssignment(long volunteerId, long facilityAssignmentId) {
		FacilityAssignment fa = facilityAssignmentDAO.findRequiredByPrimaryKey(facilityAssignmentId);
		if (fa.isInactive())
			return false;

		Volunteer v = volunteerDAO.findRequiredByPrimaryKey(volunteerId);

		List<VolunteerFacilityAssignment> list = volunteerFacilityAssignmentDAO.findByCriteria(volunteerId,
				facilityAssignmentId, null);
		VolunteerFacilityAssignment vfa;

		if (list.isEmpty()) {
			vfa = new VolunteerFacilityAssignment(v, fa);
		} else {
			vfa = list.get(0);
			vfa.setInactive(false);
		}

		vfa = volunteerFacilityAssignmentDAO.saveOrUpdate(vfa);

		if (v.getPrimaryFacility() == null)
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), false, null, true, fa.getFacility().getId());

		return true;
	}

	@Override
	public boolean deleteOrInactivateOrganization(long volunteerOrganizationId) {
		VolunteerOrganization vo = volunteerOrganizationDAO.findRequiredByPrimaryKey(volunteerOrganizationId);
		Volunteer v = vo.getVolunteer();

		AbstractBasicOrganization currentPrimary = v.getPrimaryOrganization();
		AbstractBasicOrganization orgBeingInactivated = vo.getOrganization();

		boolean reportPrimaryInactivated = orgBeingInactivated.equals(currentPrimary);

		List<WorkEntry> workEntries = workEntryDAO.findByCriteria(v.getId(), orgBeingInactivated.getId());
		if (workEntries.isEmpty()) {
			volunteerOrganizationDAO.delete(volunteerOrganizationId);
		} else {
			vo.setInactive(true);
			vo = volunteerOrganizationDAO.saveOrUpdate(vo);
		}

		volunteerDAO.flushAndRefresh(v);

		Set<VolunteerOrganization> organizationsByStatus = v.getVolunteerOrganizationsByStatus(true);
		if (organizationsByStatus.size() == 1) {
			AbstractBasicOrganization organization = organizationsByStatus.iterator().next().getOrganization();
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), true, organization.getId(), false, null);
			reportPrimaryInactivated = false;
		} else {
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), true, null, false, null);
		}

		if (organizationsByStatus.size() == 0)
			reportPrimaryInactivated = false;

		return reportPrimaryInactivated;
	}

	@Override
	public boolean deleteAssignment(long volunteerFacilityAssignmentId) {
		VolunteerFacilityAssignment vo = volunteerFacilityAssignmentDAO
				.findRequiredByPrimaryKey(volunteerFacilityAssignmentId);
		Volunteer v = vo.getVolunteer();

		Facility currentPrimary = v.getPrimaryFacility();
		Facility facBeingInactivated = vo.getFacilityAssignment().getFacility();

		boolean reportPrimaryInactivated = facBeingInactivated.equals(currentPrimary);

		if (vo.getWorkEntries().isEmpty()) {
			volunteerFacilityAssignmentDAO.delete(volunteerFacilityAssignmentId);
		} else {
			vo.setInactive(true);
			vo = volunteerFacilityAssignmentDAO.saveOrUpdate(vo);
		}

		volunteerDAO.flushAndRefresh(v);

		Set<VolunteerFacilityAssignment> assignmentsByStatus = v.getFacilityAssignmentsByStatus(true);
		Set<Facility> uniqueFacilities = new HashSet<>();
		for (VolunteerFacilityAssignment a : assignmentsByStatus)
			uniqueFacilities.add(a.getFacilityAssignment().getFacility());

		if (uniqueFacilities.size() == 1) {
			Facility facilityAssignment = uniqueFacilities.iterator().next();
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), false, null, true, facilityAssignment.getId());
			reportPrimaryInactivated = false;
		} else {
			volunteerDAO.updateFieldsWithoutVersionIncrement(v.getId(), false, null, true, null);
		}

		if (assignmentsByStatus.size() == 0)
			reportPrimaryInactivated = false;

		return reportPrimaryInactivated;
	}

}
